/************************************************************************************
TerraLib - a library for developing GIS applications.
Copyright  2001-2007 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular purpose.
The library provided hereunder is on an "as is" basis, and the authors have no
obligation to provide maintenance, support, updates, enhancements, or modifications.
In no event shall INPE and Tecgraf / PUC-Rio be held liable to any party for direct,
indirect, special, incidental, or consequential damages arising out of the use
of this library and its documentation.
*************************************************************************************/
/*! \file TeDriverCSV.h
    \brief This file contains functions to deal with a file in CSV format
*/
#ifndef __TERRALIB_INTERNAL_DRIVERCSV_H
#define __TERRALIB_INTERNAL_DRIVERCSV_H

#include "TeTable.h"
#include "TeBox.h"
#include "TeDataTypes.h"

#include <string>
using namespace std;

class TeLayer;
class TeAsciiFile;
class TeProjection;
class TeDatabase;

/** @defgroup CSV format
 *  @ingroup Formats
 *  Functions related to CSV format
 *  @{
 */
/** Creates an attribute table in a TerraLib database and imports a CSV file to this table 
	\param csvFileName			CSV file name
	\param attrTable			attribute table specifications
	\param db					pointer to a TerraLib database
	\param layer				pointer to a layer to which the table will associated
	\param separator			character used to separate the data in the CSV file 
	\param firstLineIsHeader	indicates if the first line is a header
	\param chunkSize			number of objects in memory before save them in the database
	\param createAutoNum		indicates if the table first attribute is an autonumber which must be generated  
	\return TRUE if the data was imported successfully and FALSE otherwise
*/
bool TeImportCSVFile(const string& csvFileName, TeTable& attrTable, TeDatabase* db, TeLayer* layer=0, 
					 const char& separator=',', const bool& firstLineIsHeader=false,   
					 const unsigned int& chunkSize=100, const bool& createAutoNum = false);

/** Imports a CSV file to a layer in a TerraLib database, its geometries (points) and attributes
	\param csvFileName			CSV file name pointer to a layer
	\param attributeX			attribute index which provides the coordinate X
	\param attributeY			attribute index which provides the coordinate Y
	\param attrTable			attribute table specifications
	\param db					pointer to a TerraLib database
	\param layer				pointer to an empty layer which will be filled from CSV file
	\param separator			character used to separate the data in the CSV file 
	\param firstLineIsHeader	indicates if the first line is a header
	\param chunkSize			number of objects in memory before save them in the database
	\param createAutoNum		indicates if the table first attribute is an autonumber which must be generated  
	\return TRUE if the data was imported successfully and FALSE otherwise
*/
bool TeImportCSVFile(const string& csvFileName, const int& attributeX, const int& attributeY, 
					 TeTable& attrTable, TeDatabase* db, TeLayer* layer, 
					 const char& separator=',', const bool& firstLineIsHeader=false,   
					 const unsigned int& chunkSize=100, const bool& createAutoNum = false);

/** @} */ // End of subgroup CSV format

/** \example importCSV.cpp
 Shows how to import a MID/MIF data file into a new layer and a CSV file as a second attribute table of the layer
*/

#endif

