#include "TePDIColorTransform.hpp"

#include <TeAgnostic.h>
#include "TePDIUtils.hpp"

#include <TeDefines.h>



TePDIColorTransform::TePDIColorTransform()
{
}


TePDIColorTransform::~TePDIColorTransform()
{
}


void TePDIColorTransform::ResetState( const TePDIParameters& params )
{
  TePDILevelRemap::ResetState( params );
}


bool TePDIColorTransform::CheckParameters( 
  const TePDIParameters& parameters ) const
{
  /* Checking transformation type */
  ColorTransfTypes transf_type;
  TEAGN_TRUE_OR_RETURN( parameters.GetParameter( "transf_type", transf_type ),
    "Missing parameter: transf_type" );
  TEAGN_TRUE_OR_RETURN(
    ( ( transf_type == Rgb2Ihs ) || ( transf_type == Ihs2Rgb ) ),
    "Invalid parameter: transf_type" );
    
  /* Checking input raster */

  TePDITypes::TePDIRasterPtrType input_image;
  if( ! parameters.GetParameter( "input_image", input_image ) ) {

    TEAGN_LOGERR( "Missing parameter: input_image" );
    return false;
  }
  if( ! input_image.isActive() ) {

    TEAGN_LOGERR( "Invalid parameter: input_image inactive" );
    return false;
  }
  if( input_image->params().status_ == TeRasterParams::TeNotReady ) {

    TEAGN_LOGERR( "Invalid parameter: input_image not ready" );
    return false;
  }
  TEAGN_TRUE_OR_RETURN( input_image->nBands() == 3,
    "Invalid number of bands from input image" );
    
  /* Checking input data type */
    
  TEAGN_TRUE_OR_RETURN( ( ( input_image->params().dataType_[ 0 ] ==
    input_image->params().dataType_[ 1 ] ) &&
    ( input_image->params().dataType_[ 1 ] ==
    input_image->params().dataType_[ 2 ] ) ),
    "Data type mismatch between bands for input_image" );
    
  if( transf_type == Rgb2Ihs ) {
    TEAGN_TRUE_OR_RETURN( input_image->params().dataType_[ 0 ] != TeDOUBLE,
      "Input double rasters not allowed for this algorithm" );
    TEAGN_TRUE_OR_RETURN( input_image->params().dataType_[ 1 ] != TeDOUBLE,
      "Input double rasters not allowed for this algorithm" );
    TEAGN_TRUE_OR_RETURN( input_image->params().dataType_[ 2 ] != TeDOUBLE,
      "Input double rasters not allowed for this algorithm" );        
  }
  
  /* Checking output raster */

  TePDITypes::TePDIRasterPtrType output_image;
  if( ! parameters.GetParameter( "output_image", output_image ) ) {

    TEAGN_LOGERR( "Missing parameter: output_image" );
    return false;
  }
  if( ! output_image.isActive() ) {

    TEAGN_LOGERR( "Invalid parameter: output_image inactive" );
    return false;
  }
  if( output_image->params().status_ == TeRasterParams::TeNotReady ) {

    TEAGN_LOGERR( "Invalid parameter: output_image not ready" );
    return false;
  }
  
  /* Checking photometric interpretation */
  
  for( unsigned int channel = 0 ; channel < 3 ; ++channel ) {
    TEAGN_TRUE_OR_RETURN( ( 
      ( input_image->params().photometric_[ channel ] == 
        TeRasterParams::TeRGB ) ||
      ( input_image->params().photometric_[ channel ] == 
        TeRasterParams::TeMultiBand ) ),
    "Invalid parameter - input_image (invalid photometric interpretation)" );
  }

  return true;
}


void TePDIColorTransform::MapRgb2Hsl(
  double red, double green, double blue,
  double& hue, double& sat, double& light )
{
  hue = sat = light = 0;

  double max = MAX( red, MAX( green , blue ) );
  double min = MIN( red, MIN( green , blue ) );
  
  double sum = min + max;

  light = sum / 2.;

  if( light <= 0. ) {
    return;
  }

  double range = ( max - min );

  if( range > 0 ) {
    sat = range / ( ( light <= 0.5 ) ? sum : ( 2. - max - min ) );
  } else {
    sat = range;
    return;
  }

  double internal_red = ( max - red ) / range;
  double internal_green = ( max - green ) / range;
  double internal_blue = ( max - blue ) / range;

  if( red == max ) { /* red channel is the brightest */
    hue = ( ( green == min ) ? ( 5. + internal_blue ) :
      ( 1. - internal_green ) );
  } else if( green == max ) { /* green channel is the brightest */
    hue = ( ( blue == min ) ? ( 1. + internal_red ) :
      ( 3. - internal_blue ) );
  } else { /* blue channel is the brightest */
    hue = ( ( red == min ) ? ( 3. + internal_green ) :
      ( 5. - internal_red ) );
  }

  hue /= 6.;
}


void TePDIColorTransform::MapHsl2Rgb(
  double hue, double sat, double light,
  double& red, double& green, double& blue )
{
  double value = ( light <= 5. ) ? ( light * ( 1. + sat ) ) :
    ( light + sat - ( light * sat ) );

  if( value <= 0. ) {
    red = green = blue = 0.;
  } else {
    double min;
    double sv;
    int sextant;
    double fract, vsf, mid1, mid2;

    min = light + light - value;
    sv = (value - min ) / value;
    hue *= 6.0;
    sextant = (int)hue;
    fract = hue - sextant;
    vsf = value * sv * fract;
    mid1 = min + vsf;
    mid2 = value - vsf;

    switch (sextant) {
      case 1: red = mid2;  green = value; blue = min; break;
      case 2: red = min;   green = value; blue = mid1; break;
      case 3: red = min;   green = mid2;  blue = value; break;
      case 4: red = mid1;  green = min;   blue = value; break;
      case 5: red = value; green = min;   blue = mid2; break;
      default : red = value; green = mid1;  blue = min; break;
    }
  }
}


bool TePDIColorTransform::RunImplementation()
{
  TePDITypes::TePDIRasterPtrType input_image;
  params_.GetParameter( "input_image", input_image );

  TePDITypes::TePDIRasterPtrType output_image;
  params_.GetParameter( "output_image", output_image );

  ColorTransfTypes transf_type;
  params_.GetParameter( "transf_type", transf_type );

  switch( transf_type ) {
    case Rgb2Ihs :
    {
      return RunRgb2Ihs( input_image, output_image );
      break;
    }
    case Ihs2Rgb :
    {
      return RunIhs2Rgb( input_image, output_image );
      break;
    }
    default :
    {
      TEAGN_LOG_AND_RETURN( "Invalid transformation type" );
      break;
    }
  }

  return false;
}


bool TePDIColorTransform::RunRgb2Ihs(
  TePDITypes::TePDIRasterPtrType input_image,
  TePDITypes::TePDIRasterPtrType output_image )
{
  unsigned int lines = ( unsigned int ) input_image->params().nlines_;
  unsigned int columns = ( unsigned int ) input_image->params().ncols_;
  
  /* Reseting output raster */
  
  TeRasterParams outRaster_params = output_image->params();
  
  outRaster_params.nBands( 3 );
  if( input_image->projection() != 0 ) {
    TeSharedPtr< TeProjection > proj( TeProjectionFactory::make( 
      input_image->projection()->params() ) );  
    outRaster_params.projection( proj.nakedPointer() );
  }
  outRaster_params.boxResolution( input_image->params().box().x1(), 
    input_image->params().box().y1(), input_image->params().box().x2(), 
    input_image->params().box().y2(), input_image->params().resx_, 
    input_image->params().resy_ );  
  outRaster_params.setPhotometric( TeRasterParams::TeMultiBand, -1 );

  TEAGN_TRUE_OR_RETURN( output_image->init( outRaster_params ),
    "Output raster reset error" );  
  
  double min, max;
  TEAGN_TRUE_OR_RETURN( TePDIUtils::TeGetRasterMinMaxBounds(
    input_image, 0, min, max ),
    "Unable to get channels limits" );
  double diff = max - min;
    
  double red, green, blue;
  double hue, sat, light;
  
  StartProgInt( "Converting RGB -> IHS...", lines );  
  
  TeRaster* output_image_nptr = output_image.nakedPointer();
  TeRaster* input_image_nptr = input_image.nakedPointer();
  
  /* Dummy use definition */
  
  bool in_raster_use_dummy = input_image_nptr->params().useDummy_;
  bool out_raster_use_dummy = output_image_nptr->params().useDummy_;
  double dummy_i = 0;
  double dummy_h = 0;
  double dummy_s = 0;
  if( out_raster_use_dummy ) {
    dummy_i = output_image_nptr->params().dummy_[ 0 ];
    dummy_h = output_image_nptr->params().dummy_[ 1 ];
    dummy_s = output_image_nptr->params().dummy_[ 2 ];
  }
  
  /* Raster convertion */

  for( unsigned int line = 0 ; line < lines ; ++line ) {
    TEAGN_FALSE_OR_RETURN( UpdateProgInt( line ), "Canceled by the user" );
  
    for( unsigned int column = 0 ; column < columns ; ++column ) {
      if( input_image_nptr->getElement( column, line, red, 0 ) && 
          input_image_nptr->getElement( column, line, green, 1 ) &&
          input_image_nptr->getElement( column, line, blue, 2 ) ) {
          
        MapRgb2Hsl( ( red - min ) / diff, ( green - min ) / diff,
                    ( blue - min ) / diff, hue, sat, light );      
                    
        TEAGN_TRUE_OR_RETURN( output_image_nptr->setElement(
          column, line, ( light * diff ) + min, 0 ),
          "Unable to write intensity channel for output_image" );
        TEAGN_TRUE_OR_RETURN( output_image_nptr->setElement(
          column, line, ( hue * diff ) + min, 1 ),
        "Unable to write hue channel for output_image" );
        TEAGN_TRUE_OR_RETURN( output_image_nptr->setElement(
          column, line, ( sat * diff ) + min, 2 ),
          "Unable to write saturation channel for output_image" );
      } else {
        TEAGN_TRUE_OR_RETURN( in_raster_use_dummy, "Raster read error" );
        
        TEAGN_TRUE_OR_RETURN( output_image_nptr->setElement(
          column, line, dummy_i, 0 ),
          "Unable to write intensity channel for output_image" );
        TEAGN_TRUE_OR_RETURN( output_image_nptr->setElement(
          column, line, dummy_h, 1 ),
        "Unable to write hue channel for output_image" );
        TEAGN_TRUE_OR_RETURN( output_image_nptr->setElement(
          column, line, dummy_s, 2 ),
          "Unable to write saturation channel for output_image" );
      }
    }
  }

  return true;
}


bool TePDIColorTransform::RunIhs2Rgb(
  TePDITypes::TePDIRasterPtrType input_image,
  TePDITypes::TePDIRasterPtrType output_image )
{
  unsigned int lines = ( unsigned int ) input_image->params().nlines_;
  unsigned int columns = ( unsigned int ) input_image->params().ncols_;
  
  /* Reseting output raster */
  
  TeRasterParams outRaster_params = output_image->params();
  
  outRaster_params.nBands( 3 );
  if( input_image->projection() != 0 ) {
    TeSharedPtr< TeProjection > proj( TeProjectionFactory::make( 
      input_image->projection()->params() ) );  
    outRaster_params.projection( proj.nakedPointer() );
  }
  outRaster_params.boxResolution( input_image->params().box().x1(), 
    input_image->params().box().y1(), input_image->params().box().x2(), 
    input_image->params().box().y2(), input_image->params().resx_, 
    input_image->params().resy_ );  
  outRaster_params.setPhotometric( TeRasterParams::TeMultiBand, -1 );

  TEAGN_TRUE_OR_RETURN( output_image->init( outRaster_params ),
    "Output raster reset error" );   

  double out_min, out_max;
  TEAGN_TRUE_OR_RETURN( TePDIUtils::TeGetRasterMinMaxBounds(
    output_image, 0, out_min, out_max ),
    "Unable to get output channels limits" );
  double out_diff = out_max - out_min;
  
  TeRaster* output_image_nptr = output_image.nakedPointer();
  TeRaster* input_image_nptr = input_image.nakedPointer();
  
  /* Dummy use definition */
  
  bool in_raster_use_dummy = input_image_nptr->params().useDummy_;
  bool out_raster_use_dummy = output_image_nptr->params().useDummy_;
  double dummy_r = 0;
  double dummy_g = 0;
  double dummy_b = 0;  
  if( out_raster_use_dummy ) {
    dummy_r = output_image_nptr->params().dummy_[ 0 ];
    dummy_g = output_image_nptr->params().dummy_[ 1 ];
    dummy_b = output_image_nptr->params().dummy_[ 2 ];
  }
  
  /* Raster convertion */

  double red, green, blue;
  double hue, sat, intensity;
  
  StartProgInt( "Converting IHS -> RGB...", lines );
  
  for( unsigned int line = 0 ; line < lines ; ++line ) {
    TEAGN_FALSE_OR_RETURN( UpdateProgInt( line ), "Canceled by the user" );
  
    for( unsigned int column = 0 ; column < columns ; ++column ) {
      if( input_image_nptr->getElement( column, line, intensity, 0 ) &&
          input_image_nptr->getElement( column, line, hue, 1 ) &&
          input_image_nptr->getElement( column, line, sat, 2 ) ) {

        MapHsl2Rgb( hue / out_diff, sat / out_diff, intensity / out_diff, 
          red, green, blue );
  
        TEAGN_TRUE_OR_RETURN( output_image_nptr->setElement(
          column, line, ( red * out_diff ) + out_min, 0 ),
          "Unable to write red channel for output_image" );
        TEAGN_TRUE_OR_RETURN( output_image_nptr->setElement(
          column, line, ( green * out_diff ) + out_min, 1 ),
          "Unable to write green channel for output_image" );
        TEAGN_TRUE_OR_RETURN( output_image_nptr->setElement(
          column, line, ( blue * out_diff ) + out_min, 2 ),
          "Unable to write blue channel for output_image" );
      } else {                    
        TEAGN_TRUE_OR_RETURN( in_raster_use_dummy, "Raster read error" );
      
        TEAGN_TRUE_OR_RETURN( output_image_nptr->setElement(
          column, line, dummy_r, 0 ),
          "Unable to write red channel for output_image" );
        TEAGN_TRUE_OR_RETURN( output_image_nptr->setElement(
          column, line, dummy_g, 1 ),
          "Unable to write green channel for output_image" );
        TEAGN_TRUE_OR_RETURN( output_image_nptr->setElement(
          column, line, dummy_b, 2 ),
          "Unable to write blue channel for output_image" );
      }
    }
  }

  return true;
}

