/*
TerraLib - a library for developing GIS applications.
Copyright  2001, 2002, 2003 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular
purpose. The library provided hereunder is on an "as is" basis, and the
authors have no obligation to provide maintenance, support, updates,
enhancements, or modifications.
In no event shall INPE be held liable to any party
for direct, indirect, special, incidental, or consequential damages arising
out of the use of this library and its documentation.
*/

#ifndef TEPDICOLORTRANSFORM_HPP
  #define TEPDICOLORTRANSFORM_HPP

  #include "TePDILevelRemap.hpp"
  #include "TePDIParameters.hpp"
  #include <TeSharedPtr.h>

  /**
   * @brief This is the class for raster color transforms.
   * @author Emiliano F. Castejon <castejon@dpi.inpe.br>
   *
   * @note The general required parameters are:
   *
   * @param input_image ( TePDITypes::TePDIRasterPtrType ) - 3 bands,
   * @param output_image ( TePDITypes::TePDIRasterPtrType ).
   * @param transf_type ( TePDIColorTransform::ColorTransfTypes ) - 
   * Transformation type.
   *
   * @note Channels disposition for RGB images:
   * Band0=Red , Band1=Green , Band2=Blue
   *   
   * @note Channels disposition for IHS images:
   * Band0=Intensity , Band1=Hue , Band2=Saturation
   * 
   * @ingroup PDIAlgorithms
   */
  class TePDIColorTransform : public TePDILevelRemap {
    public :

      /**
       * @enum ColorTransfTypes Color transformation types.
       */
      enum ColorTransfTypes {
        /** RGB -> IHS ( the input channels range will be used as 
          * normalization factors ) */
        Rgb2Ihs = 1,
        /** IHS -> RGB ( the output channels range will be used as 
          * normalization factors ) */
        Ihs2Rgb = 2
      };

      /**
       * @brief Default Constructor.
       *
       */
      TePDIColorTransform();

      /**
       * @briefDefault Destructor
       */
      ~TePDIColorTransform();

      /**
       * @brief Checks if the supplied parameters fits the requirements of each
       * PDI algorithm implementation.
       *
       * @note Error log messages must be generated. No exceptions generated.
       *
       * @param parameters The parameters to be checked.
       * @return true if the parameters are OK. false if not.
       */
      bool CheckParameters( const TePDIParameters& parameters ) const;

    protected :

      /**
       * @brief Runs the current algorithm implementation.
       *
       * @return true if OK. false on error.
       */
      bool RunImplementation();

      /**
       * @brief Reset the internal state to the initial state.
       *
       * @param params The new parameters referente at initial state.
       */
      void ResetState( const TePDIParameters& params );

      /**
       * @brief Colorspace mapping from RGB ( [0,1] normalized ) to
       * HSL( [0,1] normalized ).
       *
       * @param hue Hue.
       * @param sat Saturation.
       * @param light Light.
       * @param red Red.
       * @param green Green.
       * @param blue Blue.
       */
      void MapRgb2Hsl( double red, double green, double blue,
                       double& hue, double& sat, double& light );

      /**
       * @brief Colorspace mapping from HSL ( [0,1] normalized ) to
       * RGB( [0,1] normalized ).
       *
       * @param hue Hue.
       * @param sat Saturation.
       * @param light Light.
       * @param red Red.
       * @param green Green.
       * @param blue Blue.
       */
      void MapHsl2Rgb( double hue, double sat, double light,
                       double& red, double& green, double& blue );

      /**
       * @brief Runs the RGB -> IHS implementation.
       *
       * @note Error log messages will be generated. No exceptions generated.
       *
       * @param input_image Input image ( band 0 -> red, band 1 -> green
       * band 2 -> blue ).
       * @param output_image Input image ( band 0 -> intensity,
       * band 1 -> hue, band 2 -> saturation ).
       * @return true if OK. false if not.
       */
      bool RunRgb2Ihs( TePDITypes::TePDIRasterPtrType input_image,
        TePDITypes::TePDIRasterPtrType output_image );

      /**
       * @brief Runs the IHS -> RGB implementation.
       *
       * @note Error log messages will be generated. No exceptions generated.
       *
       * @param input_image Input image.
       * @param output_image Input image.
       * @return true if OK. false if not.
       */
      bool RunIhs2Rgb( TePDITypes::TePDIRasterPtrType input_image,
        TePDITypes::TePDIRasterPtrType output_image );

  };
  
/** @example TePDIColorTransform_test.cpp
 *    Shows how to use this class.
 */    

#endif //TEPDISTATISTIC_HPP
