 
#include "TePDIGarguetFusionStrategy.hpp"
#include "TePDIMathFunctions.hpp"
#include "TePDIStrategyFactory.hpp"
#include "TePDIAlgorithmFactory.hpp"
#include "TePDIUtils.hpp"
#include <TeAgnostic.h>

#include <TeBox.h>
#include <TeUtils.h>
#include <TeRasterParams.h>

TePDIGarguetFusionStrategy::TePDIGarguetFusionStrategy()
{
};      

TePDIGarguetFusionStrategy::~TePDIGarguetFusionStrategy()
{
};


bool TePDIGarguetFusionStrategy::CheckParameters( 
  const TePDIParameters& parameters ) const
{
  /* Checking input_raster1 */
  
  TePDITypes::TePDIRasterPtrType reference_raster;
  TEAGN_TRUE_OR_RETURN( parameters.GetParameter( "reference_raster", 
    reference_raster ),
    "Missing parameter: reference_raster" );
  TEAGN_TRUE_OR_RETURN( reference_raster.isActive(),
    "Invalid parameter: reference_raster inactive" );
  TEAGN_TRUE_OR_RETURN( reference_raster->params().status_ != 
    TeRasterParams::TeNotReady, 
    "Invalid parameter: reference_raster not ready" );    
    
  /* Checking input_raster2 */
  
  TePDITypes::TePDIRasterPtrType lowres_raster;
  TEAGN_TRUE_OR_RETURN( parameters.GetParameter( "lowres_raster", 
    lowres_raster ),
    "Missing parameter: lowres_raster" );
  TEAGN_TRUE_OR_RETURN( lowres_raster.isActive(),
    "Invalid parameter: lowres_raster inactive" );
  TEAGN_TRUE_OR_RETURN( lowres_raster->params().status_ != 
    TeRasterParams::TeNotReady, "Invalid parameter: lowres_raster not ready" );    
    
  /* Checking output_raster */
  
  TePDITypes::TePDIRasterPtrType output_raster;
  TEAGN_TRUE_OR_RETURN( parameters.GetParameter( "output_raster", 
    output_raster ),
    "Missing parameter: output_raster" );
  TEAGN_TRUE_OR_RETURN( output_raster.isActive(),
    "Invalid parameter: output_raster inactive" );
  TEAGN_TRUE_OR_RETURN( output_raster->params().status_ != 
    TeRasterParams::TeNotReady, "Invalid parameter: output_raster not ready" );    
    
  /* Checking bands */    
    
  int reference_raster_band = 0;    
  TEAGN_TRUE_OR_RETURN( parameters.GetParameter( "reference_raster_band", 
    reference_raster_band ), "Missing parameter : reference_raster_band" );
  TEAGN_TRUE_OR_RETURN( ( reference_raster->nBands() > reference_raster_band ),
    "Invalid parameter : reference_raster_band" );
    
  int lowres_raster_band = 0;    
  TEAGN_TRUE_OR_RETURN( parameters.GetParameter( "lowres_raster_band", 
    lowres_raster_band ), "Missing parameter : lowres_raster_band" );
  TEAGN_TRUE_OR_RETURN( ( lowres_raster->nBands() > lowres_raster_band ),
    "Invalid parameter : lowres_raster_band" ); 
    
  /* Checking input_raster1 and input_raster2 decomposition compatibility */
  
  unsigned long int r1_level = 0;
  unsigned long int r2_level = 0;
  unsigned long int min_error = 0;
  TEAGN_TRUE_OR_RETURN( TePDIUtils::FindBestDecimLevels( reference_raster,
    lowres_raster, r1_level, r2_level, min_error),
    "Unable to find a suitable decomposition level for both input rasters" );
  TEAGN_TRUE_OR_RETURN( ( min_error == 0 ),
    "Unable to find a suitable decomposition level for both input rasters" );
    
  /* Checking photometric interpretation */
  
  TEAGN_TRUE_OR_RETURN( ( 
    ( reference_raster->params().photometric_[ reference_raster_band ] == 
      TeRasterParams::TeRGB ) ||
    ( reference_raster->params().photometric_[ reference_raster_band ] == 
      TeRasterParams::TeMultiBand ) ),
  "Invalid parameter - reference_raster (invalid photometric interpretation)" );   

  TEAGN_TRUE_OR_RETURN( ( 
    ( lowres_raster->params().photometric_[ lowres_raster_band ] == 
      TeRasterParams::TeRGB ) ||
    ( lowres_raster->params().photometric_[ lowres_raster_band ] == 
      TeRasterParams::TeMultiBand ) ),
  "Invalid parameter - lowres_raster (invalid photometric interpretation)" );   
      
  return true;
}


bool TePDIGarguetFusionStrategy::Implementation( const TePDIParameters& params )
{
  TePDITypes::TePDIRasterPtrType reference_raster;
  params.GetParameter( "reference_raster", reference_raster );

  TePDITypes::TePDIRasterPtrType lowres_raster;
  params.GetParameter( "lowres_raster", lowres_raster );

  TePDITypes::TePDIRasterPtrType output_raster;
  params.GetParameter( "output_raster", output_raster );
  
  int reference_raster_band = 0;
  params.GetParameter( "reference_raster_band", reference_raster_band );

  int lowres_raster_band = 0;
  params.GetParameter( "lowres_raster_band", lowres_raster_band );
  
  TePDITypes::TePDIRasterPtrType lowlow;
  TEAGN_TRUE_OR_THROW( TePDIUtils::TeAllocRAMRaster( lowlow, 1, 1, 1, false,
    TeDOUBLE, 0 ), "Unable to create temporary low-low raster" );  
  
  /* Finding the best decomposition levels for both rasters */
  
  unsigned long int reference_raster_best_level = 0;
  unsigned long int lowres_raster_best_level = 0;
  unsigned long int error = 0;
  TePDIUtils::FindBestDecimLevels( reference_raster, lowres_raster, 
    reference_raster_best_level, lowres_raster_best_level, error );
  TEAGN_TRUE_OR_RETURN( ( ( reference_raster_best_level > 0 ) && 
    ( lowres_raster_best_level > 0 ) ),
    "Unable to find the best wavelet decomposition levels" );
    
 /* Building default wavelet filters  - Reference coef 07 */
 
  double filters_scale = 0.71;
  if( params.CheckParameter< double >( "filters_scale" ) ) {
      
    params.GetParameter( "filters_scale", filters_scale );
  }
    
  std::vector< double > a_filter_l;
  a_filter_l.push_back( -0.05 );
  a_filter_l.push_back( 0.25 );
  a_filter_l.push_back( 0.6 );
  a_filter_l.push_back( 0.25 );
  a_filter_l.push_back( -0.05 );
  
  std::vector< double > a_filter_h;
  a_filter_h.push_back( 0.0107143 );
  a_filter_h.push_back( -0.0535714 );
  a_filter_h.push_back( -0.2607143 );
  a_filter_h.push_back( 0.6071429 );
  a_filter_h.push_back( -0.2607143 );
  a_filter_h.push_back( -0.0535714 );
  a_filter_h.push_back( 0.0107143 );
  
  std::vector< double > s_filter_l;
  s_filter_l.push_back( -0.0107143 );
  s_filter_l.push_back( -0.0535714 );
  s_filter_l.push_back( 0.2607143 );
  s_filter_l.push_back( 0.6071429 );
  s_filter_l.push_back( 0.2607143 );
  s_filter_l.push_back( -0.0535714 );
  s_filter_l.push_back( -0.0107143 );
  
  std::vector< double > s_filter_h;
  s_filter_h.push_back( -0.05 );
  s_filter_h.push_back( -0.25 );
  s_filter_h.push_back( 0.6 );
  s_filter_h.push_back( -0.25 );
  s_filter_h.push_back( -0.05 );
  
  /* Checking for user supplied wavelet filters */
  
  if( params.CheckParameter< std::vector< double > >( "a_filter_l" ) ) {
      
    params.GetParameter( "a_filter_l", a_filter_l );
  }
  if( params.CheckParameter< std::vector< double > >( "a_filter_h" ) ) {
      
    params.GetParameter( "a_filter_h", a_filter_h );
  }  
  if( params.CheckParameter< std::vector< double > >( "s_filter_l" ) ) {
      
    params.GetParameter( "s_filter_l", s_filter_l );
  }
  if( params.CheckParameter< std::vector< double > >( "s_filter_h" ) ) {
      
    params.GetParameter( "s_filter_h", s_filter_h );
  }  
  
  /* starting filters */
  
  TePDIParameters params1;
  params1.SetParameter( "wavelets_type", std::string( "mallat" ) );
  params1.SetParameter( "filter_task", std::string( "SBExtract" ) );
  params1.SetParameter( "input_image", lowres_raster );
  params1.SetParameter( "band", lowres_raster_band );
  params1.SetParameter( "a_filter_l", a_filter_l );
  params1.SetParameter( "a_filter_h", a_filter_h );
  params1.SetParameter( "s_filter_l", s_filter_l );
  params1.SetParameter( "s_filter_h", s_filter_h );
  params1.SetParameter( "filters_scale", filters_scale );         
  params1.SetParameter( "levels", (int)lowres_raster_best_level );  
  params1.SetParameter( "sub_band", lowlow );  
  params1.SetParameter( "sub_band_index", (int)0 );  
  
  TePDIAlgorithm::pointer filter1( 
    TePDIAlgorithmFactory::make( "TePDIWavelets", params1) );
  TEAGN_TRUE_OR_RETURN( filter1.isActive(), 
    "Unable to create wavelet filter1" );
  filter1->ToggleProgInt( progress_interface_enabled_ );
  TEAGN_TRUE_OR_THROW( filter1->Reset( params1 ),
    "Unable to set wavelet filter 1 Parameters" );
  TEAGN_TRUE_OR_THROW( filter1->Apply(), "Wavelet Filter 1 apply error" );
  
  TePDIParameters params2;
  params2.SetParameter( "wavelets_type", std::string( "mallat" ) );
  params2.SetParameter( "filter_task", std::string( "SBSwap" ) );
  params2.SetParameter( "input_image", reference_raster );
  params2.SetParameter( "band", reference_raster_band );
  params2.SetParameter( "output_image", output_raster );
  params2.SetParameter( "a_filter_l", a_filter_l );
  params2.SetParameter( "a_filter_h", a_filter_h );
  params2.SetParameter( "s_filter_l", s_filter_l );
  params2.SetParameter( "s_filter_h", s_filter_h );    
  params2.SetParameter( "filters_scale", filters_scale );         
  params2.SetParameter( "levels", (int)reference_raster_best_level );  
  params2.SetParameter( "sub_band", lowlow ); 
  params2.SetParameter( "sub_band_index", (int)0 );  
    
  TePDIAlgorithm::pointer filter2( 
    TePDIAlgorithmFactory::make( "TePDIWavelets", params2) );
  TEAGN_TRUE_OR_RETURN( filter2.isActive(), 
    "Unable to create wavelet filter2" );
  filter2->ToggleProgInt( progress_interface_enabled_ );
  TEAGN_TRUE_OR_THROW( filter2->Reset( params2 ),
    "Unable to set wavelet filter 2 Parameters" );  
  TEAGN_TRUE_OR_THROW( filter2->Apply(), "Wavelet Filter 2 apply error" );  
    
  return true;
}


