 
#include "TePDIIHSFusionStrategy.hpp"
#include "TePDIUtils.hpp"
#include <TeAgnostic.h>
#include "TePDIColorTransform.hpp"
#include "TePDIAlgorithmFactory.hpp"
#include "TePDIPIManager.hpp"

#include <TeRasterRemap.h>


TePDIIHSFusionStrategy::TePDIIHSFusionStrategy()
{
};      


TePDIIHSFusionStrategy::~TePDIIHSFusionStrategy()
{
};


bool TePDIIHSFusionStrategy::CheckParameters( 
  const TePDIParameters& parameters ) const
{
  /* Checking input_raster1 */
  
  TePDITypes::TePDIRasterPtrType reference_raster;
  TEAGN_TRUE_OR_RETURN( parameters.GetParameter( "reference_raster", 
    reference_raster ),
    "Missing parameter: reference_raster" );
  TEAGN_TRUE_OR_RETURN( reference_raster.isActive(),
    "Invalid parameter: reference_raster inactive" );
  TEAGN_TRUE_OR_RETURN( reference_raster->params().status_ != 
    TeRasterParams::TeNotReady, 
    "Invalid parameter: reference_raster not ready" );
    
  /* Checking input_raster2 */
  
  TePDITypes::TePDIRasterPtrType lowres_raster;
  TEAGN_TRUE_OR_RETURN( parameters.GetParameter( "lowres_raster", 
    lowres_raster ),
    "Missing parameter: lowres_raster" );
  TEAGN_TRUE_OR_RETURN( lowres_raster.isActive(),
    "Invalid parameter: lowres_raster inactive" );
  TEAGN_TRUE_OR_RETURN( lowres_raster->params().status_ != 
    TeRasterParams::TeNotReady, "Invalid parameter: lowres_raster not ready" );    
  TEAGN_TRUE_OR_RETURN( ( lowres_raster->nBands() == 3 ),
    "Invalid parameter: lowres_raster must have 3 bands ( RGB )" );       
    
  /* Checking output_raster */
  
  TePDITypes::TePDIRasterPtrType output_raster;
  TEAGN_TRUE_OR_RETURN( parameters.GetParameter( "output_raster", 
    output_raster ),
    "Missing parameter: output_raster" );
  TEAGN_TRUE_OR_RETURN( output_raster.isActive(),
    "Invalid parameter: output_raster inactive" );
  TEAGN_TRUE_OR_RETURN( output_raster->params().status_ != 
    TeRasterParams::TeNotReady, "Invalid parameter: output_raster not ready" );    
    
  /* Checking bands */    
    
  int reference_raster_band = 0;    
  TEAGN_TRUE_OR_RETURN( parameters.GetParameter( "reference_raster_band", 
    reference_raster_band ), "Missing parameter : reference_raster_band" );
  TEAGN_TRUE_OR_RETURN( ( reference_raster->nBands() > reference_raster_band ),
    "Invalid parameter : reference_raster_band" );
    
  /* Checking photometric interpretation */
  
  TEAGN_TRUE_OR_RETURN( ( 
    ( reference_raster->params().photometric_[ reference_raster_band ] == 
      TeRasterParams::TeRGB ) ||
    ( reference_raster->params().photometric_[ reference_raster_band ] == 
      TeRasterParams::TeMultiBand ) ),
  "Invalid parameter - reference_raster (invalid photometric interpretation)" );   

  for( int lowres_raster_band = 0 ; 
    lowres_raster_band < lowres_raster->params().nBands() ; 
    ++lowres_raster_band ) {
    
    TEAGN_TRUE_OR_RETURN( ( 
      ( lowres_raster->params().photometric_[ lowres_raster_band ] == 
        TeRasterParams::TeRGB ) ||
      ( lowres_raster->params().photometric_[ lowres_raster_band ] == 
        TeRasterParams::TeMultiBand ) ),
    "Invalid parameter - lowres_raster (invalid photometric interpretation)" ); 
  }  
    
  return true;
}


bool TePDIIHSFusionStrategy::Implementation( const TePDIParameters& params )
{
  TePDITypes::TePDIRasterPtrType reference_raster;
  params.GetParameter( "reference_raster", reference_raster );
  
  TePDITypes::TePDIRasterPtrType lowres_raster;
  params.GetParameter( "lowres_raster", lowres_raster );
  
  TePDITypes::TePDIRasterPtrType output_raster;
  params.GetParameter( "output_raster", output_raster );
    
  int reference_raster_band = 0;    
  params.GetParameter( "reference_raster_band", reference_raster_band );
  
  /* Bringing lowres_raster to reference_raster size */

  TePDITypes::TePDIRasterPtrType upsampled_raster;
    
  double dummy_value = 0;
  if( lowres_raster->params().useDummy_ ) {
    dummy_value = lowres_raster->params().dummy_[ 0 ];
  }
  
  TeRasterParams upsampled_raster_params;
  upsampled_raster_params.nBands( 3 );
  if( reference_raster->projection() != 0 ) {
    TeSharedPtr< TeProjection > proj( TeProjectionFactory::make( 
      reference_raster->projection()->params() ) );  
    upsampled_raster_params.projection( proj.nakedPointer() );
  }
  upsampled_raster_params.boxResolution( reference_raster->params().box().x1(), 
    reference_raster->params().box().y1(), 
    reference_raster->params().box().x2(), 
    reference_raster->params().box().y2(), 
    reference_raster->params().resx_, 
    reference_raster->params().resy_ );
  upsampled_raster_params.setDataType( lowres_raster->params().dataType_[ 0 ], 
    -1 );
  if( lowres_raster->params().useDummy_ ) {
    upsampled_raster_params.setDummy( dummy_value, -1 );
  } else {
    upsampled_raster_params.useDummy_ = false;
  }
  upsampled_raster_params.setPhotometric( TeRasterParams::TeRGB, -1 );
  
  TEAGN_TRUE_OR_RETURN( TePDIUtils::TeAllocRAMRaster( upsampled_raster,
    upsampled_raster_params, TePDIUtils::TePDIUtilsAutoMemPol ),
    "Unable create the new upsampled raster RGB raster" );
   
  TeRasterRemap remap_instance( lowres_raster.nakedPointer(), 
    upsampled_raster.nakedPointer() );
  TEAGN_TRUE_OR_RETURN( remap_instance.apply( progress_interface_enabled_ ),
    "Unable to remap lowres_raster" );
    
  /* Generating the IHS raster */
  
  TePDITypes::TePDIRasterPtrType ihs_raster;
  TEAGN_TRUE_OR_RETURN( TePDIUtils::TeAllocRAMRaster( ihs_raster,
    1, 10, 10, false, TeDOUBLE, 0 ),
    "Unable create the new ihs_raster raster RGB raster" );  
  
  TePDIParameters rgb2ihs_params;
  rgb2ihs_params.SetParameter( "input_image", upsampled_raster );
  rgb2ihs_params.SetParameter( "output_image", ihs_raster );
  rgb2ihs_params.SetParameter( "transf_type", TePDIColorTransform::Rgb2Ihs );  
  
  TePDIAlgorithm::pointer IHSTransform( 
    TePDIAlgorithmFactory::make( "TePDIColorTransform", rgb2ihs_params) );
    
  IHSTransform->ToggleProgInt( progress_interface_enabled_ );
    
  TEAGN_TRUE_OR_RETURN( IHSTransform->Apply(), 
    "Unable to build IHS color space" );
    
  rgb2ihs_params.Clear();
    
  /* Swapping reference_raster into ihs_raster I component */
  
  const unsigned int lines_bound = ihs_raster->params().nlines_;
  const unsigned int cols_bound = ihs_raster->params().ncols_;
  unsigned int line = 0;
  unsigned int col = 0;
  TeRaster* inraster = reference_raster.nakedPointer();
  TeRaster* outraster = ihs_raster.nakedPointer();
  double value = 0;
  bool get_result = false;
  bool set_result = false;
  
  TePDIPIManager progress( "Swapping Intensity channel...",
    lines_bound, progress_interface_enabled_ );
  
  for( line = 0; line < lines_bound ; ++line ) {
    TEAGN_FALSE_OR_RETURN( progress.Increment(), "Canceled by the user" );
  
    for( col = 0; col < cols_bound ; ++col ) {
      get_result = inraster->getElement( col, line, value, 
        reference_raster_band );
      TEAGN_DEBUG_CONDITION( get_result, "Unable to read element" )
      
      set_result = outraster->setElement( col, line, value, 0 );
      TEAGN_DEBUG_CONDITION( set_result, "Unable to read element" )      
    }
  }
  
  progress.Toggle( false );
  
  upsampled_raster.reset();
  
  /* Switching back to RBG space */
  
  TePDIParameters ihs2rgb_params;
  ihs2rgb_params.SetParameter( "input_image", ihs_raster );
  ihs2rgb_params.SetParameter( "output_image", output_raster );
  ihs2rgb_params.SetParameter( "transf_type", TePDIColorTransform::Ihs2Rgb );  
  
  TEAGN_TRUE_OR_RETURN( IHSTransform->Reset( ihs2rgb_params ), 
    "Unable to build RGB color space" );    
  
  TEAGN_TRUE_OR_RETURN( IHSTransform->Apply(), 
    "Unable to build RGB color space" );  
     
  return true;
}

