/*
TerraLib - a library for developing GIS applications.
Copyright  2001, 2002, 2003 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular
purpose. The library provided hereunder is on an "as is" basis, and the
authors have no obligation to provide maintenance, support, updates,
enhancements, or modifications.
In no event shall INPE be held liable to any party
for direct, indirect, special, incidental, or consequential damages arising
out of the use of this library and its documentation.
*/

#ifndef TEPDIMATRIX_HPP
  #define TEPDIMATRIX_HPP
  
  #include <TeAgnostic.h>
  #include <TeSharedPtr.h>
  #include <TeMappedMemory.h>
  
  #include "TeUtils.h"
  
  #include <vector>

  /**
   * @brief This is the template class to deal with a generic matrix.
   * @author Emiliano F. Castejon <castejon@dpi.inpe.br>
   * @ingroup PDIModule
   */
  template< class T >
  class TePDIMatrix {
    public :
      /** @typedef TeSharedPtr< TePDIMatrix< T > > pointer 
          Type definition for a instance pointer */
      typedef TeSharedPtr< TePDIMatrix< T > > pointer;
      /** @typedef const TeSharedPtr< TePDIMatrix< T > > const_pointer 
          Type definition for a const instance pointer */
      typedef const TeSharedPtr< TePDIMatrix< T > > const_pointer;
      
      /**
       * @brief Memory polycy.
       */
      enum MemoryPolicy {
        /**
         * Automatic memory policy ( Try to use RAM or DISK, if there is no 
         * avaliable RAM ) -
         * DO NOT USE AutoMemPol FOR COMPLEX DATA TYPES !
         * 
         */
        AutoMemPol = 1,
        /**
         * RAM memory policy.
         */
        RAMMemPol = 2,
        /**
         * Disk memory policy ( virtual mapped memory ) -
         *  DO NOT USE DiskMemPol FOR COMPLEX DATA TYPES !
         */
        DiskMemPol = 3
      };
      
      /**
       * @brief Default Constructor.
       *
       * @param lines Number of lines.
       * @param columns Number of columns.
       * @param mp Memory policy.
       */
      TePDIMatrix( unsigned int lines = 0, unsigned int columns = 0, 
        MemoryPolicy mp = RAMMemPol );

      /**
       * @brief Alternative Constructor.
       *
       * @param external External object reference.
       * @param mp Memory policy.
       */
      TePDIMatrix( const TePDIMatrix< T >& external, 
        MemoryPolicy mp = RAMMemPol );

      /**
       * @brief Default Destructor
       */
      ~TePDIMatrix();
      
      /**
       * @brief Reset the active instance the the new parameters.
       *
       * @param lines Number of lines.
       * @param columns Number of columns.
       * @return true if OK, false on error.
       */
      bool Reset( unsigned int lines = 0, unsigned int columns = 0 );      

      /**
       * @brief Reset the active instance the the new parameters.
       *
       * @param lines Number of lines.
       * @param columns Number of columns.
       * @param mp Memory policy.
       * @return true if OK, false on error.
       */
      bool Reset( unsigned int lines, unsigned int columns,
        MemoryPolicy mp );
      
      /**
       * @brief The number of current matrix lines.
       *
       * @return The number of current matrix lines.
       */
      inline unsigned int GetLines() const;
      
      /**
       * @brief The number of current matrix columns.
       *
       * @return The number of current matrix columns
       */
      inline unsigned int GetColumns() const;
      
      /**
       * @brief Empty Matrix verification.
       *
       * @return true if the matrix is empty.
       */
      inline bool IsEmpty() const;

      /**
       * @brief Operator = overload.
       *
       * @note The external memory policy will be used as reference.
       *
       * @param external External instance reference.
       * @return A reference to the current matrix.
       */
      const TePDIMatrix< T >& operator=( const TePDIMatrix< T >& external );

      /**
       * @brief Operator () overload.
       *
       * @param line Line number.
       * @param column Column number.
       * @return A reference to the required element.
       */
      inline T& operator()( unsigned int line, unsigned int column );
      
      /**
       * @brief Operator () overload.
       *
       * @param line Line number.
       * @param column Column number.
       * @return A const reference to the required element.
       */
      inline const T& operator()( unsigned int line, 
        unsigned int column ) const; 
      
    protected :

      /**
       * @brief The internal buffer.
       */
      mutable T** buffer_;
      
      /**
       * @brief The internal buffer lines number.
       */
      mutable unsigned int buffer_lines_;
      
      /**
       * @brief The internal buffer columns number.
       */
      mutable unsigned int buffer_columns_;     
      
      /**
       * @brief The user supplied memory policy.
       */
      mutable MemoryPolicy user_mem_policy_;   
            
      /**
       * @brief The current used memory policy.
       */
      mutable MemoryPolicy current_mem_policy_;
      
      /**
       * @brief The internal mapped memory handler.
       */
      mutable std::vector< TeSharedPtr< TeMappedMemory > > 
        mapmemhandlervec_;
      
      /**
       * @brief Reset the internal variables to the initial state.
       */      
      void init();
  };

  template< class T >
  void TePDIMatrix< T >::init()
  {
    buffer_ = 0;
    buffer_lines_ = 0;
    buffer_columns_ = 0;  
    user_mem_policy_ = current_mem_policy_ = RAMMemPol;
  }
  
  
  template< class T >
  TePDIMatrix< T >::TePDIMatrix( unsigned int lines, unsigned int columns,
    MemoryPolicy mp )
  {
    init();
    
    TEAGN_TRUE_OR_THROW( Reset( lines, columns, mp ),
      "Unable to initiate the matrix object" );
  }  
  
  
  template< class T >
  TePDIMatrix< T >::TePDIMatrix( const TePDIMatrix< T >& external,
    MemoryPolicy mp )
  {
    init();
    
    TEAGN_TRUE_OR_THROW( 
      Reset( external.buffer_lines_, external.buffer_columns_, mp ),
      "Unable to initiate the matrix object" );
    
    unsigned int line;
    unsigned int column;
    
    for( line = 0 ; line < buffer_lines_ ; ++line ) {
      for( column = 0 ; column < buffer_columns_ ; ++column ) {
        buffer_[ line ][ column ] = external.buffer_[ line ][ column ];
      }
    }
  }


  template< class T >
    TePDIMatrix< T >::~TePDIMatrix()
  {
    Reset( 0, 0 );
  }

  
  template< class T >
  bool TePDIMatrix< T >::Reset( unsigned int lines, unsigned int columns )
  {
    return Reset( lines, columns, user_mem_policy_ );
  }  

  template< class T >
  bool TePDIMatrix< T >::Reset( unsigned int lines, unsigned int columns,
    MemoryPolicy mp )
  {
    /* Update the old buffer if necessary */
    
    if( ( lines != buffer_lines_ ) || ( columns != buffer_columns_ ) ||
        ( mp != user_mem_policy_ ) ) {
    
      /* free the old resources */
    
      if( buffer_ != 0 ) {
        switch( current_mem_policy_ ) {
          case RAMMemPol :
          {
            for( unsigned int line = 0 ; line < buffer_lines_ ; ++line ) {
              delete[] (buffer_[ line ]);
            }
        
            delete[] buffer_;          
            
            break;
          }
          case DiskMemPol :
          {
            mapmemhandlervec_.clear();
            delete[] buffer_;
          
            break;
          }
          default :
          {
            TEAGN_LOG_AND_THROW( "Invalid memory policy" );
            break;          
          }
        }
        
        /* Updating all global variables */
                
        buffer_ = 0;
        buffer_lines_ = 0;
        buffer_columns_ = 0;
      }
      
      /* Allocate the new resources */
      
      user_mem_policy_ = mp;
      
      if( ( lines != 0 ) && ( columns != 0 ) ) {
        /* Guessing the best memory source, if in automatic mode */
      
        if( mp == AutoMemPol ) {
          if( TeGetFreePhysicalMemory() > 
            ( lines * columns * sizeof( T ) ) ) {
            
            current_mem_policy_ = RAMMemPol;
          } else {
            current_mem_policy_ = DiskMemPol;
          }
        } else {
          current_mem_policy_ = mp;
        }
      
        switch( current_mem_policy_ ) {
          case RAMMemPol :
          {
            buffer_ = new T*[ lines ];
        
            TEAGN_TRUE_OR_RETURN( buffer_, "Memory allocation error" );
            
            bool mem_erros_ocurred = false;
            unsigned int line = 0;
        
            for( line = 0 ; line < lines ; ++line ) {
              if( mem_erros_ocurred ) {
                buffer_[ line ] = 0;
              } else {
                buffer_[ line ] = new T[ columns ];
                
                if( buffer_[ line ] == 0 ) {
                  mem_erros_ocurred = true;
                }
              }
            }
            
            if( mem_erros_ocurred ) {
              for( line = 0 ; line < lines ; ++line ) {
                if( buffer_[ line ] != 0 ) {
                  delete[] (buffer_[ line ]);
                }
              }
              
              delete[] buffer_;
              buffer_ = 0;
            
              TEAGN_LOG_AND_RETURN( "Memory allocation error" );
            }
            
            break;
          }
          case DiskMemPol :
          {
            const unsigned long int max_bytes_per_mapped_file =
              ( 1024 * 1024 * 100 );
            const unsigned long int total_bytes = lines * 
              columns * sizeof( T );
            const unsigned long int bytes_per_line = columns * sizeof( T );
            
            TEAGN_TRUE_OR_RETURN( 
              ( bytes_per_line <= max_bytes_per_mapped_file ),
              "The number of columns is too big" );
            
            unsigned long int full_mapped_files_nmb = 0;
            unsigned long int full_mapped_files_lines_nmb = 0;
            unsigned long int full_mapped_files_bytes = 0;
            unsigned long int parcial_mapped_file_bytes = 0;
            unsigned long int parcial_mapped_file_lines_nmb = 0;
            if( total_bytes > max_bytes_per_mapped_file ) {
              full_mapped_files_lines_nmb = max_bytes_per_mapped_file
                / bytes_per_line;
              full_mapped_files_nmb = lines / full_mapped_files_lines_nmb;
              full_mapped_files_bytes = full_mapped_files_lines_nmb *
                bytes_per_line;
                
              parcial_mapped_file_lines_nmb = lines - 
                ( full_mapped_files_nmb * full_mapped_files_lines_nmb );
              parcial_mapped_file_bytes = parcial_mapped_file_lines_nmb * 
                bytes_per_line;
            } else {
              parcial_mapped_file_bytes = total_bytes;
              parcial_mapped_file_lines_nmb = lines;
            }            
            
            buffer_ = new T*[ lines ];
            TEAGN_TRUE_OR_RETURN( buffer_, "Memory allocation error" );            
              
            T** buffer_walker_ptr = buffer_;
              
            /* Alocating all full mapped files */
            
            for( unsigned long int fmf_nmb = 0 ; 
              fmf_nmb < full_mapped_files_nmb ; ++fmf_nmb ) {
              
              TeSharedPtr< TeMappedMemory > mm_ptr( new TeMappedMemory );
                
              if( ( ! mm_ptr.isActive() ) ||
                  ( ! mm_ptr->reset( full_mapped_files_bytes, true ) ) ) {
                  
                mapmemhandlervec_.clear();
                delete[] buffer_;
                TEAGN_LOG_AND_RETURN( 
                  "Unable to allocate disk memory resources" );
              }
              
              mapmemhandlervec_.push_back( mm_ptr );
              
              T* start_pointer = ( T* ) mm_ptr->getPointer();
              
              for( unsigned long int curr_line = 0; 
                curr_line < full_mapped_files_lines_nmb ;
                ++curr_line ) {
                
                *buffer_walker_ptr = start_pointer;
                
                ++buffer_walker_ptr;
                start_pointer = start_pointer + columns;
              }
            }     
            
            /* Alocating all parcial mapped file */
            
            if( parcial_mapped_file_bytes > 0 ) {
              TeSharedPtr< TeMappedMemory > mm_ptr( new TeMappedMemory );
                
              if( ( ! mm_ptr.isActive() ) ||
                ( ! mm_ptr->reset( parcial_mapped_file_bytes, true ) ) ) {
                
                mapmemhandlervec_.clear();
                delete[] buffer_;
                TEAGN_LOGERR( "Unable to allocate disk memory resources" );
              }
              
              mapmemhandlervec_.push_back( mm_ptr );
              
              T* start_pointer = ( T* ) mm_ptr->getPointer();
              
              for( unsigned long int curr_line = 0 ; 
                curr_line < parcial_mapped_file_lines_nmb ;
                ++curr_line ) {
                
                *buffer_walker_ptr = start_pointer;
                
                ++buffer_walker_ptr;
                start_pointer = start_pointer + columns;
              }
            }                     
             
            break;
          }       
          default :
          {
            TEAGN_LOG_AND_THROW( "Invalid memory policy" );
            break;
          }
        }
        
        buffer_lines_ = lines;
        buffer_columns_ = columns;          
      }
    }
    
    return true;
  }
  
  
  template< class T >
  inline unsigned int TePDIMatrix< T >::GetLines() const
  {
    return buffer_lines_;
  }

  
  template< class T >
  inline unsigned int TePDIMatrix< T >::GetColumns() const
  {
    return buffer_columns_;
  }
  
  
  template< class T >
  inline bool TePDIMatrix< T >::IsEmpty() const
  {
    return ( buffer_ == 0 ) ? true : false;
  }
  

  template< class T >
  const TePDIMatrix< T >& TePDIMatrix< T >::operator=(
    const TePDIMatrix< T >& external )
  {
    TEAGN_TRUE_OR_THROW( 
      Reset( external.buffer_lines_, external.buffer_columns_,
      external.current_mem_policy_ ),
      "Unable to initiate the matrix object" );
    
    unsigned int line;
    unsigned int column;
    
    for( line = 0 ; line < buffer_lines_ ; ++line ) {
      for( column = 0 ; column < buffer_columns_ ; ++column ) {
        buffer_[ line ][ column ] = external.buffer_[ line ][ column ];
      }
    }

    return *this;
  }

  
  template< class T >
  inline T& TePDIMatrix< T >::operator()( unsigned int line, 
    unsigned int column )
  {
    TEAGN_DEBUG_CONDITION( buffer_ != 0, 
      "Trying to access an empty matrix" );
    TEAGN_DEBUG_CONDITION( line < buffer_lines_, 
      "Trying to access an invalid line [" + Te2String( line ) + "]" );
    TEAGN_DEBUG_CONDITION( column < buffer_columns_, 
      "Trying to access an invalid column [" + Te2String( column ) + "]" );
  
    return buffer_[ line ][ column ];
  }    
  

  template< class T >
  inline const T& TePDIMatrix< T >::operator()( unsigned int line, 
    unsigned int column ) const
  {
    TEAGN_DEBUG_CONDITION( buffer_ != 0, 
      "Trying to access an empty matrix" );
    TEAGN_DEBUG_CONDITION( line < buffer_lines_, 
      "Trying to access an invalid line [" + Te2String( line ) + "]" );
    TEAGN_DEBUG_CONDITION( column < buffer_columns_, 
      "Trying to access an invalid column [" + Te2String( column ) + "]" );
  
    return buffer_[ line ][ column ];
  }       
  
/** @example TePDIMatrix_test.cpp
 *    Shows how to use this class.
 */    
  
#endif //TEPDIMATRIX_HPP

