/*
TerraLib - a library for developing GIS applications.
Copyright  2001, 2002, 2003 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular
purpose. The library provided hereunder is on an "as is" basis, and the
authors have no obligation to provide maintenance, support, updates,
enhancements, or modifications.
In no event shall INPE be held liable to any party
for direct, indirect, special, incidental, or consequential damages arising
out of the use of this library and its documentation.
*/

#ifndef TEPDIPIMANAGER_HPP
  #define TEPDIPIMANAGER_HPP
  
  #include <TeSharedPtr.h>
  
  #include <string>

  /**
   * @brief This is the class for the active progress interface management.
   * @author Emiliano F. Castejon <castejon@dpi.inpe.br>
   * @ingroup PDIModule
   */
  class TePDIPIManager {
  
    public :
    
      /**
       * @brief Default Constructor
       *
       * @param message Progress window message (label).
       * @param tsteps Progress window total steps.
       * @param enabled Flag to enable ( true ) or disable ( false ).
       */
      TePDIPIManager( const std::string& message = " ", 
        unsigned long int tsteps = 0, bool enabled = false );    
    
      /**
       * @brief Default Destructor
       */
      ~TePDIPIManager();
      
      /**
       * @brief Enable / Disable the current progress interface manager.
       *
       * @param enabled Flag to enable ( true ) or disable ( false ).
       */
      void Toggle( bool enabled );  
      
      /**
       * @brief Update the current progress interface to the supplied step.
       *
       * @param step Current step.
       * @return true if the interface was canceled, false otherwise.
       */
      bool Update( unsigned long int step );      
      
      /**
       * @brief Incriments the the current progress by one step.
       * @return true if the interface was canceled, false otherwise.
       */
      bool Increment();

      /**
       * @brief Reset the current instance state.
       * 
       * @param message Message (label).
       * @param tsteps Total steps.
       */
      void Reset( const std::string& message = " ", 
        unsigned long int tsteps = 0 );

    protected :
    
      /**
       * @brief This is the class for the active progress interface management
       * parameters.
       * @author Emiliano F. Castejon <castejon@dpi.inpe.br>
       * @ingroup PDIModule
       */    
      class TePDIPIManagerSettings {
        public :
          /**
           * @brief A flag indicating if the PI manager is enabled.
           */
          bool enabled_;
          
          /**
           * @brief The current manager pointer.
           */
          TePDIPIManager* curr_manager_ptr_;          
          
          /**
           * @brief The total steps number.
           */
          unsigned long int tsteps_;
          
          /**
           * @brief The current step.
           */
          unsigned long int currstep_;
          
          /**
           * @brief The proportional current step ( 0% to 100% ).
           */
          int propcurrstep_;          
          
          /**
           * @brief The progress interface message (label).
           */
          std::string message_;
          
          /**
           * @brief Default constructor.
           */
          TePDIPIManagerSettings()
          : enabled_( false ), curr_manager_ptr_( 0 ), tsteps_( 0 ), 
            currstep_( 0 ), propcurrstep_( 0 ), message_( "" )
          {};
          
          /**
           * @brief Default desstructor.
           */
          ~TePDIPIManagerSettings() {};
      };
    
      /**
       * @brief Internal progress interface status (current instance).
       */
      mutable TePDIPIManagerSettings my_settings_;
      
      /**
       * @brief Internal progress global interface status (last active instance).
       */
      mutable TePDIPIManagerSettings old_global_settings_;
      
      /**
       * @brief Returns a reference to the global settings used by the last
       * active progress interface manager.
       */      
      static TePDIPIManagerSettings& GetGlobalSettings()
      {
        static TePDIPIManagerSettings global_settings;
        return global_settings;
      };
      
      /**
       * @brief Updates the progress interface with the new settings.
       *
       * @param settings The new progress interface manager to be applied.
       * @return true if the interface was canceled, false otherwise.
       */         
      bool UpdateProgressInterface( 
        const TePDIPIManagerSettings& settings ) const;
        
    private :
    
      /**
       * @brief Alternative Constructor
       */    
      TePDIPIManager( const TePDIPIManager& ) {};
      
      /**
       * @brief operator= overload.
       */    
      const TePDIPIManager& operator=( const TePDIPIManager& ) 
      { 
        return *this;
      };
  };

#endif
