/************************************************************************************
TerraLib - a library for developing GIS applications.
Copyright  2001-2004 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular purpose.
The library provided hereunder is on an "as is" basis, and the authors have no
obligation to provide maintenance, support, updates, enhancements, or modifications.
In no event shall INPE and Tecgraf / PUC-Rio be held liable to any party for direct,
indirect, special, incidental, or consequential damages arising out of the use
of this library and its documentation.
*************************************************************************************/

#include <TePDIRaster2Vector.hpp>
#include <TeAgnostic.h>
#include <TePDIUtils.hpp>
//#include <TePDIBSpline.hpp>

#include <TeUtils.h>
#include <shapefil.h>

#include <algorithm>
#include <string>
#include <vector>

#include <float.h>

//these are the 8 cardinals directions used to parcour an edge
#define NW 0
#define N  1
#define EN 2
#define E  3
#define SE 4
#define S  5
#define WS 6
#define W  7

/* Used by extract arcs method */

enum Directions{
  UP=1<<0,
  DOWN=1<<1,
  RIGHT=1<<2,
  LEFT=1<<3
};

TePDIRaster2Vector::TePDIRaster2Vector()
{
    tree_=NULL;
    raster_uses_dummy_ = false;
    dummy_value_ = 0;
  containerPolygons_.clear();
  rtreePolygons_=NULL;
  raster_channel_ = 0;
}

TePDIRaster2Vector::~TePDIRaster2Vector()
{
  if( tree_ != 0 ) {
    delete tree_;
  }
  
  if( rtreePolygons_ != 0 ) {
    delete rtreePolygons_;
  }
  
  containerPolygons_.clear();
}

bool
TePDIRaster2Vector :: CheckParameters( 
  const TePDIParameters& parameters ) const {
  /* Checking for general required parameters */

  //parameter input_image
  TePDITypes::TePDIRasterPtrType inRaster;
  if( ! parameters.GetParameter( "rotulated_image", inRaster ) ) {

    TEAGN_LOGERR( "Missing parameter: rotulated_image" );
    return false;
  }
  if( ! inRaster.isActive() ) {

    TEAGN_LOGERR( "Invalid parameter: rotulated_image inactive" );
    return false;
  }
  if( inRaster->params().status_ == TeRasterParams::TeNotReady ) {

    TEAGN_LOGERR( "Invalid parameter: rotulated_image not ready" );
    return false;
  }
  
  /* Checking output_polsets */
  
  TEAGN_TRUE_OR_RETURN( 
    parameters.CheckParameter< TePDITypes::TePDIPolSetMapPtrType >( 
    "output_polsets" ), "Missing parameter: output_polsets" );
  
  TePDITypes::TePDIPolSetMapPtrType output_polsets;
  parameters.GetParameter( "output_polsets", output_polsets );
            
  TEAGN_TRUE_OR_RETURN( output_polsets.isActive(),
    "Invalid parameter output_polsets : Inactive polygon set pointer" );

  /* Checking channel */
  
  unsigned int channel = 0;
  
  TEAGN_TRUE_OR_RETURN( 
    parameters.GetParameter( "channel", channel ), 
    "Missing parameter: channel" );
    
  TEAGN_TRUE_OR_RETURN( ( channel < (unsigned int)inRaster->nBands() ),
    "Invalid parameter: channel" );
    
  return true;
}


bool TePDIRaster2Vector::RunImplementation()
{
  params_.GetParameter( "rotulated_image", raster_ );
        
  raster_uses_dummy_ = raster_->params().useDummy_;
  dummy_value_ = 0;
  if( raster_uses_dummy_ ) {
    dummy_value_ = raster_->params().dummy_[ 0 ];
  }

  chanfer_=false;
  resolution_=false;
        
  //attributes of the raster
  
  resx_ = raster_->params().resx_ ;
  resy_ = raster_->params().resy_ ;

  const TeBox raster_bbox=raster_->params().boundingBox();
  xref_ = raster_bbox.lowerLeft().x_ + 0.5*this->resx_;
  yref_ = raster_bbox.lowerLeft().y_ - 0.5*this->resy_;
  nCols_  = raster_->params().ncols_;
  nLines_ = raster_->params().nlines_;
  
  /* Updating the directions  */
  
  directions_[ 0 ] = TeCoord2D( -1, -1 );  /* NW - 0 */
  directions_[ 1 ] = TeCoord2D(  0, -1 );  /* N  - 1 */
  directions_[ 2 ] = TeCoord2D(  1, -1 );  /* EN - 2 */
  directions_[ 3 ] = TeCoord2D(  1,  0 );  /* E  - 3 */
  directions_[ 4 ] = TeCoord2D(  1,  1 );  /* SE - 4 */
  directions_[ 5 ] = TeCoord2D(  0,  1 );  /* S  - 5 */
  directions_[ 6 ] = TeCoord2D( -1,  1 );  /* WS - 6 */
  directions_[ 7 ] = TeCoord2D( -1,  0 );  /* W  - 7 */

  //declarations
        
  double val;
  int i,j;

  /* Creating a new RTree */
  
  if( rtreePolygons_ != 0 ) {
    delete rtreePolygons_;
  }  
  rtreePolygons_ = new TeSAM::TeRTree<unsigned int, 8, 4>(
    raster_->params().boundingBox());
    
  int countObjects = 0;
  vector<unsigned int> indexVec;

  //if we want to smooth the arcs we need the tree of nodes
  
  if ( chanfer_) {
    TeFindNodes();
  }
        
  StartProgInt( "Vectorizing...", nLines_ );
                
  //we scan the rotulated image in row order to vectorize the cells
  
  for ( j = 0 ; j < nLines_ ; j++) {
    TEAGN_FALSE_OR_RETURN( UpdateProgInt( j ), "Canceled by the user" );
    
    for ( i = 0 ; i < nCols_ ; i++) {
      /* Virifying if the current point is already inside a polygon generated
         before */
         
      TeCoord2D coordMatrix(i, j);
      TeCoord2D coordWorld = raster_->index2Coord(coordMatrix);
      TeBox boxPoint(coordWorld, coordWorld);
      bool exist = false;  
      
      indexVec.clear();
      rtreePolygons_->search( boxPoint, indexVec );
      
      if( ! raster_->getElement(i,j,val,raster_channel_) ) {
        TEAGN_TRUE_OR_RETURN( raster_uses_dummy_, "Raster read error" );
        val = dummy_value_;
      }      
      
      if ( ! indexVec.empty() ) {
        unsigned int indexVec_index = 0;        

        while( indexVec_index < indexVec.size() ) {
          TEAGN_DEBUG_CONDITION( 
            ( indexVec[ indexVec_index ] < containerPolygons_.size() ),
            "Invalid cadidates vector index" );
    
          const polygonStruct& pol_str_ref = 
            containerPolygons_[ indexVec[ indexVec_index ] ];
      
          if ( val == pol_str_ref.value_ ) {
            if( TePDIUtils::TeRelation( coordWorld, 
              *( pol_str_ref.indexer_ ) ) == TeINSIDE ) {
                 
              exist = true;
              break;
            }        
          }
          
          ++indexVec_index;
        }                
      }

      if( ! exist ) {
         //we found a new class
         
        TeLine2D line; 
         
        if( TeDetectEdge( i, j, line ) ) {
          TePolygon poly;
          poly.add( line );
          
          /* Verifying if the polygon is a hole of other one */
            
          unsigned int indexVec_index = 0;
          
          while( indexVec_index < indexVec.size() ) {
            polygonStruct& pol_struct_ref = 
              containerPolygons_[ indexVec[ indexVec_index ] ];
            TeTileIndexer& indexer_ref = *( pol_struct_ref.indexer_ );
            TePolygon& pol_ref = pol_struct_ref.poly_;
          
            if( TePDIUtils::TeRelation( coordWorld, indexer_ref ) == 
              TeINSIDE ) {
            
              pol_ref.add( poly.first() );
              indexer_ref.update( pol_ref.size() - 1 );
                
              break;
            }
            
            ++indexVec_index;
          }
          
          // inserting the new polygon into tree and container
          
          if( raster_uses_dummy_ && ( val == dummy_value_ ) ) {
            poly.objectId( "none" );
          } else {
            ++countObjects;
            poly.objectId( Te2String( countObjects ) );
          }
          
          polygonStruct dummy_ps;
          containerPolygons_.push_back( dummy_ps );
          
          containerPolygons_[ containerPolygons_.size() - 1 ].reset( poly, 
            (int)val, raster_->params().resy_ );
    
          rtreePolygons_->insert( poly.box(), containerPolygons_.size() - 1 );        
        } else {
          i = nCols_;
          j = nLines_;

          TEAGN_LOGERR( "Edge detection error" )
        }
      }
    }
  }
        
  /* Generating output data */
  
  TePDITypes::TePDIPolSetMapPtrType output_polsets;
  params_.GetParameter( "output_polsets", output_polsets );
      
  output_polsets->clear();
  
  std::map< double, TePolygonSet >::iterator psm_it;
    
  std::vector<polygonStruct>::iterator cpit = containerPolygons_.begin();
  std::vector<polygonStruct>::iterator cpit_end = containerPolygons_.end();
  
  double value = 0;
    
  while( cpit != cpit_end ) {
    value = (double)cpit->value_;
  
    if( ( ! raster_uses_dummy_ ) || ( value != dummy_value_ ) ) {
      psm_it = output_polsets->find( value );
      
      if( psm_it == output_polsets->end() ) {
        TePolygonSet dummy_ps;
        dummy_ps.add( cpit->poly_ );
        
        (*output_polsets)[ value ] = dummy_ps;
      } else {
        psm_it->second.add( cpit->poly_ );
      }
    }
    
    ++cpit;
  }
                
  return true;
}

void
TePDIRaster2Vector::ResetState( const TePDIParameters& params )
{
  if ( tree_ != 0 ) {
    tree_->clear(); 
  }
        
  if( rtreePolygons_ != 0 ) {
    rtreePolygons_->clear();
  }        
  
  params.GetParameter( "channel", raster_channel_ );
}


bool
TePDIRaster2Vector::TeExtractArcs(TeLine2D &line, TeLineSet &arcs,
  vector<bool> &reversed)
{

        //declarations
        unsigned int indice;
        Node *Nd;
        bool reverse=false;
        Directions precDir, nextDir;
        TeCoord2D P, precP, nextP;
        TeLine2D arc;

        if ((tree_==NULL) || (line.size()<3) ){
                //nothing to do
                arc.copyElements(line);
                arcs.add(arc);
                reversed.push_back(reverse);
                return true;
        }

    //indice of the first node
        for ( indice=0; indice < line.size() ; indice++)
        {
                if (tree_->isPresent(line[indice]) !=NULL)
                        break;
        }
    if (indice==line.size()){
                arc.copyElements(line);
                arcs.add(arc);  //this is an island
                reversed.push_back(reverse);
                return true;
        }

    line.erase(line.end()-1);
    //int z= line.size();
    for (unsigned int i=0; i< line.size() ; i++){
                P=line[indice];
                arc.add(P);
                if ((Nd=tree_->isPresent(P))!=NULL){
                        precP=line[indice!=0?indice-1:line.size()-1];
                        if (precP.x() < P.x())
                                precDir=LEFT;
                        else if (precP.x() > P.x())
                                precDir=RIGHT;
                        else if (precP.y() < P.y())
                                precDir=DOWN;
                        else
                                precDir=UP;

                        if (reverse)
                                TeReverseLine(arc);
                        else
                                Nd->seenFrom_ = Nd->seenFrom_ | precDir;

                        if (i!=0)
                        {
                                TeLine2D cp;
                                cp.copyElements(arc);
                                arcs.add(cp); reversed.push_back(reverse);
                                arc.clear();
                                arc.add(P);
                        }

                        nextP=line[indice!=line.size()-1?indice+1:0];
                        if (nextP.x() > P.x())
                                nextDir=RIGHT;
                        else if (nextP.x() < P.x())
                                nextDir=LEFT;
                        else if (nextP.y() > P.y())
                                nextDir=UP;
                        else
                                nextDir=DOWN;

                        if (Nd->seenFrom_ & nextDir)
                                reverse=true;
                        else {
                                reverse=false;
                                Nd->seenFrom_=Nd->seenFrom_ | nextDir;
                        }
                }
                indice++;
            indice=indice%line.size();
        }

        P=line[indice];
        arc.add(P);
        if (reverse)
                TeReverseLine(arc);
        TeLine2D cp;
        cp.copyElements(arc);
        arcs.add(cp); reversed.push_back(reverse);
        arc.clear();
        return true;
}


bool TePDIRaster2Vector::TeDetectEdge(long i, long j, TeLine2D &line )
{
  TEAGN_DEBUG_RETURN( startingEdgeTest( i, j ), 
    "Starting edge detection error at x=" + Te2String( i ) + " line=" +
    Te2String( j ) )
    
  line.clear();
    
  /* Current polygon pixel values */
  
  double pol_pixels_value = 0;
  
  if( ! raster_->getElement(i, j, pol_pixels_value, raster_channel_ ) ) {
    TEAGN_TRUE_OR_RETURN( raster_uses_dummy_, "Raster read error" );
    pol_pixels_value = dummy_value_;
  }  
  
  /* Generating chaing code by following the polygon borders */
  
  short curr_dir = E;
  short new_dir = E;
  short new_test_start_dir = E;
  short curr_pixel_corner = NW;
  
  short pinit_leaving_dir = W;
  bool pinit_leaving_dir_set = false;
  
  TeCoord2D curr_chain_p = TeCoord2D( 0, 0 );

  int curr_x_index = i;
  int curr_y_index = j;
  int next_x_index = i;
  int next_y_index = j;
  
  double curr_pixel_value = 0;
  
  bool look_for_next_pixel = true;
  bool next_pixel_found = false;
  bool must_add_curr_chain_p = false;
  
  line.add( curr_chain_p );
  
  while( look_for_next_pixel ) {
    /* Finding the next direction */
        
    new_test_start_dir = ( curr_dir + 6 ) % 8;
    new_dir = new_test_start_dir;
    next_pixel_found = false;
        
    do {
      next_x_index = curr_x_index + (int)( directions_[new_dir].x() );
      next_y_index = curr_y_index + (int)( directions_[new_dir].y() );
            
      if( ( next_x_index > (-1) ) && ( next_y_index > (-1) ) && 
        ( next_x_index < nCols_ ) && ( next_y_index < nLines_ ) ) {
              
        if( ! raster_->getElement( next_x_index, next_y_index, 
          curr_pixel_value, raster_channel_ ) ) {
                        
          TEAGN_TRUE_OR_RETURN( raster_uses_dummy_, "Raster read error" );
          curr_pixel_value = dummy_value_;  
        }
                        
        if ( curr_pixel_value == pol_pixels_value ){
          next_pixel_found = true;
          break;
        }
      }
            
      new_dir = ( new_dir + 2 ) % 8;
    } while( new_dir != new_test_start_dir );
    
    /* Generating the polygon line following the current state */
    
      if( ! next_pixel_found ) {
        // there is only one point inside the polygon
        // that is the last one
        
        curr_chain_p += ( directions_[ E ] );
        line.add( curr_chain_p );
        
        curr_chain_p += ( directions_[ S ] );
        line.add( curr_chain_p );
        
        curr_chain_p += ( directions_[ W ] );
        line.add( curr_chain_p );
        
        curr_chain_p += ( directions_[ N ] );
        line.add( curr_chain_p );                        
                      
        look_for_next_pixel = false;
      } else if( ( curr_x_index == i ) && ( curr_y_index == j ) &&
        ( pinit_leaving_dir == new_dir ) ) {
      
        /* We are back to the fist pixel again
          ( going to the same direction ) */
        
        switch( curr_pixel_corner ) {
          case EN :
          {
            curr_chain_p += ( directions_[ S ] );
            line.add( curr_chain_p );
        
            curr_chain_p += ( directions_[ W ] );
            line.add( curr_chain_p );
        
            curr_chain_p += ( directions_[ N ] );
            line.add( curr_chain_p );
            
            break;
          }
          case SE :
          { 
            curr_chain_p += ( directions_[ W ] );
            line.add( curr_chain_p );
        
            curr_chain_p += ( directions_[ N ] );
            line.add( curr_chain_p );
            
            break;
          }
          case WS :
          {
            curr_chain_p += ( directions_[ N ] );
            line.add( curr_chain_p );
                      
            break;
          }
          default :
          {
            TEAGN_LOG_AND_RETURN( "Invalid pixel corner=" + 
              Te2String( curr_pixel_corner) + " curr_dir=" + 
              Te2String( curr_dir ) );
            break;
          }
        }      
       
        look_for_next_pixel = false;   
      } else {
        /* Updating current chain point and current pixel corner */
        
        if( curr_dir == new_dir ) {
          curr_chain_p += ( directions_[ curr_dir ] );
          
          must_add_curr_chain_p = true;
        } else {
          if( must_add_curr_chain_p ) {
            line.add( curr_chain_p );
            
            must_add_curr_chain_p = false;
          }
        
          switch( new_dir ) {
            case E:
            {
              switch( curr_pixel_corner ) {
                case NW :
                {
                  curr_chain_p += ( directions_[ E ] );
                  line.add( curr_chain_p );
                        
                  break;
                }
                case EN :
                {
                  curr_pixel_corner = NW;
                  break;
                }
                case SE :
                { 
                  curr_chain_p += ( directions_[ W ] );
                  line.add( curr_chain_p );
                  
                  curr_chain_p += ( directions_[ N ] );
                  line.add( curr_chain_p );
                  
                  curr_chain_p += ( directions_[ E ] );
                  line.add( curr_chain_p );                                                      
                  
                  curr_pixel_corner = NW;
                  break;
                }
                case WS :
                {
                  curr_chain_p += ( directions_[ N ] );
                  line.add( curr_chain_p );
                  
                  curr_chain_p += ( directions_[ E ] );
                  line.add( curr_chain_p );                    
                  
                  curr_pixel_corner = NW;
                  break;
                }
                default :
                {
                  TEAGN_LOG_AND_THROW( "Invalid pixel corner" )
                  break;
                }
              }
              
              break;
            }
            case S:
            {
              switch( curr_pixel_corner ) {
                case NW :
                {
                  curr_chain_p += ( directions_[ E ] );
                  line.add( curr_chain_p );
                  
                  curr_chain_p += ( directions_[ S ] );
                  line.add( curr_chain_p );  
                                  
                  curr_pixel_corner = EN;  
                  break;
                }
                case EN :
                {
                  curr_chain_p += ( directions_[ S ] );
                  line.add( curr_chain_p );
                                  
                  break;
                }
                case SE :
                {
                  curr_pixel_corner = EN;
                  break;
                }
                case WS :
                {
                  curr_chain_p += ( directions_[ N ] );
                  line.add( curr_chain_p );
                  
                  curr_chain_p += ( directions_[ E ] );
                  line.add( curr_chain_p );
                  
                  curr_chain_p += ( directions_[ S ] );
                  line.add( curr_chain_p );                 
                
                  curr_pixel_corner = EN;
                  break;
                }
                default :
                {
                  TEAGN_LOG_AND_THROW( "Invalid pixel corner" )
                  break;
                }
              }
              
              break;
            }    
            case W:
            {
              switch( curr_pixel_corner ) {
                case NW :
                {
                  curr_chain_p += ( directions_[ E ] );
                  line.add( curr_chain_p );
                  
                  curr_chain_p += ( directions_[ S ] );
                  line.add( curr_chain_p );
                  
                  curr_chain_p += ( directions_[ W ] );
                  line.add( curr_chain_p ); 
                                  
                  curr_pixel_corner = SE;  
                  break;
                }
                case EN :
                {
                  curr_chain_p += ( directions_[ S ] );
                  line.add( curr_chain_p );
                  
                  curr_chain_p += ( directions_[ W ] );
                  line.add( curr_chain_p ); 
                                  
                  curr_pixel_corner = SE;
                  break;
                }
                case SE :
                {
                  curr_chain_p += ( directions_[ W ] );
                  line.add( curr_chain_p ); 
                                  
                  break;
                }
                case WS :
                {
                  curr_pixel_corner = SE;
                  break;
                }
                default :
                {
                  TEAGN_LOG_AND_THROW( "Invalid pixel corner" )
                  break;
                }
              }
              
              break;
            }
            case N:
            {
              switch( curr_pixel_corner ) {
                case NW :
                {
                  curr_pixel_corner = WS;  
                  break;
                }
                case EN :
                {
                  curr_chain_p += ( directions_[ S ] );
                  line.add( curr_chain_p );
                  
                  curr_chain_p += ( directions_[ W ] );
                  line.add( curr_chain_p );
                  
                  curr_chain_p += ( directions_[ N ] );
                  line.add( curr_chain_p );
                                  
                  curr_pixel_corner = WS;  
                  break;
                }
                case SE :
                {
                  curr_chain_p += ( directions_[ W ] );
                  line.add( curr_chain_p );
                  
                  curr_chain_p += ( directions_[ N ] );
                  line.add( curr_chain_p );
                                  
                  curr_pixel_corner = WS;  
                  break;
                }
                case WS :
                {
                  curr_chain_p += ( directions_[ N ] );
                  line.add( curr_chain_p );                

                  break;
                }
                default :
                {
                  TEAGN_LOG_AND_THROW( "Invalid pixel corner" )
                  break;
                }
              }
            
              break;
            }
            default :
            {
              TEAGN_LOG_AND_THROW( "Invalid new_dir" )
              break;
            }
          }
        }
          
        /* Updating the leaving direction from the initial point */
        
        if( ( ! pinit_leaving_dir_set ) && ( curr_x_index == i ) && 
            ( curr_y_index == j ) ) {
              
          pinit_leaving_dir_set = true;
          pinit_leaving_dir = new_dir;
        }           
          
        /* updating the current state */
        
        curr_x_index = next_x_index;
        curr_y_index = next_y_index;  
        
        curr_dir = new_dir;         
      }
  }
    
  /* Changing lines coords to world coords */
  
  const TeCoord2D chain_init_p = raster_->index2Coord( TeCoord2D( 
    ( (double)i ) - 0.5, ( (double)j ) - 0.5 ) );  
    
  TeLine2D::iterator it = line.begin();
  TeLine2D::iterator it_end = line.end();
  
  double min_x = DBL_MAX;
  double min_y = DBL_MAX;
  double max_x = DBL_MIN;
  double max_y = DBL_MIN;
  
  while( it != it_end ) {
    it->x_ = chain_init_p.x_ + ( it->x_ * resx_ );
    it->y_ = chain_init_p.y_ - ( it->y_ * resy_ );
    
    if( it->x_ < min_x ) min_x = it->x_;
    if( it->y_ < min_y ) min_y = it->y_;

    if( it->x_ > max_x ) max_x = it->x_;
    if( it->y_ > max_y ) max_y = it->y_;
    
    ++it;
  }
  
  /* Updating the line box */
  
  TeBox newbox( min_x, min_y, max_x, max_y );
  line.setBox( newbox );
  
  /* Closing the line, if necessary */
  
  if( line.size() > 1 ) {
    if( ( line[ 0 ].x() != line[ line.size() - 1 ].x() ) ||
      ( line[ 0 ].y() != line[ line.size() - 1 ].y() ) ) {
      
      line.add( line[ 0 ] );
    }
  }  

  return true;
}


//! the function TeFindNode build the Te2DTree of node
//  where a node is a point frontiere between at least 3 different cells
//  or 2 if it is located on the border
bool
TePDIRaster2Vector::TeFindNodes()
{

        int i, j;
        short tr;

        double  xref2=xref_-0.5*resx_;
        double  yref2=yref_+0.5*resy_;

        tree_= new TePDI2DTree<Node, TeCoord2D>(new Node(nCols_/2 , -1.0), 200, 
          true);

        //we scan the image in the raw order to find the nodes
        double c1,c2,c3,c4;
        
        StartProgInt( "Finding nodes...", nLines_ + nCols_ + nCols_ );
        
        for (j=1; j< nLines_; j++){
                IncProgInt();
        
                //beginning of the line
                if( ! raster_->getElement(0,j-1,c1,raster_channel_) ) {
                  c1 = dummy_value_;
                }
                if( ! raster_->getElement(0,j,c2,raster_channel_) ) {
                  c2 = dummy_value_;
                }
                
                if (c1 != c2) {
                  tree_->addObject(new Node(0*resx_ +xref2 ,(nLines_ - j ) * 
                    resy_ +yref2 ) );
                }

                //middle of the line
                for (i=1; i<nCols_; i++){
                        if( ! raster_->getElement(i,j-1,c3,raster_channel_) ) {
                          c3 = dummy_value_;
                        }
                        if( ! raster_->getElement(i,j,c4,raster_channel_) ) {
                          c4 = dummy_value_;
                        }
                        
                        //tr is the number of transitions
                        tr=0;
                        if (c1!=c2) tr++;
                        if (c1!=c3) tr++;
                        if (c2!=c4) tr++;
                        if (c3!=c4) tr++;
                        if (tr>2) tree_->addObject(new Node(i*resx_ + xref2 , (nLines_-j)*resy_ + yref2 ));
                        c1=c3;
                        c2=c4;
                }
                //end of the line
                if (c1!=c2) {
                        tree_->addObject(new Node(nCols_*resx_ +xref2 , (nLines_-j)*resy_ + yref2 ));
                }
        }


        //the first line
        if( ! raster_->getElement(0,0,c1,raster_channel_) ) {
          c1 = dummy_value_;
        }
  
        for (i=1; i<nCols_; i++){
                IncProgInt();
        
                raster_->getElement(i,0,c2,raster_channel_);
                if (c1!=c2){
                        tree_->addObject(new Node(i*resx_ +xref2 , (nLines_-0)*resy_+yref2 ));
                }
                c1=c2;
        }


        //the last line
        if( ! raster_->getElement(0,nLines_-1,c1,raster_channel_) ) {
          c1 = dummy_value_;
        }
  
        for (i=1; i<nCols_; i++){
                IncProgInt();
        
                if( ! raster_->getElement(i,nLines_-1,c2,raster_channel_) ) {
                  c2 = dummy_value_;
                }
                
                if (c1!=c2) {tree_->addObject(new Node(i*resx_ +xref2 , 0*resy_ + yref2 ));}
                c1=c2;
        }
        tree_->addObject(new Node(0*resx_      + xref2,    nLines_*resy_ +yref2 ));
        tree_->addObject(new Node(nCols_*resx_ + xref2,    nLines_*resy_ +yref2 ));
        tree_->addObject(new Node(0*resx_      + xref2,    0*resy_      +yref2 ));
        tree_->addObject(new Node(nCols_*resx_ + xref2,    0*resy_      +yref2 ));
        return true;
}

//! This function breaks the corners of the line
//  for each point Pi, Pi= (P[i-1]+P[i])/2
//  it needs the same distance between each point in the line
//  of input
bool
TePDIRaster2Vector::TeChanfer(TeLine2D& line)
{

        TeLine2D cp; TeCoord2D P;
    cp.copyElements(line);   //copy the elements of the line
    line.clear();
    line.add(cp[0]);

    for (unsigned int i=1 ; i < cp.size(); i++ )
        {
                //Pi= (P[i-1]+P[i])/2
                P.setXY( (cp[i-1].x()+cp[i].x())/2, (cp[i-1].y()+cp[i].y())/2);
                line.add(P);
        }
        //we add the last point
    line.add(cp[cp.size()-1]);

    return true;
}


//build a  closed polygon from a lineset and a vector of bool that
// inform when one line has been reversed
bool
TePDIRaster2Vector::TeBuildPolygon(TeLineSet &arcs, vector<bool> &reversed,
  TePolygon &poly )
{
        TeLine2D line, arc;
        poly.clear();

        for (unsigned int i =  0 ; i < arcs.size() ; i++)
        {
                arc = arcs[i];
                if (reversed[i])
                        // arc[i] has been reversed before the simplification
                        TeReverseLine(arc);
                for (unsigned int j=0; j < arc.size() ; j++)
                        line.add(arc[j]);
        }
        arcs.clear();
        reversed.clear();
        if (!line.isRing())
                line.add(line[0]);
        TeLinearRing ring(line);
        poly.add(ring);
        return true;
}


void TePDIRaster2Vector::locatePoint( const TeCoord2D& coord2Find, 
  const vector<polygonStruct>&  containerPolygons,
  const vector<unsigned int>& cadidates, bool& exist, 
  const int& curValue)
{
  exist = false;
  
  for ( unsigned int i = 0 ; i < cadidates.size() ; ++i ) {
    TEAGN_DEBUG_CONDITION( ( cadidates[ i ] < containerPolygons.size() ),
      "Invalid cadidates vector index" );
  
    const polygonStruct& pol_str_ref = containerPolygons[ cadidates[ i ] ];
    
    if (curValue == pol_str_ref.value_) {
      if( TePDIUtils::TeRelation( coord2Find, *( pol_str_ref.indexer_ ) ) ==
        TeINSIDE ) {
           
        exist = true;
        return;
      }        
    }
  }
}


bool TePDIRaster2Vector::startingEdgeTest( const int& x, 
  const int& y )
{
  int nlines = raster_->params().nlines_;
  int ncols = raster_->params().ncols_;

  TEAGN_TRUE_OR_RETURN( ( y < nlines ), 
    "Invalid col [" + Te2String( x ) + "]" );
  TEAGN_TRUE_OR_RETURN( ( x < ncols ), 
    "Invalid line [" + Te2String( y ) + "]" );
    
  if( ( x == 0 ) || ( y == 0 ) || ( x == ( ncols - 1 ) ) || 
    ( y == ( nlines - 1 ) ) ) {
    
    return true;
  } else {
    double test_val = 0;
    double current_val = 0;  
  
    if( ! raster_->getElement(x, y, current_val, raster_channel_ ) ) {
      TEAGN_TRUE_OR_RETURN( raster_uses_dummy_, "Raster read error" );
      current_val = dummy_value_;
    }
    
    if( ! raster_->getElement(x - 1, y, test_val, raster_channel_ ) ) {
      TEAGN_TRUE_OR_RETURN( raster_uses_dummy_, "Raster read error" );
      test_val = dummy_value_;
    }  
  
    TEAGN_TRUE_OR_RETURN( ( current_val != test_val ), 
      "The given point isn't a starting edge [x=" + Te2String( x ) +
      " y=" + Te2String( y ) + "]" );
      
    return true;
  }
}
