/************************************************************************************
TerraLib - a library for developing GIS applications.
Copyright  2001-2004 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular purpose.
The library provided hereunder is on an "as is" basis, and the authors have no
obligation to provide maintenance, support, updates, enhancements, or modifications.
In no event shall INPE and Tecgraf / PUC-Rio be held liable to any party for direct,
indirect, special, incidental, or consequential damages arising out of the use
of this library and its documentation.
*************************************************************************************/

/**
 * @file TePDIRaster2Vector.h
 *
 * @brief This file contains functions to support vectorizing of
 * raster data into a TerraLib database
 * the input raster is a 4-connected rotulated image :
 *               *each region is for 4-connected
 *               *each region has a unique identifier
 */

#ifndef  __TERRALIB_RASTER2VECTOR_H
#define  __TERRALIB_RASTER2VECTOR_H

#include "TePDIParameters.hpp"
#include "TePDIAlgorithm.hpp"

#include "data_structs/TeTileIndexer.h"
#include "data_structs/TePDI2DTree.hpp"

#include <TeRaster.h>
#include <TeLayer.h>
#include <TeRTree.h>


/**
  * @brief Support for a node ( node is a frontiere between 3 or 4 different regions ).
  * @author Nicolas Despres <nicolasdespres@wanadoo.fr>
  * @ingroup PDIModule
  *
  */
class Node
{
public:
        /** @brief Default constructor. */
        Node(){}

        /**
         * @brief Alternative Constructor
         *
         * @param x: abscissa of the node
         * @param y: ordinate of the node
         */
        Node(double x, double y)
        {
                coord_ = TeCoord2D(x,y);
                seenFrom_=0;
        }

        /** @brief Empty destructor. */
        ~Node() { }

        /** @brief the coordinate of the node */
        TeCoord2D coord_;

        /** @brief only the firts 4 bits of the char are used :
         * seenFrom = |left|right|up|down|.|.|.|.|
         */
        unsigned char seenFrom_;
};

/**
  * @brief Raster vectorizing.
  * @author Emiliano F. Castejon <castejon@dpi.inpe.br>
  * @author Eric S. Abreu <eric@dpi.inpe.br>
  * @author Gilberto Ribeiro de Quiroz <gribeiro@dpi.inpe.br>
  * @author Nicolas Despres <nicolasdespres@wanadoo.fr>
  * @ingroup PDIAlgorithms
  *
  * @note The general required parameters by TePDIRaster2Vector :
  *
  * @param rotulated_image (TePDITypes::TePDIRasterPtrType),
  * The rotulated_image raster ( 1 band, not floating point, 4-connected 
  * - each region is for 4-connected ).
  * @param output_polsets ( TePDITypes::TePDIPolSetMapPtrType ) - 
  * The user supplied output polygon sets map where the generated polygons 
  * will be stored (each polygon set contains polygons related to the
  * same pixel value).
  * @param channel (unsigned int) - The raster channel to process.
  *
  */
class TePDIRaster2Vector : public TePDIAlgorithm
{
  public:
   /** @typedef Type definition for a instance pointer */
   typedef TeSharedPtr< TePDIRaster2Vector > pointer;
   /** @typedef Type definition for a const instance pointer */
   typedef const TeSharedPtr< TePDIRaster2Vector > const_pointer;

   /**
    * @brief Default Constructor.
    */
   TePDIRaster2Vector();

   /**
    * @brief Default destructor.
    */
   ~TePDIRaster2Vector();

   /**
    * @brief Checks if the supplied parameters fits the requirements of each
    * PDI algorithm implementation.
    *
    * @note Error log messages must be generated. No exceptions generated.
    *
    * @param parameters The parameters to be checked.
    * @return true if the parameters are OK. false if not.
    */
    bool CheckParameters( const TePDIParameters& parameters ) const;

  protected:

    /**
     * @brief Flag indication for dummy value use ( rotulated image ).
     */
    bool raster_uses_dummy_;
    
    /**
     * @brief The used dummy value.
     */
    double dummy_value_;
    
    /**
     * @brief Directions vector.
     */ 
    TeCoord2D directions_[8];  
    
    /**
     * @brief Runs the current algorithm implementation.
     *
     * @return true if OK. false on error.
     */
    bool RunImplementation();

    /**
     * @brief Reset the internal state to the initial state.
     *
     * @param params The new parameters referente at initial state.
     */
    void ResetState( const TePDIParameters& params );

        /**
         * @brief TeDetectEdge function that detects a edge of a cell in Raster and 
         * write his shape in output line.
         *
         * @param i abscissa( column ) of the upper-left point of the shape
         * @param j ordinate( line ) of the upper-left point of the shape
         * @param line 2D Line.
         * @return true if ok, otherwise false
         */
        bool TeDetectEdge(long i, long j, TeLine2D &line );
          
        /**
         * @brief build the 2DTree of nodes
         *
         * @return true if ok, otherwise false
         */
        bool TeFindNodes();

        /**
         * @brief extratcs the arcs from the polyline ptlist
         * where an arc is polyline between 2 nodes of the precedent build 2DTree
         *
         * @param ptlist: reference to the input polyline
         * @param arcs: list of extracted arcs
         * @param reversed: vector of boolean to inform which line has been inversed
         * @return true if ok, otherwise false
         */
        bool TeExtractArcs(TeLine2D& ptlist, TeLineSet &arcs, vector<bool> &reversed);

        /**
         * @brief This function breaks the corners of the polyline line for each
     * point Pi, Pi= (P[i-1]+P[i])/2
     * it needs the same distance between each point in the line of input
         *
         * @param line: polyline to chanfer
         * @return true if ok, otherwise false
         */
        bool TeChanfer(TeLine2D &line);

        /**
         * @brief build a polyline from a list of arcs and vector of bool
         *
         * @param arcs: reference to the input list of arcs
         * @param reversed: vector of boolean to inform which line has been inversed.
         * @param poly Polygon.
         * @return true if ok, otherwise false
         */
        bool TeBuildPolygon(TeLineSet &arcs, vector<bool> &reversed, TePolygon &poly );

        /** @brief pointer to the input image */
        TePDITypes::TePDIRasterPtrType raster_;
        /** @brief boolean that is true when we smooth with chanfer algo */
        bool  chanfer_;
        /** @brief resolution of the b-spline */
        short resolution_;

        /** @brief resolution x */
        double resx_;
        
        /** @brief resolution y */
        double resy_;
        
        /** @brief lower-left reference point x */
        double xref_;
         
        /** @brief lower-left reference point y */
        double yref_;
        
        /** @brief number of lines */
        long   nLines_;
        
        /** @brief number of cols */
        long   nCols_;

        /** @brief number of cols */
        unsigned int raster_channel_;
        
        /** @brief TeDTree were we store the nodes to have a quick search
         *  in the future it could be replace by a quadtree that would give better results 
         */
        TePDI2DTree<Node,TeCoord2D>* tree_;

    /** 
     *  @brief A RTree instance pointer to optimize the searching of points
     *  inside already created polygons 
     */
    TeSAM::TeRTree<unsigned int, 8, 4>*    rtreePolygons_;
    
    /**
     *  @brief A polygon container node class. 
     */    
    class polygonStruct
    {
      public:
      
       /**
        * @brief The value ( color ) related to a polygon.
        */
       int value_;

       /**
        * @brief The polygon tile indexer Y Axis resolution.
        */
       double tile_indexer_dy_;
       
       /**
        * @brief The stored polygon instance.
        */
       TePolygon poly_;
       
       /**
        * @brief The polygon tile indexer pointer.
        */
       TeSharedPtr< TeTileIndexer > indexer_;
       
       /**
        * @brief Default constructor.
        */       
       polygonStruct()
       {
         reset( poly_, 0, 1.0 );
       };
       
       /**
        * @brief Alternative constructor.
        * @param external External reference.
        */       
       polygonStruct( const polygonStruct& external )
       {
         reset( external.poly_, external.value_, external.tile_indexer_dy_ );
       }
       
       /**
        * @brief Alternative constructor.
        * @param p Polygon reference.
        * @param v V.
        * @param tile_indexer_dy Tile indexer dy.
        */       
       polygonStruct( const TePolygon& p, const int& v, 
          const double& tile_indexer_dy )
       {
         reset( p, v, tile_indexer_dy );
       }
       
       /**
        * @brief Operator= overload.
        * @param external External reference.
        * @return a reference to the external object.
        */         
       const polygonStruct& operator=( const polygonStruct& external )
       {
         reset( external.poly_, external.value_, external.tile_indexer_dy_ );
         return external;
       }
       
       /**
        * @brief Reset the current instance.
        * @param p Polygon reference.
        * @param v V.
        * @param tile_indexer_dy Tile indexer dy.
        */       
       void reset( const TePolygon& p, const int& v, 
          const double& tile_indexer_dy )
       {
         poly_ = p;
         value_ = v;
         tile_indexer_dy_ = tile_indexer_dy;
         indexer_.reset( new TeTileIndexer( poly_, tile_indexer_dy_ ) );       
       }
    };
    
    /** @brief Vector of all polygons */
    vector<polygonStruct>  containerPolygons_;

    /**
     * Verifies if a given point is inside of one the candidate polygons.
     *
     * @param coord2Find The point coordinates.
     * @param containerPolygons The container of all polygons.
     * @param cadidates A vector of candidates polygons indexes.
     * @param exist Will be true if the point was found inside one of candidate
     * polygons.
     * @param curValue The value assotiated with the point (and the related
     * polygon).
     */    
    void locatePoint( const TeCoord2D& coord2Find, 
      const vector<polygonStruct>&  containerPolygons,
      const vector<unsigned int>& cadidates, bool& exist, 
      const int& curValue );
      
    /**
     * @brief Tests if the current point is a edge start.
     * @param x X coord.
     * @param y Y coord.
     * @return true if the current point is a edge start, false if not.
     */
    bool startingEdgeTest( const int& x, const int& y );

};


/** @example TePDIRaster2Vector_test.cpp
 *    Shows how to use this class.
 */  
 
#endif
