#include "TePDIUtils.hpp"

#include "TePDIInterpolator.hpp"
#include "TePDIPIManager.hpp"

#include <TeDecoderMemory.h>
#include <TeDecoderMemoryMap.h>
#include <TeDecoderTIFF.h>
#include <TeRaster.h>
#include <TeRasterParams.h>
#include <TeDataTypes.h>
#include <TeUtils.h>
#include <TeDefines.h>
#include <TeVectorRemap.h>
#include <TeRasterTransform.h>
#include <TeRasterRemap.h>
#include <TeProgress.h>
#include <TeAgnostic.h>

#include <float.h>
#include <math.h>
#include <stdio.h>


namespace TePDIUtils{


  bool TeAllocRAMRaster(
    TePDITypes::TePDIRasterPtrType& template_raster,
    TePDITypes::TePDIRasterPtrType& RAMRaster,
    unsigned int bands, bool force_new_dt, bool force_mm,
    TeDataType pixel_type )
  {
    TEAGN_TRUE_OR_RETURN( template_raster.isActive(),
      "Inactive Input Raster" );
    TEAGN_TRUE_OR_RETURN( template_raster->params().status_ != TeRasterParams::TeNotReady,
      "Input not Ready" );

    TeRasterParams temp_params = template_raster->params();

    if( bands != 0 ) {
      temp_params.nBands( bands );
    }

    if( force_new_dt ) {
      temp_params.setDataType( pixel_type, -1 );
    }

    return TeAllocRAMRaster( RAMRaster, temp_params, force_mm );
  }


  bool TeAllocRAMRaster(
    TePDITypes::TePDIRasterPtrType& RAMRaster,
    unsigned int bands, unsigned int lines, unsigned int columns,
    bool force_mm, TeDataType pixel_type, TePDIRgbPalette* palette )
  {
    TEAGN_TRUE_OR_RETURN( bands > 0, "Invalid bands number" );
    TEAGN_TRUE_OR_RETURN( lines > 0, "Invalid lines number" );
    TEAGN_TRUE_OR_RETURN( columns > 0, "Invalid columns number" );

    TeRasterParams temp_params;

    temp_params.nBands( bands );

    temp_params.setDataType( pixel_type, -1 );
    
    temp_params.setNLinesNColumns( lines, columns );
    
    if( palette != 0 ) {
      temp_params.setPhotometric( TeRasterParams::TePallete );

      temp_params.lutr_.clear();
      temp_params.lutg_.clear();
      temp_params.lutb_.clear();

      TePDIRgbPalette::iterator pal_it = palette->begin();
      TePDIRgbPalette::iterator pal_it_end = palette->end();

      for( unsigned int lut_index = 0 ; lut_index < palette->size() ;
           ++lut_index ) {

        temp_params.lutr_.push_back( (unsigned short)pal_it->second.red_ );
        temp_params.lutg_.push_back( (unsigned short)pal_it->second.green_ );
        temp_params.lutb_.push_back( (unsigned short)pal_it->second.blue_ );

        ++pal_it;
      }
    }    

    return TeAllocRAMRaster( RAMRaster, temp_params, force_mm );
  }
  
  bool TeAllocRAMRaster( TePDITypes::TePDIRasterPtrType& RAMRaster,
    const TeRasterParams& params, bool force_mm )
  {
    return TeAllocRAMRaster( RAMRaster, params,
      ( force_mm ? TePDIUtilsDiskMemPol : TePDIUtilsAutoMemPol ) );
  }

  
  bool TeAllocRAMRaster( TePDITypes::TePDIRasterPtrType& RAMRaster,
    const TeRasterParams& params, TePDIUtilsMemPol mempol )
  {
    TeRasterParams internal_params = params;
    
    internal_params.mode_ = 'c';
  
    if( mempol == TePDIUtilsRamMemPol ) {
      internal_params.decoderIdentifier_ = "MEM";
    } else if( mempol == TePDIUtilsDiskMemPol ) {
      internal_params.decoderIdentifier_ = "MEMMAP";
      
      TEAGN_TRUE_OR_RETURN( TeGetTempFileName( internal_params.fileName_ ),
        "Unable to get temporary file name" );         
    } else {
      /* Auto mem pol */
      internal_params.decoderIdentifier_ = "SMARTMEM";
    }
    
    if( RAMRaster.isActive() ) {
      RAMRaster->updateParams( internal_params );
      return RAMRaster->init( internal_params );
    } else {
      RAMRaster.reset( new TeRaster( internal_params ) );
      return RAMRaster->init();
    }
  }
    

  bool TeRaster2Geotiff(
    TePDITypes::TePDIRasterPtrType& in_raster,
    const std::string& file_name, TeDataType pixel_type )
  {
    TEAGN_TRUE_OR_RETURN( in_raster.isActive(),
      "Inactive Input Raster" );
    TEAGN_TRUE_OR_RETURN( in_raster->params().status_ != 
      TeRasterParams::TeNotReady, "Input not Ready" );
    TEAGN_TRUE_OR_RETURN( file_name.size() != 0,
      "Invalid file name" );

    TeRasterParams temp_params = in_raster->params();

    temp_params.mode_ = 'c';
    temp_params.fileName_ = file_name;
    temp_params.setDataType( pixel_type );
    
    if( ( temp_params.nBands() == 3 ) &&
      ( temp_params.photometric_[ 0 ] == TeRasterParams::TeMultiBand ) &&
      ( temp_params.photometric_[ 1 ] == TeRasterParams::TeMultiBand ) &&
      ( temp_params.photometric_[ 2 ] == TeRasterParams::TeMultiBand ) ) {
      
      temp_params.setPhotometric( TeRasterParams::TeRGB, -1 );
    }
    
    /* Creating disk output raster */
    
    TePDITypes::TePDIRasterPtrType outRaster( new TeRaster() );

    outRaster->setDecoder( new TeDecoderTIFF( temp_params ) );

    TEAGN_TRUE_OR_RETURN( outRaster->init(),
      "Unable to init GeoTIFF Raster" );
      
    return TeCopyRasterBands( in_raster, outRaster, true, true );
  }


  bool TeCopyRasterPixels(
    TePDITypes::TePDIRasterPtrType& source_raster,
    unsigned int source_band,
    TePDITypes::TePDIRasterPtrType& target_raster,
    unsigned int target_band,
    bool discretize_levels,
    bool trunc_levels )
  {
    TEAGN_TRUE_OR_RETURN( source_raster.isActive(),
      "Inactive Input Raster" );
    TEAGN_TRUE_OR_RETURN( target_raster.isActive(),
      "Inactive Output Raster" );
    TEAGN_TRUE_OR_RETURN( source_raster->params().status_ != TeRasterParams::TeNotReady,
      "Input not Ready" );
    TEAGN_TRUE_OR_RETURN( target_raster->params().status_ == TeRasterParams::TeReadyToWrite,
      "OutPut not Ready" );

    TeRasterParams& source_params = source_raster->params();
    TeRasterParams& target_params = target_raster->params();

    TEAGN_TRUE_OR_RETURN( (int)source_band < source_params.nBands(),
      "Invalid source band" );
    TEAGN_TRUE_OR_RETURN( (int)target_band < target_params.nBands(),
      "Invalid target band" );
    TEAGN_TRUE_OR_RETURN( (
      ( source_params.nlines_ == target_params.nlines_ ) &&
      ( source_params.ncols_ == target_params.ncols_ ) ),
      "Input and Output Raster's have different dimentions" );
      
    /* Dummy support */
    
    const bool source_raster_uses_dummy = source_raster->params().useDummy_;
    double source_raster_dummy = 0;
    if( source_raster_uses_dummy ) {
      source_raster_dummy = source_raster->params().dummy_[ source_band ];
    }
    
    const bool target_raster_uses_dummy = target_raster->params().useDummy_;
    double target_raster_dummy = 0;
    if( target_raster_uses_dummy ) {
      target_raster_dummy = target_raster->params().dummy_[ target_band ];
    } else if( source_raster_uses_dummy ) {
      target_raster_dummy = source_raster_dummy;
    }

    /* Defining the valid values range base on the output raster pixel type */

    double min_value = 0;
    double max_value = 0;

    TEAGN_TRUE_OR_RETURN(
      TeGetRasterMinMaxBounds( target_raster, target_band, min_value, max_value ),
      "Unable to get raster level bounds" );

    /* Pixel by pixel copy loop */

    double pixel_value;

    for( int line = 0 ; line < source_params.nlines_ ; ++line ) {
      for( int column = 0 ; column < source_params.ncols_ ; ++column ) {
        if( source_raster->getElement( column, line, pixel_value, 
          source_band ) ) {
          
          if( discretize_levels ) {
            pixel_value = TeRound( pixel_value );
          }

          if( trunc_levels ) {
            if( pixel_value < min_value ) {
              pixel_value = min_value;
            } else if( pixel_value > max_value ) {
              pixel_value = max_value;
            }
          }

          TEAGN_TRUE_OR_RETURN( target_raster->setElement( column, line,
            pixel_value, target_band ), "Pixel copy error" );
        } else {
          TEAGN_TRUE_OR_RETURN( source_raster_uses_dummy,
            "Unable to read from input raster" );
            
          TEAGN_TRUE_OR_RETURN( target_raster->setElement( column, line,
            target_raster_dummy, target_band ), "Pixel copy error" );          
        }
      }
    }

    return true;
  }


  bool TeCopyRasterBands(
    TePDITypes::TePDIRasterPtrType& source_raster,
    TePDITypes::TePDIRasterPtrType& target_raster,
    bool discretize_levels,
    bool trunc_levels )
  {
    TEAGN_TRUE_OR_RETURN( source_raster.isActive(),
      "Inactive Input Raster" );
    TEAGN_TRUE_OR_RETURN( target_raster.isActive(),
      "Inactive Output Raster" );
    TEAGN_TRUE_OR_RETURN( source_raster->params().status_ != TeRasterParams::TeNotReady,
      "Input not Ready" );
    TEAGN_TRUE_OR_RETURN( target_raster->params().status_ == TeRasterParams::TeReadyToWrite,
      "OutPut not Ready" );

    TeRasterParams& source_params = source_raster->params();
    TeRasterParams& target_params = target_raster->params();

    TEAGN_TRUE_OR_RETURN(
      ( ( source_params.nBands() == target_params.nBands() ) &&
      ( source_params.nlines_ == target_params.nlines_ ) &&
      ( source_params.ncols_ == target_params.ncols_ ) ),
      "Input and Output Raster's have different dimentions" );

    for( int band = 0 ; band < source_params.nBands() ; ++band ) {
      TEAGN_TRUE_OR_RETURN( TeCopyRasterPixels( source_raster, band, target_raster,
        band, discretize_levels,trunc_levels ), "Unable to copy band " +
        Te2String( band ) );
    }

    return true;
  }

  
  bool TeGetRasterMinMaxBounds( TePDITypes::TePDIRasterPtrType& raster,
    unsigned int band, double& min, double& max )
  {
    TEAGN_TRUE_OR_RETURN( raster.isActive(),
      "Inactive Input Raster" );

    TeRasterParams& params = raster->params();

    TEAGN_TRUE_OR_RETURN( (int)band < params.nBands(),
      "Invalid source band" );

    switch( params.dataType_[ band ] ) {
      case TeBIT :
      {
        max = 1;
        min = 0;
        break;
      }
      case TeUNSIGNEDCHAR :
      {
        max = pow( 2., (double)sizeof( unsigned char ) * 8. ) - 1;
        min = 0;
        break;
      }
      case TeUNSIGNEDSHORT :
      {
        max = pow( 2., (double)sizeof( short ) * 8. ) - 1;
        min = 0;
        break;
      }
      case TeUNSIGNEDLONG :
      {
        max = pow( 2., (double)sizeof( unsigned long ) * 8. ) - 1;
        min = 0;
        break;
      }
      case TeCHAR :
      {
        max = ( pow( 2., (double)sizeof( char ) * 8. ) ) / 2;
        min = -1 * max;
        break;
      }
      case TeSHORT :
      {
        max = ( pow( 2., (double)sizeof( short ) * 8. ) ) / 2;
        min = -1 * max;
        break;
      }
      case TeINTEGER :
      {
        max = ( pow( 2., (double)sizeof( int ) * 8. ) ) / 2;
        min = -1 * max;
        break;
      }
      case TeLONG :
      {
        max = ( pow( 2., (double)sizeof( long ) * 8. ) ) / 2;
        min = -1 * max;
        break;
      }
      case TeFLOAT :
      {
        max = FLT_MAX;
        min = -1 * max;
        break;
      }
      case TeDOUBLE :
      {
        max = DBL_MAX;
        min = -1 * max;
        break;
      }
      default :
      {
        TEAGN_LOG_AND_RETURN( "Invalid raster pixel type" );
        break;
      }
    }

    return true;
  }
  
  
  unsigned long int TeGetRasterReqMem( TeRasterParams& params )
  {
    unsigned long int req_mem = 0;

    for( int band = 0 ; band < params.nBands() ; ++band ) {
      req_mem += params.elementSize( band ) * 
        params.nlines_ * params.ncols_;
    }
    
    req_mem += sizeof( unsigned short ) * params.lutr_.size();
    req_mem += sizeof( unsigned short ) * params.lutg_.size();
    req_mem += sizeof( unsigned short ) * params.lutb_.size();  
    
    return req_mem;
  }

  
  bool rasterCompare( TePDITypes::TePDIRasterPtrType& raster1,
    TePDITypes::TePDIRasterPtrType& raster2 )
  {
    TEAGN_TRUE_OR_THROW( raster1.isActive(), "Raster1 not active" );
    TEAGN_TRUE_OR_THROW( raster2.isActive(), "Raster2 not active" );
    TEAGN_TRUE_OR_THROW( raster1->params().status_ != TeRasterParams::TeNotReady,
      "Raster1 not ready" );
    TEAGN_TRUE_OR_THROW( raster2->params().status_ != TeRasterParams::TeNotReady,
      "Raster2 not ready" );

    TeRasterParams par1 = raster1->params();
    TeRasterParams par2 = raster2->params();

    if( ( par1.nBands() != par2.nBands() ) ||
        ( par1.ncols_ != par2.ncols_ ) ||
        ( par1.nlines_ != par2.nlines_ ) ) {

      return false;
    }

    for( int bandt = 0 ; bandt < par1.nBands() ; ++bandt ) {
      if( ( par1.dataType_[ bandt ] != par2.dataType_[ bandt ] ) ||
          ( par1.photometric_[ bandt ] != par2.photometric_[ bandt ] ) ) {

        return false;
      }
    }

    unsigned int lines = ( unsigned int ) par1.nlines_;
    unsigned int columns = ( unsigned int ) par1.ncols_;
    double value1 = 0;
    double value2 = 0;

    for( int band = 0 ; band < par1.nBands() ; ++band ) {
      for( unsigned int line = 0 ; line < lines ; ++line ) {
        for( unsigned int column = 0 ; column < columns ; ++column ) {
          if( raster1->getElement( column, line, value1, band ) !=
              raster2->getElement( column, line, value2, band ) ) {

            return false;
          }

          if( value1 != value2 ) {
            return false;
          }
        }
      }
    }

    return true;
  }
  
  
  bool IsFloatBand( const TePDITypes::TePDIRasterPtrType& raster, 
    unsigned int band  )
  {
    TEAGN_TRUE_OR_THROW( raster.isActive(),
      "Inactive Input Raster" );

    TeRasterParams& params = raster->params();

    TEAGN_TRUE_OR_THROW( (int)band < params.nBands(),
      "Invalid source band" );

    switch( params.dataType_[ band ] ) {
      case TeBIT :
      {
        return false;
        break;
      }
      case TeUNSIGNEDCHAR :
      {
        return false;
        break;
      }
      case TeUNSIGNEDSHORT :
      {
        return false;
        break;
      }
      case TeUNSIGNEDLONG :
      {
        return false;
        break;
      }
      case TeCHAR :
      {
        return false;
        break;
      }
      case TeSHORT :
      {
        return false;
        break;
      }
      case TeINTEGER :
      {
        return false;
        break;
      }
      case TeLONG :
      {
        return false;
        break;
      }
      case TeFLOAT :
      {
        return true;
        break;
      }
      case TeDOUBLE :
      {
        return true;
        break;
      }
      default :
      {
        TEAGN_LOG_AND_THROW( "Invalid raster pixel type" );
        return false;
        break;
      }
    }
  }
  
  
  void MapCoords2RasterIndexes( const TeCoord2D& g_in, 
    const TePDITypes::TePDIRasterPtrType& raster, TeCoord2D& g_out )
  {
    TEAGN_TRUE_OR_THROW( raster.isActive(), "Inactive raster" );
    
    g_out = raster->coord2Index( g_in );
  };
  
  
  void MapCoords2RasterIndexes( const TeBox& box_in, 
      const TePDITypes::TePDIRasterPtrType& raster, TeBox& box_out )
  {
    TEAGN_TRUE_OR_THROW( raster.isActive(), "Inactive raster" );
    
    TeBox temp_box( raster->coord2Index( box_in.lowerLeft() ),
      raster->coord2Index( box_in.upperRight() ) );
      
    box_out = temp_box;
  };  
  
  
  void MapRasterIndexes2Coords( const TeCoord2D& g_in, 
    const TePDITypes::TePDIRasterPtrType& raster, TeCoord2D& g_out )
  {
    TEAGN_TRUE_OR_THROW( raster.isActive(), "Inactive raster" );
      
    g_out = raster->index2Coord( g_in );
  };  
  
  
  void MapRasterIndexes2Coords( const TeBox& box_in, 
      const TePDITypes::TePDIRasterPtrType& raster, TeBox& box_out )
  {
    TEAGN_TRUE_OR_THROW( raster.isActive(), "Inactive raster" );
    
    TeBox temp_box( raster->index2Coord( box_in.lowerLeft() ),
      raster->index2Coord( box_in.upperRight() ) );
      
    box_out = temp_box;
  };   
  
  
  void makeSegmentSet( const TeLinearRing& g_in, TeLineSet& g_out )
  {
    g_out.clear();
    
    const int bound = g_in.size() - 1 ;
    
    for( int index = 0 ; index < bound ; ++index ) {
      TeLine2D line;
      
      line.add( g_in[ index ] );
      line.add( g_in[ index + 1 ] );
      
      g_out.add( line );
    }
  }
  
  
  bool IsDiscreteLut( const TePDITypes::TePDILutType& lut )
  {
    TEAGN_TRUE_OR_THROW( ( lut.size() > 0 ), "Emptry lut" );
  
    TePDITypes::TePDILutType::const_iterator it = lut.begin();
    const TePDITypes::TePDILutType::const_iterator it_end = lut.end();
  
    while( it != it_end ) {
      if( it->first != floor( it->first ) ) {
        return false;
      }
      
      ++it;
    }
    
    return true;
  }
  

  bool FindBestDecimLevels( 
    const TePDITypes::TePDIRasterPtrType& raster1,
    const TePDITypes::TePDIRasterPtrType& raster2,
    unsigned long int& r1_level, unsigned long int& r2_level,
    unsigned long int& decim_error )
  {
    TEAGN_TRUE_OR_RETURN( raster1.isActive(), "raster1 inactive" );
    TEAGN_TRUE_OR_RETURN( raster2.isActive(), "raster2 inactive" );
  
    const unsigned long int r1_lines = raster1->params().nlines_;
    const unsigned long int r1_cols = raster1->params().ncols_;
    const unsigned long int r2_lines = raster2->params().nlines_;
    const unsigned long int r2_cols = raster2->params().ncols_;
  
    const unsigned long int r1_max_levels = MIN(
      TePDIMathFunctions::DecimLevels( r1_lines ),
      TePDIMathFunctions::DecimLevels( r1_cols ) );  
    const unsigned long int r2_max_levels = MIN(
      TePDIMathFunctions::DecimLevels( r2_lines ),
      TePDIMathFunctions::DecimLevels( r2_cols ) );
    
    if( ( r1_max_levels < 1 ) || ( r2_max_levels < 1 ) ) {
      r1_level = r2_level = 0;
      return false;
    } else {
      bool best_level_diff_found = false;
      
      unsigned long int curr_diff = 0;
      long int curr_r1_lines = r1_lines;
      long int curr_r1_cols = r1_cols;
      long int curr_r2_lines = 0;
      long int curr_r2_cols = 0;
      
      for( unsigned long int curr_r1_level = 1 ; 
        curr_r1_level <= r1_max_levels ; 
        ++curr_r1_level ) {
        
        curr_r1_lines = ( long int )
          ceil( ( (double)curr_r1_lines ) / 2 );
        curr_r1_cols = ( long int )
          ceil( ( (double)curr_r1_cols ) / 2 );      
          
        curr_r2_lines = r2_lines;
        curr_r2_cols = r2_cols;        
      
        for( unsigned long int curr_r2_level = 1 ; 
          curr_r2_level <= r2_max_levels ; 
          ++curr_r2_level ) {
        
          curr_r2_lines = ( long int )
            ceil( ( (double)curr_r2_lines ) / 2 );
          curr_r2_cols = ( long int )
            ceil( ( (double)curr_r2_cols ) / 2 );
            
          curr_diff = ABS( curr_r1_lines - curr_r2_lines ) +
            ABS( curr_r1_cols - curr_r2_cols );
            
          if( best_level_diff_found ) {
            if( curr_diff < decim_error ) {
              r1_level = curr_r1_level;
              r2_level = curr_r2_level;
              decim_error = curr_diff;
            }
          } else {
            r1_level = curr_r1_level;
            r2_level = curr_r2_level;
            decim_error = curr_diff;
            best_level_diff_found = true;
          }
        }
      }
      
      return true;
    }
  }
  
  bool buildDetailedBox( const TePDITypes::TePDIRasterPtrType& raster,
    TePolygon& box )
  {
    box.clear();
    
    TEAGN_TRUE_OR_RETURN( raster.isActive(), "raster inactive" );
    
    bool raster_is_projected = true;
    if( raster->params().projection() == 0 ) {
      raster_is_projected = false;
    } else if( raster->params().projection()->name() == "NoProjection" ) {
      raster_is_projected = false;
    }
    
    TeLinearRing lr;
    double lines = (double)raster->params().nlines_;
    double cols = (double)raster->params().ncols_;
    double offset = 0;
    
    /* Adding top box border ( direction left -> right )*/
    
    for( offset = 0 ; offset < cols ; ++offset ) {
      if( raster_is_projected ) {
        lr.add( raster->index2Coord( TeCoord2D( offset, 0 ) ) );
      } else {
        lr.add( TeCoord2D( offset, 0 ) );
      }
    }
    
    /* Adding right box border ( direction up -> down) */
    
    for( offset = 1 ; offset < lines ; ++offset ) {
      if( raster_is_projected ) {
        lr.add( raster->index2Coord( TeCoord2D( cols - 1, offset ) ) );
      } else {
        lr.add( TeCoord2D( cols - 1, offset ) );
      }
    }
    
    /* Adding bottom box border ( direction right ->left )*/
    
    for( offset = 1 ; offset < cols ; ++offset ) {
      if( raster_is_projected ) {
        lr.add( raster->index2Coord( 
          TeCoord2D( cols - 1 - offset, lines - 1 ) ) );
      } else {
        lr.add( TeCoord2D( cols - 1 - offset, lines - 1 ) );
      }
    }      
    
    /* Adding left box border ( bottom -> top )*/
    
    for( offset = 1 ; offset < lines ; ++offset ) {
      if( raster_is_projected ) {
        lr.add( raster->index2Coord( 
          TeCoord2D( 0, lines - 1 - offset ) ) );
      } else {
        lr.add( TeCoord2D( 0, lines - 1 - offset ) );
      }
    }
        
    box.add( lr );
    
    return true;
  }
  
  
  bool buildDetailedBBox( const TePDITypes::TePDIRasterPtrType& raster,
    TePolygon& box )
  {
    box.clear();
    
    TEAGN_TRUE_OR_RETURN( raster.isActive(), "raster inactive" );
    
    bool raster_is_projected = true;
    if( raster->params().projection() == 0 ) {
      raster_is_projected = false;
    } else if( raster->params().projection()->name() == "NoProjection" ) {
      raster_is_projected = false;
    }    
    
    TeLinearRing lr;
    const double lines = (double)raster->params().nlines_;
    const double cols = (double)raster->params().ncols_;
    double offset = 0;
    
    /* Adding top box border ( direction left -> right )*/
    
    for( offset = 0 ; offset <= cols ; ++offset ) {
      if( raster_is_projected ) {
        lr.add( raster->index2Coord( TeCoord2D( ( offset - 0.5 ), -0.5 ) ) );
      } else {
        lr.add( TeCoord2D( ( offset - 0.5 ), -0.5 ) );
      }
    }
    
    /* Adding right box border ( direction up -> down) */
    
    for( offset = 0 ; offset < lines ; ++offset ) {
      if( raster_is_projected ) {
        lr.add( raster->index2Coord( 
          TeCoord2D( ( cols - 0.5 ), offset + 0.5) ) );
      } else {
        lr.add( TeCoord2D( ( cols - 0.5 ), offset + 0.5) );
      }
    }
    
    /* Adding bottom box border ( direction right ->left )*/
    
    for( offset = 0 ; offset < cols ; ++offset ) {
      if( raster_is_projected ) {
        lr.add( raster->index2Coord( 
          TeCoord2D( ( cols - 1.5 - offset ), ( lines - 0.5 ) ) ) );
      } else {
        lr.add( TeCoord2D( ( cols - 1.5 - offset ), ( lines - 0.5 ) ) );
      }
    }      
    
    /* Adding left box border ( bottom -> top )*/
    
    for( offset = 0 ; offset < lines ; ++offset ) {
      if( raster_is_projected ) {
        lr.add( raster->index2Coord( 
          TeCoord2D( -0.5, ( lines - 1.5 - offset ) ) ) );
      } else {
        lr.add( TeCoord2D( -0.5, ( lines - 1.5 - offset ) ) );
      }
    }
        
    box.add( lr );
    
    return true;
  }  
  
  
  void getGeodeticPixelRes( TeRasterParams& params,
    double& x_res, double& y_res )
  {
    TEAGN_TRUE_OR_THROW( ( params.projection() != 0 ),
      "Missing projection" )
    TEAGN_TRUE_OR_THROW( ( params.projection()->name() != "NoProjection" ),
      "Invalid projection" );
      
    TeCoord2D pixel00( 0, 0 );
    TeCoord2D rightpixel( 1, 0 );
    TeCoord2D bottompixel( 0, 1 );
      
    pixel00 = params.index2Coord( pixel00 );
    rightpixel = params.index2Coord( rightpixel );
    bottompixel = params.index2Coord( bottompixel );
    
    pixel00 = params.projection()->PC2LL( pixel00 );
    rightpixel = params.projection()->PC2LL( rightpixel );
    bottompixel = params.projection()->PC2LL( bottompixel );
    
    x_res = TeDistance( rightpixel, pixel00 );
    y_res = TeDistance( bottompixel, pixel00 );  
  }
  
    
  bool reprojectRaster( const TePDITypes::TePDIRasterPtrType& input_raster,
    TeProjection& target_projection, double target_resx, double target_resy,
    TePDITypes::TePDIRasterPtrType& output_raster )
  {
    TEAGN_TRUE_OR_RETURN( input_raster.isActive(), 
      "input_raster not active" );
    TEAGN_TRUE_OR_RETURN( output_raster.isActive(), "Raster2 not active" );
    TEAGN_TRUE_OR_RETURN( 
      input_raster->params().status_ != TeRasterParams::TeNotReady,
      "input_raster not ready" );
    TEAGN_TRUE_OR_RETURN( 
      output_raster->params().status_ != TeRasterParams::TeNotReady,
      "output_raster not ready" ); 
    TEAGN_TRUE_OR_RETURN( ( input_raster->params().projection() != 0 ),
      "Missing input_raster projection" )
    TEAGN_TRUE_OR_RETURN( 
      ( input_raster->params().projection()->name() != "NoProjection" ),
      "Invalid input_raster projection" );
      
    /* Building the remmaped box */
      
    TePolygon original_box_pol;
    TEAGN_TRUE_OR_RETURN( TePDIUtils::buildDetailedBox( input_raster,
      original_box_pol ), 
      "Unable to build detailed bounding box" );
        
    TePolygon reprojected_box_pol;  
    TeVectorRemap( original_box_pol, input_raster->params().projection(),
      reprojected_box_pol, &target_projection );    
      
    /* Remmaping raster */
  
    TeRasterParams output_raster_params = output_raster->params();
    
    output_raster_params.projection( &target_projection );
    
    output_raster_params.nBands( input_raster->params().nBands() );
    output_raster_params.setDataType( 
      input_raster->params().dataType_[ 0 ] );
    output_raster_params.setPhotometric( 
      input_raster->params().photometric_[ 0 ] );
    if( input_raster->params().useDummy_ ) {
      output_raster_params.setDummy( 
        input_raster->params().dummy_[ 0 ], -1 );
    }    
    
    TeBox reprojected_box = reprojected_box_pol.box();
    output_raster_params.boxResolution( reprojected_box.x1(), 
      reprojected_box.y1(), reprojected_box.x2(), reprojected_box.y2(), 
      target_resx, target_resy );
      
    TEAGN_TRUE_OR_RETURN( output_raster->init( output_raster_params ),
      "Output raster init error" );
      
    TeRasterTransform transf( input_raster.nakedPointer(),
      output_raster.nakedPointer() );
    transf.setTransfFunction( TeRasterTransform::TeBand2Band );
      
    TeRasterRemap remap( input_raster.nakedPointer(), 
      output_raster.nakedPointer() ) ;
    remap.setTransformer( &transf );
    
    bool progress_interface_enabled = false;
    if( TeProgress::instance() != 0 ) {
      progress_interface_enabled = true;
    }
    
    TEAGN_TRUE_OR_RETURN( remap.apply( progress_interface_enabled ),
      "Raster remapping error" );
      
    return true;
  }
  
  
  short TeRelation( const TeCoord2D& c, const TeTileIndexer& indexer )
  {
    TeTileIndexer::TeTileSegIndex* tindex_ptr;
    const TePolygon& pol = indexer.getPol();
    
    indexer.getTile( c.y(), &tindex_ptr );
    
    if( tindex_ptr == 0 ) {
      return TeOUTSIDE;
    }
    
    register double ty, tx;
    register bool inside_flag = false;
    register int yflag0, yflag1;
  
    TeLinearRing::iterator vtx0, vtx1;
  
    tx = c.x();
    ty = c.y();
    
    for( unsigned int i = 0 ; i < tindex_ptr->size() ; ++i ) {
      TEAGN_DEBUG_CONDITION( ( ( *tindex_ptr )[ i ].first < pol.size() ),
        "Invalid ring index" );
      TEAGN_DEBUG_CONDITION( ( ( ( *tindex_ptr )[ i ].second + 1 ) < 
        pol[ ( *tindex_ptr )[ i ].first ].size() ),
        "Invalid segment index" );
    
      const TeCoord2D& vtx0 = 
        pol[ ( *tindex_ptr )[ i ].first ][ ( *tindex_ptr )[ i ].second ];
      const TeCoord2D& vtx1 = 
      pol[ ( *tindex_ptr )[ i ].first ][ ( *tindex_ptr )[ i ].second + 1 ];
      
      yflag0 = (vtx0.y() >= ty);
      
      yflag1 = (vtx1.y() >= ty);
    
      if(yflag0 != yflag1)
      {
        /* TODO - Check Boundary case */
      
        if(((vtx1.y() - ty) * (vtx0.x() - vtx1.x()) >=
              (vtx1.x() - tx) * (vtx0.y() - vtx1.y())) == yflag1)
        {
          inside_flag = !inside_flag ;
        }
      }
    
    }
    
    if( inside_flag ) {
      return TeINSIDE;
    }
    
    return TeOUTSIDE;
  }  
  
  
  void TeSegSetIntersection( const TeLineSet& ls1, const TeLineSet& ls2,
    TePointSet& ps )
  {
    ps.clear();
    
    const unsigned int ls1_size = ls1.size();
    const unsigned int ls2_size = ls2.size();
    
    unsigned int ls1_index = 0;
    unsigned int ls2_index = 0;
     
    bool line1_is_vertical = false;
    double line1_vertical_x_value = 0;
    double line1_m = 0;
    double line1_q = 0;
    
    double line2_m = 0;
    double line2_q = 0;    
    
    TeCoord2D candidate_pt;
    double canditate_pt_x = 0;
    bool candidate_pt_found = false;
    
    double line1_min_x = 0;
    double line1_min_y = 0;
    double line1_max_x = 0;
    double line1_max_y = 0;
    
    double line2_min_x = 0;
    double line2_min_y = 0;
    double line2_max_x = 0;
    double line2_max_y = 0;
    
    for( ls1_index = 0 ; ls1_index < ls1_size ; ++ls1_index ) {
      const TeLine2D& line1 = ls1[ ls1_index ];
      
      TEAGN_TRUE_OR_THROW( ( line1.size() == 2 ), "Invalid line size" );
      
      const TeCoord2D& line1_p1 = line1[ 0 ];
      const TeCoord2D& line1_p2 = line1[ 1 ];
      
      line1_min_x = MIN( line1_p1.x(), line1_p2.x() );
      line1_min_y = MIN( line1_p1.y(), line1_p2.y() );
      line1_max_x = MAX( line1_p1.x(), line1_p2.x() );
      line1_max_y = MAX( line1_p1.y(), line1_p2.y() );
      
      if( line1_p1.x() == line1_p2.x() ) {
        line1_is_vertical = true;
        
        line1_vertical_x_value = line1_p1.x();
      } else {
        line1_is_vertical = false;
        
        line1_m = ( line1_p1.y() - line1_p2.y() ) /
          ( line1_p1.x() - line1_p2.x() );
        line1_q = line1_p1.y() - ( line1_m * line1_p1.x() );
      }
    
      for( ls2_index = 0 ; ls2_index < ls2_size ; ++ls2_index ) {
        candidate_pt_found = false;
        const TeLine2D& line2 = ls2[ ls2_index ];
        
        TEAGN_TRUE_OR_THROW( ( line2.size() == 2 ), "Invalid line size" );
        
        const TeCoord2D& line2_p1 = line2[ 0 ];
        const TeCoord2D& line2_p2 = line2[ 1 ];
        
        if( line2_p1.x() == line2_p2.x() ) {
          if( ! line1_is_vertical ) {
            candidate_pt.setXY( line2_p1.x(), ( line1_m * line2_p1.x() ) +
              line1_q );
            candidate_pt_found = true;
          }
        } else {
          line2_m = ( line2_p1.y() - line2_p2.y() ) /
            ( line2_p1.x() - line2_p2.x() );
          line2_q = line2_p1.y() - ( line2_m * line2_p1.x() );        
        
          if( line1_is_vertical ) {
            candidate_pt.setXY( line1_vertical_x_value, 
              ( line2_m * line1_vertical_x_value ) + line2_q );
            candidate_pt_found = true;
          } else {
            if( line1_m != line2_m ) {
              canditate_pt_x = (  line2_q - line1_q ) / ( line1_m - line2_m );
              candidate_pt.setXY( canditate_pt_x, ( line2_m * canditate_pt_x ) + 
                line2_q );
              candidate_pt_found = true;
            }
          }
        }
        
        if( candidate_pt_found ) {
          line2_min_x = MIN( line2_p1.x(), line2_p2.x() );
          line2_min_y = MIN( line2_p1.y(), line2_p2.y() );
          line2_max_x = MAX( line2_p1.x(), line2_p2.x() );
          line2_max_y = MAX( line2_p1.y(), line2_p2.y() );        
          
          if( ( candidate_pt.x() >= line1_min_x  ) &&
            ( candidate_pt.y() >= line1_min_y  ) &&
            ( candidate_pt.x() <= line1_max_x  ) &&
            ( candidate_pt.y() <= line1_max_y  ) &&
            ( candidate_pt.x() >= line2_min_x  ) &&
            ( candidate_pt.y() >= line2_min_y  ) &&
            ( candidate_pt.x() <= line2_max_x  ) &&
            ( candidate_pt.y() <= line2_max_y  ) ) {
            
            ps.add( candidate_pt );
          }
          
        }
      }
    }
  
  }
  
  TeDataType chooseBestPixelType( TePDITypes::TePDIRasterVectorType& rasters )
  {
    TeDataType best_dt = TeUNSIGNEDCHAR;
    double best_min = 0;
    double best_max = 255;
  
    TePDITypes::TePDIRasterVectorType::iterator it = rasters.begin();
    TePDITypes::TePDIRasterVectorType::iterator it_end = rasters.end();
    
    double curr_min = 0;
    double curr_max = 0;
    
    unsigned int bands_number = 0;
    unsigned int curr_band = 0;
    
    while( it != it_end ) {
      bands_number = (*it)->params().nBands();
      
      curr_band = 0;
      
      while( curr_band < bands_number ) {
        if( TeGetRasterMinMaxBounds( *it, curr_band, curr_min, curr_max ) ) {
          if( curr_min == best_min ) {
            if( curr_max > best_max ) {
              best_dt = (*it)->params().dataType_[ curr_band ];
              best_max = curr_max;
              best_min = curr_min;
            }
          } else if( curr_max == best_max ) {
            if( curr_min < best_min ) {
              best_dt = (*it)->params().dataType_[ curr_band ];
              best_max = curr_max;
              best_min = curr_min;
            }
          } else {
            if( ( curr_min < best_min ) || ( curr_max > best_max ) ) {
              best_dt = (*it)->params().dataType_[ curr_band ];
              best_max = curr_max;
              best_min = curr_min;
            }
          }
        }
        
        ++curr_band;
      }
      
      ++it;
    }
    
    return best_dt;
  }
  
  
  void getBandMinMaxValues( const TePDITypes::TePDIRasterPtrType& raster, 
    unsigned int channel, bool progress_enabled, unsigned int sample_step,  
    double& min, double& max )
  {
    TEAGN_TRUE_OR_THROW( raster.isActive(),
      "Inactive Input Raster" );

    TeRasterParams& params = raster->params();

    TEAGN_TRUE_OR_THROW( (int)channel < params.nBands(),
      "Invalid input channel" );  
  
    if ( params.decoderIdentifier_ == "DB" ) {
      
      min = params.vmin_[ channel ];
      max = params.vmax_[ channel ];
    } else {
      min = DBL_MAX;
      max = DBL_MIN;  
      
      const unsigned int in_lines_number = params.nlines_;
      const unsigned int in_columns_number = params.ncols_;
      const bool use_dummy = params.useDummy_;
  
      TePDIPIManager progress( "Guessing channel bounds...", 
        (unsigned int)ceil( ( (double)in_lines_number ) / 
        ( (double)( sample_step + 1 ) ) ), progress_enabled );      
    
      unsigned int rasterline = 0;
      unsigned int rastercolumn = 0;
      double current_raster_level = 0;
      TeRaster& in_raster_ref = *(raster.nakedPointer());
      
      for( rasterline = 0 ; rasterline < in_lines_number ; 
        rasterline += ( 1 + sample_step ) ) {
        
        for( rastercolumn = 0 ; rastercolumn < in_columns_number ; 
          rastercolumn += ( 1 + sample_step ) ) {
  
          if( in_raster_ref.getElement( rastercolumn, rasterline,
              current_raster_level, channel ) ) {
              
            if( current_raster_level > max ) {
              max = current_raster_level;
            }
            if( current_raster_level < min ) {
              min = current_raster_level;
            }
          } else {
            TEAGN_TRUE_OR_THROW( use_dummy,
              "Unable to read from input raster at line=" +
              Te2String( rasterline ) + " column=" +
              Te2String( rastercolumn ) + "channel=" +
              Te2String( channel ) );
          }
        }
        
        progress.Increment();
      }
    }      
  }
  
  
  bool convert2MultiBand( const TePDITypes::TePDIRasterPtrType& in_raster,
    bool progress_enabled, TePDITypes::TePDIRasterPtrType& out_raster )
  {
    TEAGN_TRUE_OR_THROW( in_raster.isActive(), 
      "in_raster pointer inactive" );
    TEAGN_TRUE_OR_THROW( ( in_raster->params().projection() != 0 ), 
      "in_raster do not have a projection" );
      
    /* Checking if we need a conversion */
      
    int in_raster_band = 0;
    bool conversion_required = false;
        
    for( in_raster_band = 0 ; in_raster_band < in_raster->params().nBands() ;
      ++in_raster_band ) {
      
      if( ( in_raster->params().photometric_[ in_raster_band ] !=
        TeRasterParams::TeMultiBand ) &&
        ( in_raster->params().photometric_[ in_raster_band ] !=
        TeRasterParams::TeRGB ) ) {
      
        conversion_required = true;
        in_raster_band = in_raster->params().nBands();
      }
    }
  
    if( conversion_required ) {
      TeRasterParams out_raster_params;
    
      if( out_raster.isActive() ) {
        TEAGN_TRUE_OR_THROW( ( out_raster->params().status_ != 
          TeRasterParams::TeNotReady ), "out_raster not ready" );
          
        out_raster_params = out_raster->params();
      }
      
      if( in_raster->params().photometric_[ 0 ] == 
        TeRasterParams::TePallete ) { 
                 
        out_raster_params.nBands( 3 );
      } else {
        out_raster_params.nBands( in_raster->params().nBands() );
      }
                  
      TeSharedPtr< TeProjection > proj( TeProjectionFactory::make( 
        in_raster->projection()->params() ) );          
      out_raster_params.projection( proj.nakedPointer() );      
      
      out_raster_params.boxResolution( in_raster->params().box().x1(), 
        in_raster->params().box().y1(), in_raster->params().box().x2(), 
        in_raster->params().box().y2(), 
        in_raster->params().resx_, in_raster->params().resy_ );      
      
      out_raster_params.setPhotometric( TeRasterParams::TeMultiBand );
      
      out_raster_params.setDataType( in_raster->params().dataType_[ 0 ], -1 );
      
      if( in_raster->params().useDummy_ ) {
        out_raster_params.setDummy( in_raster->params().dummy_[ 0 ], -1 );
      } else {
        out_raster_params.useDummy_ = false;
      }
        
      if( out_raster.isActive() ) {
        TEAGN_TRUE_OR_RETURN( out_raster->init( out_raster_params ),
          "Output raster init error" );
      } else {
        TEAGN_TRUE_OR_RETURN( TeAllocRAMRaster( out_raster,
          out_raster_params, TePDIUtilsAutoMemPol ),
          "Output raster allocation error" );
      }
          
      /* Creating the transformer instance */
        
      TeRasterTransform transformer( in_raster.nakedPointer(), 
        out_raster.nakedPointer() );
          
      if( in_raster->params().photometric_[ 0 ] == 
        TeRasterParams::TePallete ) {
          
        transformer.setTransfFunction( TeRasterTransform::TePall2Three );
      } else {
        transformer.setTransfFunction( TeRasterTransform::TeBand2Band );
      }
        
      /* Creating the remapper instance */
          
      TeRasterRemap remapper( in_raster.nakedPointer(), 
        out_raster.nakedPointer(), progress_enabled );
        
      remapper.setTransformer( &transformer );
      
      TEAGN_TRUE_OR_RETURN( remapper.apply( progress_enabled),
        "Raster remapping error" );
    } else {
      out_raster = in_raster;
    }
    
    return true;
  }
  
  bool compareGeometryPoints( const TeLinearRing& geom1, 
    const TeLinearRing& geom2 )
  {
    unsigned int geom1_size = geom1.size();
    
    if( geom1_size == geom2.size() ) {
    
      unsigned int index2 = 0;
      bool point_found = false;
      
      for( unsigned int index1 = 0 ; index1 < geom1_size ; ++index1 ) {
        point_found = false;
        
        for( index2 = 0 ; index2 < geom1_size ; ++index2 ) {
          if( geom1[ index1 ] == geom2[ index2 ] ) {
            point_found = true;
            break;    
          }
        }
        
        if( ! point_found ) {
          return false;
        }
      }
      
      return true;
    } else {
      return false;
    }
  }
  

  bool resampleRaster( const TePDITypes::TePDIRasterPtrType& input_raster,
    const TePDITypes::TePDIRasterPtrType& output_raster,
    double x_resolution_ratio, double y_resolution_ratio, 
    TePDIInterpolator::InterpMethod interpol, bool enable_progress ) 
  {
      
    TEAGN_TRUE_OR_RETURN( input_raster.isActive(), "Inactive input raster" );
    TEAGN_TRUE_OR_RETURN( 
      input_raster->params().status_ != TeRasterParams::TeNotReady,
      "Input raster not ready" )
    
    TEAGN_TRUE_OR_RETURN( output_raster.isActive(), 
      "Inactive output raster" );
    TEAGN_TRUE_OR_RETURN( 
      output_raster->params().status_ != TeRasterParams::TeNotReady,
      "Output raster not ready" )
      
    TEAGN_TRUE_OR_RETURN( ( x_resolution_ratio > 0.0 ), "Invalid X ratio" )
    TEAGN_TRUE_OR_RETURN( ( y_resolution_ratio > 0.0 ), "Invalid X ratio" )
    
    /* Reseting the output raster to the new geometry */
    
    TeRasterParams& input_raster_params = input_raster->params();
    
    TeRasterParams new_output_raster_params = output_raster->params();
    
    new_output_raster_params.nBands( input_raster_params.nBands() );
    new_output_raster_params.setPhotometric( 
      input_raster_params.photometric_[ 0 ], -1 );
    if( input_raster_params.projection() != 0 ) {
      new_output_raster_params.projection( input_raster_params.projection() );
    }
    new_output_raster_params.boxResolution( 
      input_raster_params.box().x1(), 
      input_raster_params.box().y1(), input_raster_params.box().x2(), 
      input_raster_params.box().y2(), 
      ( input_raster_params.resx_ * x_resolution_ratio ), 
      ( input_raster_params.resy_ * y_resolution_ratio ) );
      
    TEAGN_TRUE_OR_RETURN( output_raster->init( new_output_raster_params ),
      "Output raster reset error" )
      
    /* interpolating pixel values */
    
    const TeRasterParams& output_raster_params = output_raster->params();
    
    const unsigned int out_lines = output_raster_params.nlines_;
    const unsigned int out_cols = output_raster_params.ncols_;
    const unsigned int out_bands = output_raster_params.nBands();
    
    double in_col = 0;
    double in_line = 0;
    unsigned int out_col = 0;
    unsigned int out_line = 0;
    unsigned int out_band = 0;
    double value = 0;
    
    double dummy_value = 0;
    if( input_raster_params.useDummy_ ) {
      dummy_value = input_raster_params.dummy_[ 0 ];
    }
    
    TeRaster& output_raster_ref = *( output_raster.nakedPointer() );
    
    TePDIInterpolator interpolator;
    TEAGN_TRUE_OR_RETURN( interpolator.reset( input_raster, interpol,
      dummy_value), "Interpolator reset error" )
      
    TePDIPIManager progress( "Ressampling raster", ( out_bands * out_lines ),
      enable_progress );
    
    for( out_band = 0 ; out_band < out_bands ; ++out_band ) {
      for( out_line = 0 ; out_line < out_lines ; ++out_line ) {
        in_line = ( (double)out_line) * y_resolution_ratio;
      
        for( out_col = 0 ; out_col < out_cols ; ++out_col ) {
          in_col = ( (double)out_col) * x_resolution_ratio;
        
          interpolator.interpolate( in_line, in_col, out_band, value );
      
          TEAGN_TRUE_OR_RETURN( output_raster_ref.setElement( out_col, 
            out_line, value, out_band ), "Output raster writing error" )
        }
        
        TEAGN_FALSE_OR_RETURN( progress.Increment(), "Canceled by user" );
      }
    }
    
    return true;
  }
  
} // namespace TeUtils

