#ifndef TETILEINDEXER_H
#define TETILEINDEXER_H

#include <TeGeometry.h>
#include <TeException.h>

#include <algorithm>

#include <math.h>

#include <cassert>

/**
 * @class Polygon tile indexing class for optmized geometrical relational tests.
 * @author Gilberto Ribeiro <gribeiro@dpi.inpe.br>
 * @note The related polygon instance must always be valid ( only the polygon
 * reference is stored internally ).
 */
class TeTileIndexer
{
  public :
  
    /**
     * @typedef Indexed elements node type( pair< ring index, seg index  ).
     */
    typedef vector< pair< unsigned int, unsigned int > > TeTileSegIndex;

  protected:
    
    //!< Reference polygon.
    const TePolygon& refPol_;
    
    
    //!< Tile resolution along "y" axis.
    double dy_;
    
    /** The number of tile lines */
    unsigned int numTileLines_;

    //!< Each tile segments index vector.
    std::vector<TeTileSegIndex*> tileIndex_;
    
    TeTileIndexer( const TeTileIndexer& );
    
    /** Copy overload */
    const TeTileIndexer& operator=( const TeTileIndexer& );
    
    /*! Gets tile index intervals in y direction for a given segment (it does 
        NOT need to be oriented).
      \param c1         First segment coordinate.
      \param c2         Second segment coordinate.
      \param firstTile  The first tile index that this segment intersects.
      \param lastTile   The last tile index that this segment intersects.
     */
    void getTileIndex(const TeCoord2D& c1, const TeCoord2D& c2, 
      unsigned int& firstTile, unsigned int& lastTile) const
    {
      double lowerY = ((c1.y_ < c2.y_) ? c1.y_ : c2.y_);
      double upperY = ((c1.y_ > c2.y_) ? c1.y_ : c2.y_);

      firstTile = int((lowerY - refPol_.box().y1()) / dy_);
      lastTile  = int((upperY - refPol_.box().y1()) / dy_);  
    }    
    
    //! Gets tile index for y coordinate value.
    /*!
      \param y         Value of "y" coordinate.
      \param tileIndex Index of corresponding tile.
     */
    void getTileIndex(const double& y, unsigned int& tileIndex) const
    {
      tileIndex = int( (y - refPol_.box().y1() ) / dy_ );
    }    
    
  public:
  
    //! Alternative Constructor.
    /*!
      \param pol The polygon to index.
      \param dy Tile size along "y" axis.
     */
    TeTileIndexer( const TePolygon& pol, const double& dy )
      : refPol_( pol )
    {
      if( dy <= 0 ) {
        throw( TeException( UNKNOWN_ERROR_TYPE, "Invalid tile index dy" ) );
      }
      
      dy_ = dy;
      
      update();
    }    

    ~TeTileIndexer()
    {
      vector<TeTileSegIndex*>::iterator it = tileIndex_.begin();
      vector<TeTileSegIndex*>::iterator it_end = tileIndex_.end();
      
      while( it != it_end ) {
        delete ( *it );
        
        ++it;
      }
    }
    
    //! Update the tile index with the information of the supplied ring.
    void update( const unsigned int& ring_index )
    {
      assert( ring_index < refPol_.size() );
      
      unsigned int numSegments = 0;

      unsigned int num_points = refPol_[ring_index].size();
        
      if( num_points > 0 ) {
        numSegments = num_points - 1;
  
        for(unsigned int j = 0; j < numSegments; ++j)
        {
          // creates a pointer to the segment
          pair<unsigned int, unsigned int> segPointer(ring_index, j);
  
          // finds the tiles that this segments intersects
          unsigned int firstTileIndex = 0;
          unsigned int lastTileIndex = 0;
            
          getTileIndex( refPol_[ring_index][j], refPol_[ring_index][j + 1], 
            firstTileIndex, lastTileIndex);
              
          assert( firstTileIndex < numTileLines_ );
          assert( lastTileIndex < numTileLines_ );
  
          // associates the pointer segment to the tiles
          for(unsigned int k = firstTileIndex; k <= lastTileIndex; ++k)
          {
            tileIndex_[k]->push_back(segPointer);
          }
        }
      }
    }     

    //! Update the tile index using all polygon rings.
    void update()
    {
      /* Erasing old data */
      vector<TeTileSegIndex*>::iterator it = tileIndex_.begin();
      vector<TeTileSegIndex*>::iterator it_end = tileIndex_.end();
      
      while( it != it_end ) {
        delete *it;
        ++it;
      }    
    
      tileIndex_.clear();
      
      /* Building new index */
      
      numTileLines_ = 1 + 
        ( (unsigned int)floor( refPol_.box().height() / dy_ ) );
            
      for(unsigned int i = 0; i < numTileLines_; ++i)
      {
        tileIndex_.push_back( new TeTileSegIndex );
      }

      const unsigned int rings_number = refPol_.size();

      // for each polygon ring, we need to index its segments
      for(unsigned int i = 0; i < rings_number; ++i)
      {
        update( i );
      }
    }
    
    /*! Gets tile index.
      \param y Y value.
      \param index Output tile pointer ( NULL if not found ).
    */    
    void getTile( const double& y, TeTileSegIndex** index ) const
    {
      unsigned int tidx;
      getTileIndex( y, tidx );
      
      if( tidx < tileIndex_.size() ) {
        ( *index ) = tileIndex_[ tidx ];
      } else {
        ( *index ) = 0;
      }
    }
    
    const TePolygon& getPol() const
    {
      return refPol_;
    }

};


#endif
