#include <TeFileTheme.h>
#include <TeDatabase.h>
#include <TeSTElementSet.h>
#include <TeGeoDataDriver.h>

bool 
TeFileTheme::createFileThemeTable(TeDatabase* db)
{
	if (!db)
		return false;

	if(db->tableExist("te_file_theme"))
		return true;

	TeAttributeList attList;

	TeAttribute att1;
	att1.rep_.name_ = "theme_id";
	att1.rep_.isAutoNumber_ = false;
	att1.rep_.isPrimaryKey_ = true;
	att1.rep_.null_ = false;
	att1.rep_.type_ = TeINT;
	att1.rep_.numChar_ = 0;
	attList.push_back(att1);

	TeAttribute att2;
	att2.rep_.name_ = "file_name";
	att2.rep_.isAutoNumber_ = false;
	att2.rep_.isPrimaryKey_ = false;
	att2.rep_.null_ = false;
	att2.rep_.type_ = TeSTRING;
	att2.rep_.numChar_ = 255;
	attList.push_back(att2);

	TeAttribute att3;
	att3.rep_.name_ = "geom_rep";
	att3.rep_.isAutoNumber_ = false;
	att3.rep_.isPrimaryKey_ = false;
	att3.rep_.null_ = false;
	att3.rep_.type_ = TeINT;
	att3.rep_.numChar_ = 0;
	attList.push_back(att3);

	TeAttribute att4;
	att4.rep_.name_ = "proj_id";
	att4.rep_.isAutoNumber_ = false;
	att4.rep_.isPrimaryKey_ = false;
	att4.rep_.null_ = false;
	att4.rep_.type_ = TeINT;
	att4.rep_.numChar_ = 0;
	attList.push_back(att4);

	if (!db->createTable("te_file_theme", attList)||
		!db->createRelation("fk_file_theme_id", "te_file_theme", "theme_id", "te_theme", "theme_id", true) ||
		!db->createRelation("fk_file_theme_proj_id", "te_file_theme", "proj_id", "te_projection", "projection_id", true))
		return false;
	return 1;
}



TeFileTheme::TeFileTheme(const string& name, TeViewNode* parent, int view, int id):
		TeAbstractTheme(name, parent, view, id, TeFILETHEME),
		data_(0), 
		fileName_(""), 
		localDb_(0),
		numObjects_(0),
		projection_(0),
		reps_(TeGEOMETRYNONE),
		dataDriver_(0)
{}

TeFileTheme::TeFileTheme(const std::string& themeName, const std::string& fileName, TeDatabase* localDb, TeViewNode* parent, int view):
		TeAbstractTheme(themeName, parent, view, 0, TeFILETHEME), 
		data_(0),
		fileName_(fileName), 
		localDb_(localDb),
		numObjects_(0),
		projection_(0),
		reps_(TeGEOMETRYNONE),
		dataDriver_(0)
{}

TeFileTheme::~TeFileTheme()
{
	if(data_)
		delete data_;
	data_ = 0;
	dataDriver_ = 0; // we do not delete the pointer??? Shall we?
}

TeFileTheme& 
TeFileTheme::operator=(TeFileTheme& rhs)
{
	if ( this != &rhs )
	{
		*(TeAbstractTheme*)this = rhs;
		fileName_ = rhs.fileName_;
		localDb_ = rhs.localDb_;

		if(data_)
			delete data_;
		data_ = rhs.data_;  //FUNCATE: fazer a copia desse ponteiro 
	}
	return *this;
}

void
TeFileTheme::setDataDriver(TeGeoDataDriver* dataDriver)
{ 
	dataDriver_ = dataDriver;
	if (dataDriver_)
	{
		TeBox bb;
		dataDriver_->getDataInfo(numObjects_, bb, reps_);
        this->setThemeBox(bb);
		projection_ = dataDriver_->getDataProjection();
	}
}

TeGeoDataDriver* 
TeFileTheme::getDataDriver()
{
	return dataDriver_;
}

TeSTElementSet* 
TeFileTheme::getData()
{
	return data_;
}

void 
TeFileTheme::setFileName(const std::string& fileName)
{
	fileName_ = fileName;
}

std::string 
TeFileTheme::getFileName()
{
	return fileName_;
}

void 
TeFileTheme::setLocalDatabase(TeDatabase* db)
{	
	this->localDb_ = db; 
}

TeDatabase* 
TeFileTheme::getLocalDatabase()
{ 
	return localDb_; 
}

void 
TeFileTheme::clearData() 
{ 
	data_->clear(); 
}

unsigned int 
TeFileTheme::getNumberOfObjects()
{	
	return numObjects_; 
}

bool
TeFileTheme::loadDataFromFile()
{
    if(!dataDriver_)
		return false;

	data_ = new TeSTElementSet();

	if (!dataDriver_->loadData(data_))
	{
		data_ = 0;
		return false;
	}
	return true;
}

bool
TeFileTheme::load()
{
	if(!localDb_)
		return false;

	//second: load specific theme information from te_file_theme table
	TeDatabasePortal* portal = localDb_->getPortal();
	if(!portal)
		return false;

	//table: theme_id, file_name, geom_rep
	std::string sql = " SELECT te_file_theme.*, ";	// 0 - 3
	sql += " te_projection.* ";						// 4 - 
	sql += " FROM te_file_theme, te_projection ";
	sql += " WHERE te_file_theme.proj_id = te_projection.projection_id ";
	sql += " AND theme_id = "+ Te2String(id());

	if(!portal->query(sql) || !portal->fetchRow())
	{
		delete portal;
		return false;
	}

	fileName_ = portal->getData(1);
	visibleRep(portal->getInt(2));

	FILE* fp = fopen(fileName_.c_str(), "r");
	if(!fp)
		return false;
	else
		fclose(fp);

	TeProjection* proj = 0;
	portal->getProjection(&proj, 4);
	return true;
}

bool 
TeFileTheme::save()
{
	if(!localDb_)
		return false;

	//insert theme in database 
	if(id()==0)
	{
		if(!localDb_->insertTheme(this)) //updateThemeTable
		{
			localDb_->deleteTheme(this->id());
			return false;
		}
	}

	if (!projection_)
		projection_ = new TeNoProjection();
	if (projection_->id() <= 0)
		localDb_->insertProjection(projection_);

	string strSQL;
	strSQL  = "INSERT INTO te_file_theme (theme_id, file_name, geom_rep, proj_id) VALUES (";
	strSQL += Te2String(id());
	strSQL += ", '";
	strSQL += fileName_;
	strSQL += "', ";
	strSQL += Te2String(visibleRep());
	strSQL += ", ";
	strSQL += Te2String(projection_->id());
	strSQL += ")";

	if(!localDb_->execute(strSQL))
	{
		localDb_->deleteTheme(this->id());
		return false;
	}
	return true;
}

TeProjection* 
TeFileTheme::getThemeProjection()
{	
	return projection_; 
}

void 
TeFileTheme::setThemeProjection(TeProjection* proj)
{ 
	projection_ = proj; 
}

TeGeomRep 
TeFileTheme::getGeomReps()
{
	return reps_;
}

bool 
TeFileTheme::buildGrouping(const TeGrouping& , TeSelectedObjects , vector<double>* )
{
	return true;
}

bool 
TeFileTheme::buildGrouping(const TeGrouping& , TeChronon , vector<map<string, string> >& )
{
	return true;
}

bool 
TeFileTheme::saveGrouping(TeSelectedObjects )
{
	return true;
}
        

void 
TeFileTheme::setLegendsForObjects()
{
}

bool 
TeFileTheme::deleteGrouping()
{
	return true;
}

bool 
TeFileTheme::locatePolygon(TeCoord2D &pt, TePolygon &polygon, const double& tol)
{
    TePolygonSet aux;
	if(!locatePolygonSet(pt, tol, aux))
		return false;
	if(aux.empty())
		return false;
	
	polygon = aux[0];
	return true;	
}

bool 
TeFileTheme::locatePolygonSet(TeCoord2D& /*pt*/, double /*tol*/, TePolygonSet& /*polygons*/)
{
	if(!data_)
		return false;
//	if(!data_->locateGeometry(pt, polygons, tol))
//		return false;
	return true;
}

bool 
TeFileTheme::locateLine(TeCoord2D& /*pt*/, TeLine2D& line, const double& /*tol*/)
{
	TeLineSet aux;
	if(!data_)
		return false;
//	if(!data_->locateGeometry(pt, aux, tol))
//		return false;
	if(aux.empty())
		return false;
	line = aux[0];
	return true;
}

bool 
TeFileTheme::locatePoint(TeCoord2D& /*pt*/, TePoint &point, const double& /*tol*/)
{
	TePointSet aux;
	if(!data_)
		return false;
//	if(!data_->locateGeometry(pt, aux, tol))
//		return false;
	if(aux.empty())
		return false;
	point = aux[0];
	return true;
}

bool 
TeFileTheme::locateCell(TeCoord2D&, TeCell&, const double&)
{
	return false;
}


std::set<std::string> 
TeFileTheme::getObjects(TeSelectedObjects )
{
	std::set<std::string> set;
	return set;
}


std::set<std::string> 
TeFileTheme::getObjects(const vector<string>& )
{
	std::set<std::string> set;
	return set;
}


std::vector<std::string>  
TeFileTheme::getItemVector(TeSelectedObjects )
{
	std::vector<std::string> set;
	return set;
}


std::vector<std::string>  
TeFileTheme::getItemVector(const set<string>& )
{
	std::vector<std::string> set;
	return set;
}

void
TeFileTheme::setOwnLegendsForObjects()
{

}