/*
TerraLib - a library for developing GIS applications.
Copyright  2001, 2002, 2003 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular
purpose. The library provided hereunder is on an "as is" basis, and the
authors have no obligation to provide maintenance, support, updates,
enhancements, or modifications.
In no event shall INPE be held liable to any party
for direct, indirect, special, incidental, or consequential damages arising
out of the use of this library and its documentation.
*/

#include "TeGeometricTransformation.h"

#include <TeAgnostic.h>

#include <math.h>
#include <float.h>

#include <iostream>


TeGeometricTransformation::TeGeometricTransformation()
{
}


TeGeometricTransformation::~TeGeometricTransformation()
{
}


void TeGeometricTransformation::getParameters( TeGTParams& params ) const
{
  params = internal_params_;
}


bool TeGeometricTransformation::reset( const TeGTParams& newparams )
{
  /* If previous calculated parameters were supplied, no need to do calcules */

  if( ( newparams.direct_parameters_.Nrow() > 0 ) && 
    ( newparams.direct_parameters_.Ncol() > 0 ) &&
    ( newparams.inverse_parameters_.Nrow() > 0 ) &&
    ( newparams.inverse_parameters_.Ncol() > 0 ) 
    ) {
    
    internal_params_ = newparams;
    return true;
  } else {
    /* No previous parameters given - Need to calculate the new transformation
       parameters */
       
    if( newparams.max_dmap_error_ < 0 ) {
      throw TeException( UNKNOWN_ERROR_TYPE, 
        "Invalid maximum allowed direct mapping error" );    
      return false;
    }
    
    if( newparams.max_imap_error_ < 0 ) {
      throw TeException( UNKNOWN_ERROR_TYPE, 
        "Invalid maximum allowed inverse mapping error" );    
      return false;
    }    
    
    const unsigned int req_tie_pts_nmb = getMinRequiredTiePoints();

    if( newparams.tiepoints_.size() < req_tie_pts_nmb ) {
      return false;
    } else {
      if( newparams.auto_refine_ ) {
        /* Initiating seed */
        
        std::vector<unsigned int> comb_seed_vec;
        
        for( unsigned int comb_seed_vec_index = 0 ; 
          comb_seed_vec_index < req_tie_pts_nmb ;
          ++comb_seed_vec_index ) {
          
          comb_seed_vec.push_back( 0 );
        }
        
        /* Trying to find the best tie-points by building 
          the transformation with the highest number of
          tie-points, but with an acceptable mapping error */
          
        std::vector< TeCoordPair > best_tie_points;
        
        const unsigned int req_tie_pts_nmb = getMinRequiredTiePoints();
        const unsigned int tiepoints_size = newparams.tiepoints_.size();
        unsigned int tiepoints_index = 0;
        bool point_already_present = false;
//        TeMatrix curr_direct_parameters;
//        TeMatrix curr_inverse_parameters;        
//        std::vector< TeCoordPair > curr_tie_points;
        TeGTParams curr_params = newparams;
        
        while( recombineSeed( comb_seed_vec, 0, tiepoints_size ) ) {
          /* Extracting tie-points from the original vector */
          
          curr_params.tiepoints_.clear();
          
          for( unsigned int comb_seed_vec_index_2 = 0 ; 
            comb_seed_vec_index_2 < req_tie_pts_nmb ;
            ++comb_seed_vec_index_2 ) {
            
            curr_params.tiepoints_.push_back( 
              newparams.tiepoints_[ comb_seed_vec[ comb_seed_vec_index_2 ] - 
              1 ] );
          }
          
          /* Trying to generate a valid transformation */
          
          if( computeParameters( curr_params ) ) {
          
            if( ( newparams.max_dmap_error_ >= getDirectMappingError( 
              curr_params ) ) &&
              ( newparams.max_imap_error_ >= getInverseMappingError( 
              curr_params ) ) ) {
            
              /* Trying to insert more tie-points into current 
                 transformation */
             
              for( tiepoints_index = 0 ; tiepoints_index < tiepoints_size ;
                ++tiepoints_index ) {
                
                /* Verifying if the current tie-point is already present */
                
                point_already_present = false;
                
                for( unsigned int comb_seed_vec_index_3 = 0 ; 
                  comb_seed_vec_index_3 < req_tie_pts_nmb ;
                  ++comb_seed_vec_index_3 ) {
                
                  if( tiepoints_index == ( 
                    comb_seed_vec[ comb_seed_vec_index_3 ] - 1 ) ) {
                  
                    point_already_present = true;
                    break;
                  }
                }
                
                if( ! point_already_present ) {
                  curr_params.tiepoints_.push_back( 
                    newparams.tiepoints_[ tiepoints_index ] );
                  
                  /* Verifying if the new tie-point insertion does not generate 
                    an invalid transformation */
                  
                  if( computeParameters( curr_params ) ) {
                      
                    if( ( newparams.max_dmap_error_ < 
                      getDirectMappingError( curr_params ) ) ||
                      ( newparams.max_imap_error_ < 
                      getInverseMappingError( curr_params ) ) ) {
                      
                      curr_params.tiepoints_.pop_back();
                    }
                  } else {
                    curr_params.tiepoints_.pop_back();
                  }
                }
              }
              
              /* A valid transformation was generated, now verifying if the
                number of tie-poits is greater then the current best 
                transformation
              */            
              
              if( curr_params.tiepoints_.size() > best_tie_points.size() ) {
                best_tie_points = curr_params.tiepoints_;
              }          
            }
          }
        }
        
        if( best_tie_points.size() >= req_tie_pts_nmb ) {
          curr_params.tiepoints_ = best_tie_points;
        
          if( computeParameters( curr_params ) ) {
            if( ( newparams.max_dmap_error_ >= getDirectMappingError( 
              curr_params ) ) &&
              ( newparams.max_imap_error_ >= getInverseMappingError( 
              curr_params ) ) ) {
              
              internal_params_ = curr_params;
              
              return true;
            }
          }
        }
        
        return false;      
      } else {
        /* No auto refinning required */
        
        TeGTParams curr_params = newparams;
        
        if( computeParameters( curr_params  ) ) {
          
          if( ( newparams.max_dmap_error_ >= 
            getDirectMappingError( curr_params ) ) &&
            ( newparams.max_imap_error_ >= 
            getInverseMappingError( curr_params ) ) ) {
          
            internal_params_ = curr_params;
            
            return true;
          } else {
            return false;
          }
        } else {
          return false;
        }
      }
    }
  }
}


double TeGeometricTransformation::getDirectMappingError(  
  const TeGTParams& params ) const
{
  unsigned int tiepoints_size = params.tiepoints_.size();
  
  TEAGN_DEBUG_CONDITION( ( tiepoints_size > 0 ),
    "Invalid tie-points size" )
  
  double max_error = 0;
  double current_error = 0;
  
  for( unsigned int tpindex = 0 ; tpindex < tiepoints_size ; ++tpindex ) {
    current_error = getDirectMappingError( params.tiepoints_[ tpindex ], 
      params );
    
    if( current_error > max_error ) {
      max_error = current_error;
    }
  }
  
  return max_error;
}


double TeGeometricTransformation::getInverseMappingError(  
  const TeGTParams& params ) const
{
  unsigned int tiepoints_size = params.tiepoints_.size();
  
  TEAGN_DEBUG_CONDITION( ( tiepoints_size > 0 ),
    "Invalid tie-points size" )

  
  double max_error = 0;
  double current_error = 0;
  
  for( unsigned int tpindex = 0 ; tpindex < tiepoints_size ; ++tpindex ) {
    current_error = getInverseMappingError( params.tiepoints_[ tpindex ], 
      params );
    
    if( current_error > max_error ) {
      max_error = current_error;
    }
  }
  
  return max_error;
}


double TeGeometricTransformation::getDirectMappingError( 
  const TeCoordPair& tie_point, const TeGTParams& params ) const
{
  TeCoord2D direct_mapped_point;

  directMap( params, tie_point.pt1, direct_mapped_point );
    
  double diff_x = tie_point.pt2.x() - direct_mapped_point.x();
  double diff_y = tie_point.pt2.y() - direct_mapped_point.y();
    
  return hypot( diff_x, diff_y );
}


double TeGeometricTransformation::getInverseMappingError( 
  const TeCoordPair& tie_point, const TeGTParams& params ) const
{
  TeCoord2D inverse_mapped_point;

  inverseMap( params, tie_point.pt2, inverse_mapped_point );
    
  double diff_x = tie_point.pt1.x() - inverse_mapped_point.x();
  double diff_y = tie_point.pt1.y() - inverse_mapped_point.y();
    
  return hypot( diff_x, diff_y );
}


bool TeGeometricTransformation::recombineSeed( std::vector<unsigned int>& seed, 
  const unsigned int& seedpos, const unsigned int& elements_nmb ) const
{
  unsigned int seed_size = seed.size();
  
  if( seedpos >= seed_size ) {
    return false;
  }

  if( seed[ seedpos ]  >= ( elements_nmb - seed_size + seedpos + 1 ) ) {
    if( seedpos == seed_size - 1 ) {
      return false;
    } else if( seedpos == 0 ) {
      return recombineSeed( seed, seedpos + 1, elements_nmb ) ;
    } else {
      return recombineSeed( seed, seedpos + 1, elements_nmb ) ;
    };
  } else if( seed[ seedpos ]  == 0 ) {
    if( seedpos == 0 ) {
      seed[ seedpos ] = 1 ;
      return recombineSeed( seed, seedpos + 1, elements_nmb );
    } else if( seedpos == seed_size - 1 ) {
      seed[ seedpos ] = seed[ seedpos - 1 ] + 1;
      return true;
    } else {
      seed[ seedpos ] = seed[ seedpos - 1 ] + 1;
      seed[ seedpos + 1 ] = 0;
      return recombineSeed( seed, seedpos + 1, elements_nmb );
    }
  } else {
    if( seedpos == seed_size - 1 ) {
      seed[ seedpos ] = seed[ seedpos ] + 1;
      return true;
    } else if( seedpos == 0 ) {
      if( recombineSeed( seed, seedpos + 1, elements_nmb ) ) {
        return true;
      } else {
        seed[ seedpos ] = seed[ seedpos ] + 1;
        seed[ seedpos + 1 ] = 0;
        return recombineSeed( seed, seedpos + 1, elements_nmb );
      }
    } else {
      if( recombineSeed( seed, seedpos + 1, elements_nmb ) ) {
        return true;
      } else {
        seed[ seedpos ] = seed[ seedpos ] + 1;
        seed[ seedpos + 1 ] = 0;
        return recombineSeed( seed, seedpos + 1, elements_nmb );
      }
    }
  }
}


TeGeometricTransformation* TeGeometricTransformation::DefaultObject( 
  const TeGTParams& )
{ 
  throw TeException( UNKNOWN_ERROR_TYPE, 
      "Trying to create an invalid TeGemetricTransformation instance" );
  
  return 0;
}; 


