/************************************************************************************
TerraLib - a library for developing GIS applications.
Copyright  2001-2007 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular purpose.
The library provided hereunder is on an "as is" basis, and the authors have no
obligation to provide maintenance, support, updates, enhancements, or modifications.
In no event shall INPE and Tecgraf / PUC-Rio be held liable to any party for direct,
indirect, special, incidental, or consequential damages arising out of the use
of this library and its documentation.
*************************************************************************************/

#include "TeMappedMemory.h"
#include "TeException.h"
#include "TeErrorLog.h"
#include "TeUtils.h"
#include "TeAgnostic.h"

#if defined WIN32
  #include <windows.h>
#elif defined __unix__
  #include <sys/types.h>
  #include <sys/stat.h>
  #include <fcntl.h>
  #include <unistd.h>
#else
  #error "Unsupported platform"
#endif


void TeMappedMemory::init()
{
  #if defined WIN32
    m_hFile_ = 0;
    m_hMapping_ = 0;
    m_lpszFile_ = 0;
  #elif defined __unix__
    m_hFile_ = 0;
    m_lpszFile_ = 0;
  #else
    #error "Unsupported platform"
  #endif
  
  disk_file_name_.clear();
  mapping_is_active_ = false;
  curr_size_ = 0;
  must_delete_file_ = false;
};


TeMappedMemory::TeMappedMemory()
{
  init();
}


TeMappedMemory::~TeMappedMemory()
{
  reset();
}


void TeMappedMemory::reset()
{
  TEAGN_TRUE_OR_THROW( toggle( false ), "Error disabling mapped memory" );
  
  if( must_delete_file_ ) {
    remove( disk_file_name_.c_str() );
  }
  
  init();
}


bool TeMappedMemory::reset( unsigned long int size, bool enabled )
{
  reset();
  
  std::string disk_file_name;
      
  TEAGN_TRUE_OR_RETURN( TeGetTempFileName( disk_file_name ),
    "Unable to get temporary file name" );
    
  return( reset( disk_file_name, size, false, enabled ) );
}


bool TeMappedMemory::reset( const std::string& filename, 
  unsigned long int size, bool keep_disk_file, bool enabled )
{
  reset();

  if( filename.empty() || ( size == 0 ) ) {
    TeErrorLog::instance().insert( UNKNOWN_ERROR_TYPE,
      "Invalid mapped file name or invalid file size" ); 
            
    return false;
  } else {
    TEAGN_TRUE_OR_RETURN( createNewDiskFile( filename, size ),
      "Unable to create memory mapped disk file" );
      
    disk_file_name_ = filename;
    must_delete_file_ = ( ! keep_disk_file );
    
    if( toggle( enabled ) ) {
      return true;
    } else {
      reset();
      return false;
    }
  }
}


bool TeMappedMemory::reset( const std::string& filename, bool enabled )
{
  reset();
  
  disk_file_name_ = filename;
  
  if( toggle( enabled ) ) {
    return true;
  } else {
    reset();
    return false;
  }  
}


bool TeMappedMemory::toggle( bool enabled )
{
  if( enabled  ) {
    if( mapping_is_active_ ) {
      return true;    
    } else {
      /* Activating mapping */    
      
      if( ! TeCheckFileExistence( disk_file_name_.c_str() ) ) {
        return false;
      }
      
      unsigned long int filesize = TeGetFileSize( disk_file_name_ );
  
      if( filesize > 0 ) {
        #if defined WIN32
          DWORD dwCreationDisposition = OPEN_EXISTING;
          DWORD dwDesiredAccess = GENERIC_READ | GENERIC_WRITE;         
          DWORD flProtect = PAGE_READWRITE;
          DWORD dwDesiredAccessV = FILE_MAP_WRITE; 
      
          HANDLE m_hFile = CreateFileA(
            disk_file_name_.c_str(), dwDesiredAccess, 
            FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, dwCreationDisposition,
            0, NULL);
        
          if( m_hFile == INVALID_HANDLE_VALUE ) {
            TeErrorLog::instance().insert( UNKNOWN_ERROR_TYPE,
              "Temporary file creation error" );        
       
            return false;
          }
        
          DWORD m_dwSize = GetFileSize(m_hFile, NULL);
      
          if( m_dwSize == 0xffffffff ) {
            TeErrorLog::instance().insert( UNKNOWN_ERROR_TYPE,
              "Invalid file size" );        

            CloseHandle(m_hFile);
        
            return false;
          }        
      
          HANDLE m_hMapping = CreateFileMapping( m_hFile, NULL, flProtect,
            0, 0, NULL );
              
          if( m_hMapping == NULL ) {
            CloseHandle(m_hFile);
        
            TeErrorLog::instance().insert( UNKNOWN_ERROR_TYPE,
              "Mapping creation error" );
              
            return false;        
          }
        
          LPVOID m_lpszFile = (LPVOID) MapViewOfFile( m_hMapping, 
            dwDesiredAccessV, 0, 0, 0);
      
          if( m_lpszFile == 0 ) {
            CloseHandle( m_hMapping );
            CloseHandle( m_hFile );
        
            TeErrorLog::instance().insert( UNKNOWN_ERROR_TYPE,
              "Mapping creation error" );
            
            return false;        
          }        
      
          m_hFile_ = m_hFile;
          m_hMapping_ = m_hMapping;
          m_lpszFile_ = m_lpszFile;
        #elif defined __unix__
          int m_hFile = open( disk_file_name_.c_str(), O_RDWR );

          if( m_hFile == -1 ) {
            TeErrorLog::instance().insert( UNKNOWN_ERROR_TYPE,
              "Temporary file creation error" );
            
            return false;         
          }
     
          void* m_lpszFile  =  mmap(0, (long)filesize, (PROT_READ | PROT_WRITE), 
            MAP_SHARED, m_hFile, 0);
      
          if( m_lpszFile == ((void*)-1) ) {
            close( m_hFile );
        
            TeErrorLog::instance().insert( UNKNOWN_ERROR_TYPE,
              "Mapping creation error" );
            
            return false;            
          }
        
          m_hFile_ = m_hFile;
          m_lpszFile_ = m_lpszFile;
        #else
          #error "Unsupported platform"
        #endif
    
        mapping_is_active_ = true;
        curr_size_ = filesize;
      }      
      
      return true;
    }
  } else {
    if( mapping_is_active_ ) {
      /* Disabling mapping */
      
      #if defined WIN32
        if( ! UnmapViewOfFile( m_lpszFile_ ) ) {
          throw TeException( UNKNOWN_ERROR_TYPE, "Unable to unmap disk file",
            false );
        }
          
        CloseHandle( m_hMapping_ );
        CloseHandle( m_hFile_ );
      #elif defined __unix__
        if( munmap( m_lpszFile_ , (long)curr_size_ ) == -1) {
          throw TeException( UNKNOWN_ERROR_TYPE, "Unable to unmap disk file",
            false );
        }
            
        close(m_hFile_);
      #else
        #error "Unsupported platform"
      #endif
      
      mapping_is_active_ = false;
      curr_size_ = 0;
      
      return true;
    } else {
      return true;
    }
  }
}


void* TeMappedMemory::getPointer() const
{
  if( mapping_is_active_ ) {
    return m_lpszFile_;
  } else {
    throw TeException( UNKNOWN_ERROR_TYPE, 
      "Trying to access an inactive mapping", false );
  
    return 0;
  }
}


std::string TeMappedMemory::getFileName() const
{
  if( mapping_is_active_ ) {
    return disk_file_name_;
    
  } else {
    TEAGN_LOG_AND_THROW( "Trying to access an inactive mapping" );
      
    return std::string( "dummy string ");
  }
}


bool TeMappedMemory::createNewDiskFile( const std::string& filename,
  unsigned long int size ) const
{
  #if defined WIN32
    DWORD dwCreationDisposition = OPEN_ALWAYS;
    DWORD dwDesiredAccess = GENERIC_READ | GENERIC_WRITE;         
    DWORD flProtect = PAGE_READWRITE;
    DWORD dwDesiredAccessV = FILE_MAP_WRITE; 
        
    HANDLE m_hFile = CreateFileA(
      filename.c_str(), dwDesiredAccess, 
      FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, dwCreationDisposition,
      0, NULL);
          
    if( m_hFile == INVALID_HANDLE_VALUE ) {
      TeErrorLog::instance().insert( UNKNOWN_ERROR_TYPE,
        "Mapped file creation error" );        
          
      return false;
    }
          
    DWORD m_dwSize = GetFileSize(m_hFile, NULL);
        
    if( m_dwSize == 0xffffffff ) {
      TeErrorLog::instance().insert( UNKNOWN_ERROR_TYPE,
        "Invalid file size" );        
        
      CloseHandle(m_hFile);
          
      return false;
    }        
        
    if ( m_dwSize < size ) {
      char lpBuffer[1024];
      int nNumberOfBytesToWrite = size - m_dwSize;
      unsigned long nNumberOfBytesWritten;
          
      for (;nNumberOfBytesToWrite > 1024; nNumberOfBytesToWrite-=1024) {
        if ( ! WriteFile( m_hFile, lpBuffer, 1024, &nNumberOfBytesWritten, 
            NULL ) ) {
                 
          CloseHandle(m_hFile);
             
          TeErrorLog::instance().insert( UNKNOWN_ERROR_TYPE,
            "Unable to write to temporary file" );
                
          return false;
        }
      }
          
      if ( ! WriteFile( m_hFile, lpBuffer, nNumberOfBytesToWrite, 
        &nNumberOfBytesWritten, NULL ) ) {
               
        CloseHandle(m_hFile);
            
        TeErrorLog::instance().insert( UNKNOWN_ERROR_TYPE,
          "Unable to write to temporary file" );
              
        return false;
      }
    }
  
    CloseHandle(m_hFile);
  #elif defined __unix__
    int m_hFile = open( filename.c_str(),O_RDWR|O_CREAT,S_IRWXU);

    if( m_hFile == -1 ) {
      TeErrorLog::instance().insert( UNKNOWN_ERROR_TYPE,
        "Temporary file creation error" );
              
      return false;         
    }
    
    off_t seek_off = ( off_t )( size - 1 );
          
    if( lseek(m_hFile, seek_off, SEEK_SET) == -1 ) {
      close( m_hFile );
          
      TeErrorLog::instance().insert( UNKNOWN_ERROR_TYPE,
        "Temporary file seek error" );
              
      return false;           
    }
        
    unsigned char c = '\0';
    

    if( write( m_hFile, (void*)&c, sizeof( unsigned char ) ) == -1 ) {
      close( m_hFile );
          
      TeErrorLog::instance().insert( UNKNOWN_ERROR_TYPE,
        "Temporary file write error" );
              
      return false;            
    }  
    
    close( m_hFile );
  #else
    #error "Unsupported platform"
  #endif
  
  return true;
}


unsigned long int TeMappedMemory::size() const
{
  return curr_size_;
}
