/************************************************************************************
TerraLib - a library for developing GIS applications.
Copyright � 2001-2007 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular purpose.
The library provided hereunder is on an "as is" basis, and the authors have no
obligation to provide maintenance, support, updates, enhancements, or modifications.
In no event shall INPE and Tecgraf / PUC-Rio be held liable to any party for direct,
indirect, special, incidental, or consequential damages arising out of the use
of this library and its documentation.
*************************************************************************************/

/*! \file TeMultiContainer.h
   \brief A container class to store multiple types os object copies.
*/


#ifndef TEMULTICONTAINER_H
  #define TEMULTICONTAINER_H

  #include "TeSharedPtr.h"
  #include "TeThreadLock.h"
  
  #include <time.h>

  #include <vector>

  /**
    * @brief Multi-container node interface.
    * @author Emiliano F. Castejon <castejon@dpi.inpe.br>
    * @ingroup Utils
    */    
  class TeMCNodeInterface {
    public :
    
      /**
        * @brief Default Constructor.
        */        
      TeMCNodeInterface() {};
      
      /**
        * @brief Default Destructor.
        */        
      virtual ~TeMCNodeInterface() {};
      
      /**
        * @brief Copies the the current node instance by calling
        * the operator= from the contained object instance.
        * @return A pointer to the new node copy.
        */            
      virtual TeSharedPtr< TeMCNodeInterface > clone() const = 0;
      
    private :
    
      /**
      * @brief operator= overload.
      *
      * @param ext_instance External instance reference.
      * @return The external instance reference.
      */
      const TeMCNodeInterface& operator=( 
        const TeMCNodeInterface& ext_instance );        
  };


  /**
    * @brief Multi-container node class.
    * @author Emiliano F. Castejon <castejon@dpi.inpe.br>
    * @ingroup Utils
    */       
  template< typename ObjectT >
  class TeMCNode : public TeMCNodeInterface {
    public :
    
      /**
        * @brief Default Constructor.
        */        
      TeMCNode() {};
      
      /**
        * @brief Default Destructor.
        */        
      ~TeMCNode() {};
      
      /**
        * @brief Copies the the current node instance by calling
        * the operator= from the contained object instance.
        * @return A pointer to the new node copy.
        */              
      TeSharedPtr< TeMCNodeInterface > clone() const
      {
        TeMCNode< ObjectT >* new_node_ptr =
          new TeMCNode< ObjectT >;
          
        TeSharedPtr< ObjectT > new_obj_ptr( new ObjectT );
        ( *new_obj_ptr )= ( *obj_ptr_ );
        
        new_node_ptr->setObjPtr( new_obj_ptr );
        
        TeSharedPtr< TeMCNodeInterface > new_nodei_ptr(
          new_node_ptr );
        
        return new_nodei_ptr;
      };
      
      /**
        * @brief Set the internal object pointer.
        * @param ptr Object pointer.
        */        
      void setObjPtr( const TeSharedPtr< ObjectT >& ptr ) 
      { 
        obj_ptr_ = ptr;
      };
      
      /**
        * @brief Get the internal object pointer.
        * @return The object pointer.
        */        
      TeSharedPtr< ObjectT > getObjPtr() const
      { 
        return obj_ptr_; 
      };
      
    protected :
      
      /**
        * @brief The internal object pointer.
        */
      TeSharedPtr< ObjectT > obj_ptr_;

  }; 

 /**
  * @brief A container class to store multiple types os object copies.
  * @author Emiliano F. Castejon <castejon@dpi.inpe.br>
  * @note This is a thread-safe class.
  * @ingroup Utils
  */
  template< typename TeMultiContainerKeyT >
  class TeMultiContainer {
  
    public :
    
      /** @typedef TeSharedPtr< TeMultiContainer > pointer 
          Type definition for a instance pointer */
      typedef TeSharedPtr< TeMultiContainer > pointer;
      
      /** @typedef const TeSharedPtr< TeMultiContainer > const_pointer 
          Type definition for a const instance pointer */
      typedef const TeSharedPtr< TeMultiContainer > const_pointer;
      
      /**
       * @brief Default Constructor
       */
      TeMultiContainer();
      
      /**
       * @brief Alternative Constructor
       * @param external External reference.
       */
      TeMultiContainer( 
        const TeMultiContainer< TeMultiContainerKeyT >& external );      

      /**
       * @brief  Default Destructor
       */
      ~TeMultiContainer();
                           
      /**
       * @brief Operator == overload.
       *
       * @param ext_instance External instance reference.
       * @return true if this instance have the same internal
       * update time that the external instance.
       */
      bool operator==( 
        const TeMultiContainer< TeMultiContainerKeyT >& ext_instance ) const;
      
      /**
       * @brief Operator != overload.
       *
       * @param ext_instance External instance reference.
       * @return true if this instance don't have the same internal
       * update time that the external instance.
       */
      bool operator!=( 
        const TeMultiContainer< TeMultiContainerKeyT >& ext_instance ) const;
      
      /**
       * @brief operator= overload.
       *
       * @param ext_instance External instance reference.
       * @return The external instance reference.
       */
      const TeMultiContainer& operator=( 
        const TeMultiContainer< TeMultiContainerKeyT >& ext_instance );
      
      /**
       * @brief Clear all contents.
       *
       */
      void clear();
                         
      /**
       * @brief Store a object copy.
       *
       * @param obj_key Object key.
       * @param obj_reference Object instance.
       */                         
      template< typename ObjectT >
      void store( const TeMultiContainerKeyT& obj_key,
                         const ObjectT& obj_reference );
      
      /**
       * @brief Retrive a copy of a stored object.
       *
       * @param obj_key Object key.
       * @param obj_reference Object instance.
       * @return true if OK, false if the parameter was no found or error.
       */      
      template< typename ObjectT >
      bool retrive( const TeMultiContainerKeyT& obj_key,
        ObjectT& obj_reference ) const;
        
      /**
       * @brief Retrive copies of all stored objects of a defined type.
       *
       * @param objs_vector A vector with copies of all found objects.
       * @note An empty vector will be returned if no object was found.
       */      
      template< typename ObjectT >
      void multiRetrive( std::vector< std::pair< TeMultiContainerKeyT, 
        ObjectT > >& objs_vector ) const;        
     
      /** 
       * @brief Checks if a object is stored.
       * 
       * @param obj_key Object key.
       * @note The object type must be specified.
       * @return true if OK, false if the object is not stored.
       */
      template< typename ObjectT >
      bool isStored( const TeMultiContainerKeyT& obj_key ) const;
      
    protected :
        
      /** @typedef typename std::pair< TeMultiContainerKeyT, TeSharedPtr< TeMCNodeInterface > > VectorKeyT
        Internal vector key type definition. */
      typedef typename std::pair< TeMultiContainerKeyT, 
        TeSharedPtr< TeMCNodeInterface > > VectorKeyT;

      /** @typedef typename std::vector< VectorKeyT > VectorT
        Internal vector type definition. */
      typedef typename std::vector< VectorKeyT > VectorT;
      
      /** @brief The internal vector instance. */
      VectorT objects_vector_;    
      
      /**
       * @brief The last update time.
       */
      time_t last_up_time_;
      
      /**
       * @brief This instance locking mutex.
       */    
      mutable TeThreadLock this_lock_instance_;      

      /**
       * @brief Updates the internal last update time.
       *
       * @note Needed by comparison between two multi containers.
       */
      void update_time();            

};


template< typename TeMultiContainerKeyT >
TeMultiContainer< TeMultiContainerKeyT >::TeMultiContainer()
{
  last_up_time_ = 0;
}


template< typename TeMultiContainerKeyT >
TeMultiContainer< TeMultiContainerKeyT >::TeMultiContainer( 
  const TeMultiContainer< TeMultiContainerKeyT >& external )
{
  last_up_time_ = 0;
  
  operator=( external );
}


template< typename TeMultiContainerKeyT >
TeMultiContainer< TeMultiContainerKeyT >::~TeMultiContainer()
{
  clear();
}


template< typename TeMultiContainerKeyT >
void TeMultiContainer< TeMultiContainerKeyT >::clear()
{
  this_lock_instance_.lock();
  
  objects_vector_.clear();
  
  this_lock_instance_.unLock();
}


template< typename TeMultiContainerKeyT >
bool TeMultiContainer< TeMultiContainerKeyT >::operator==( 
  const TeMultiContainer< TeMultiContainerKeyT >& ext_instance ) const
{
  if( last_up_time_ == ext_instance.last_up_time_ ) {
    return true;
  } else {
    return false;
  }
}


template< typename TeMultiContainerKeyT >
bool TeMultiContainer< TeMultiContainerKeyT >::operator!=( 
  const TeMultiContainer< TeMultiContainerKeyT >& ext_instance ) const
{
  if( last_up_time_ == ext_instance.last_up_time_ ) {
    return false;
  } else {
    return true;
  }
}


template< typename TeMultiContainerKeyT >
const TeMultiContainer< TeMultiContainerKeyT >& 
TeMultiContainer< TeMultiContainerKeyT >::operator=( 
  const TeMultiContainer< TeMultiContainerKeyT >& ext_instance )
{
  if( ( &ext_instance ) != this ) {
    this_lock_instance_.lock();
  
    objects_vector_.clear();
    
    ext_instance.this_lock_instance_.lock();
  
    typename VectorT::const_iterator objects_vector_it = 
      ext_instance.objects_vector_.begin();
    typename VectorT::const_iterator objects_vector_it_end = 
      ext_instance.objects_vector_.end();
    
    VectorKeyT auxkey;
          
    while( objects_vector_it != objects_vector_it_end ) {
      auxkey.first = objects_vector_it->first;
      auxkey.second = objects_vector_it->second->clone();
      
      objects_vector_.push_back( auxkey );
          
      ++objects_vector_it;
    }
  
    last_up_time_ = ext_instance.last_up_time_;
  
    ext_instance.this_lock_instance_.unLock();
    
    this_lock_instance_.unLock();
  }

  return *this;
}


template< typename TeMultiContainerKeyT >
void TeMultiContainer< TeMultiContainerKeyT >::update_time()
{
  last_up_time_ = time( 0 );
}


template< typename TeMultiContainerKeyT >
template< typename ObjectT >
void TeMultiContainer< TeMultiContainerKeyT >::store( 
  const TeMultiContainerKeyT& obj_key, const ObjectT& obj_reference )
{
  this_lock_instance_.lock();
  
  typename VectorT::iterator objects_vector_it = 
    objects_vector_.begin();
  typename VectorT::iterator objects_vector_it_end = 
    objects_vector_.end();
  
  TeMCNode< ObjectT >* node_ptr = 0;
  
  while( objects_vector_it != objects_vector_it_end ) {
    if( objects_vector_it->first == obj_key ) {
      node_ptr = dynamic_cast< TeMCNode< ObjectT >* >( 
        objects_vector_it->second.nakedPointer() );
    
      if( node_ptr != 0 ){
        break;
      }
    }
  
    ++objects_vector_it;
  }
  
  if( node_ptr == 0 ) {
    TeSharedPtr< ObjectT > new_obj_ptr( new ObjectT );
    (*new_obj_ptr) = obj_reference;
  
    node_ptr = new TeMCNode< ObjectT >;
    node_ptr->setObjPtr( new_obj_ptr );
    
    VectorKeyT auxkey;
    auxkey.first = obj_key;
    auxkey.second = 
      TeSharedPtr< TeMCNodeInterface >( 
        (TeMCNodeInterface*)node_ptr );
  
    objects_vector_.push_back( auxkey );
  } else {
    (*( node_ptr->getObjPtr() ) ) = obj_reference;
  }
  
  update_time();
  
  this_lock_instance_.unLock();
}


template< typename TeMultiContainerKeyT >
template< typename ObjectT >
bool TeMultiContainer< TeMultiContainerKeyT >::retrive(
  const TeMultiContainerKeyT& obj_key, ObjectT& obj_reference ) const
{
  this_lock_instance_.lock();
  
  typename VectorT::const_iterator objects_vector_it = 
    objects_vector_.begin();
  typename VectorT::const_iterator objects_vector_it_end = 
    objects_vector_.end();
  
  TeMCNode< ObjectT >* node_ptr = 0;
  
  while( objects_vector_it != objects_vector_it_end ) {
    if( objects_vector_it->first == obj_key ) {
      node_ptr = dynamic_cast< TeMCNode< ObjectT >* >( 
        objects_vector_it->second.nakedPointer() );
    
      if( node_ptr != 0 ){
        break;
      }
    }
  
    ++objects_vector_it;
  }
  
  if( node_ptr == 0 ) {
    TEAGN_LOGWARN( "Trying to retrive an object not stored" );
    
    this_lock_instance_.unLock();
      
    return false;
  } else {
    obj_reference = ( *( node_ptr->getObjPtr() ) );
    
    this_lock_instance_.unLock();
    
    return true;
  }
}


template< typename TeMultiContainerKeyT >
template< typename ObjectT >
void TeMultiContainer< TeMultiContainerKeyT >::multiRetrive(
  std::vector< std::pair< TeMultiContainerKeyT, 
  ObjectT > >& objs_vector ) const
{
  objs_vector.clear();
  
  this_lock_instance_.lock();
  
  typename VectorT::const_iterator objects_vector_it = 
    objects_vector_.begin();
  typename VectorT::const_iterator objects_vector_it_end = 
    objects_vector_.end();
  
  TeMCNode< ObjectT >* node_ptr = 0;
  
  while( objects_vector_it != objects_vector_it_end ) {
    node_ptr = dynamic_cast< TeMCNode< ObjectT >* >( 
      objects_vector_it->second.nakedPointer() );
  
    if( node_ptr != 0 ){
      std::pair< TeMultiContainerKeyT, ObjectT > temp_pair;
      temp_pair.first = objects_vector_it->first;
      temp_pair.second = ( *( node_ptr->getObjPtr() ) );
    
      objs_vector.push_back( temp_pair );
    }
  
    ++objects_vector_it;
  }
  
  this_lock_instance_.unLock();
}


template< typename TeMultiContainerKeyT >
template< typename ObjectT >
bool TeMultiContainer< TeMultiContainerKeyT >::isStored( 
  const TeMultiContainerKeyT& obj_key ) const
{
  this_lock_instance_.lock();
  
  typename VectorT::const_iterator objects_vector_it = 
    objects_vector_.begin();
  typename VectorT::const_iterator objects_vector_it_end = 
    objects_vector_.end();
  
  TeMCNode< ObjectT >* node_ptr = 0;
  
  while( objects_vector_it != objects_vector_it_end ) {
    if( objects_vector_it->first == obj_key ) {
      node_ptr = dynamic_cast< TeMCNode< ObjectT >* >( 
        objects_vector_it->second.nakedPointer() );
    
      if( node_ptr != 0 ){
        this_lock_instance_.unLock();
      
        return true;
      }
    }
  
    ++objects_vector_it;
  }
  
  this_lock_instance_.unLock();
  
  return false;
}

/** @example TeMultiContainer_test.cpp
 *    Shows how to use this class.
 */  

#endif

