#include "TeThread.h"
#include "TeAgnostic.h"

  #if TePLATFORM == TePLATFORMCODE_MSWINDOWS

  #elif TePLATFORM == TePLATFORMCODE_LINUX

  #else
    #error "ERROR: Unsupported platform"
  #endif

TeThread::TeThread( TeThreadStartFunctT start_func_ptr )
{
  TEAGN_TRUE_OR_THROW( ( start_func_ptr != 0 ),
    "Invalid thread start function pointer" )
  start_func_ptr_ = start_func_ptr;
  
  curr_status_ = TeThreadStopped;
  curr_return_value_ = false;
  curr_priority_ = TeThreadNormalPriority;
  
  #if TePLATFORM == TePLATFORMCODE_MSWINDOWS
    thread_handler_ = 0;
  #elif TePLATFORM == TePLATFORMCODE_LINUX
    thread_id_ = 0;
  #else
    #error "ERROR: Unsupported platform"
  #endif  
}


TeThread::~TeThread()
{
  TEAGN_TRUE_OR_THROW( ( curr_status_ == TeThreadStopped ),
    "Trying to delete an running thread handler instance" )
    
  freeResources();
}


bool TeThread::start( const TeThreadParameters& params )
{
  if( curr_status_ != TeThreadStopped ) {
    return false;
  } else {
    freeResources();
    
    curr_return_value_ = false;
    params_ = params;
    
    internal_thread_exec_params_.thread_user_params_ptr_ = &params_;
    internal_thread_exec_params_.thread_start_func_ptr_ = start_func_ptr_;
    internal_thread_exec_params_.thread_status_ptr_ = &curr_status_;
    internal_thread_exec_params_.thread_return_value_ptr_ = 
      &curr_return_value_;
    
    #if TePLATFORM == TePLATFORMCODE_MSWINDOWS
      LPDWORD thread_id = 0;

      thread_handler_ = 0;
      thread_handler_ = CreateThread(NULL, 0,
        (LPTHREAD_START_ROUTINE)startFunctWrapper,
        (LPVOID)(&internal_thread_exec_params_), 0, (LPDWORD) &thread_id);
      TEAGN_TRUE_OR_RETURN( ( thread_handler_ != 0 ),
        "Thread creation failed" )
    #elif TePLATFORM == TePLATFORMCODE_LINUX
      TEAGN_TRUE_OR_THROW( ( pthread_attr_init(&thread_attr_) == 0 ),
        "Unable to init thread attributes" )
      TEAGN_TRUE_OR_THROW( ( pthread_attr_setdetachstate( &thread_attr_, 
        PTHREAD_CREATE_JOINABLE ) == 0 ),
        "Unable to set thread detach attribute" );
      TEAGN_TRUE_OR_THROW( ( pthread_attr_setschedpolicy( &thread_attr_, 
        SCHED_RR ) == 0 ),
        "Unable to set thread scheddule policy attribute" );   
             
      int pol = 0;
      TEAGN_DEBUG_CONDITION( ( pthread_attr_getschedpolicy( &thread_attr_, 
        &pol ) == 0 ),
        "Unable to get thread scheddule policy attribute" );
      TEAGN_DEBUG_CONDITION( ( pol == SCHED_RR ),
        "Invalid thread scheddule policy attribute" );     
        
      TEAGN_TRUE_OR_RETURN( ( pthread_create( &thread_id_, &thread_attr_, 
        startFunctWrapper, (void *)(&internal_thread_exec_params_) ) == 0 ),
        "Thread creation failed" )
    #else
      #error "ERROR: Unsupported platform"
    #endif
    
    curr_status_ = TeThreadRunning;
    
    if( curr_priority_ != TeThreadNormalPriority ) {
      TEAGN_TRUE_OR_LOG( setPriority( curr_priority_ ),
        "Unable to set thread priority" );
    }
  
    return true;  
  }
}


TeThread::TeThreadStatus TeThread::getStatus() const
{
  return curr_status_;
}


bool TeThread::getReturnValue() const
{
  return curr_return_value_;
}


TeThread::TeThreadPriority TeThread::getPriority() const
{
  return curr_priority_;
}


bool TeThread::setPriority( TeThread::TeThreadPriority new_priority )
{
  if( curr_status_ == TeThreadRunning ) {
    #if TePLATFORM == TePLATFORMCODE_MSWINDOWS
      int new_prio_value = THREAD_PRIORITY_NORMAL;
      
      switch( new_priority ) {
        case TeThreadIdlePriority :
        {
          new_prio_value = THREAD_PRIORITY_IDLE;
          break;
        }
        case TeThreadBelowNormalPriority :
        {
          new_prio_value = THREAD_PRIORITY_BELOW_NORMAL;
          break;
        }
        case TeThreadNormalPriority :
        {
          new_prio_value = THREAD_PRIORITY_NORMAL;
          break;
        }
        case TeThreadAboveNormalPriority :
        {
          new_prio_value = THREAD_PRIORITY_ABOVE_NORMAL;
          break;
        }
        case TeThreadTimeCriticalPriority :
        {
          new_prio_value = THREAD_PRIORITY_TIME_CRITICAL;
          break;
        }
        default :
        {
          TEAGN_LOG_AND_THROW( "Invalid thread priority" )
          break;
        }                                      
      }
      
      if( SetThreadPriority( thread_handler_, new_prio_value ) ) {
        curr_priority_ = new_priority;
        return true;
      } else {
        return false;
      }      
    #elif TePLATFORM == TePLATFORMCODE_LINUX
      int new_prio_value = 0;
      
      int min_prio = sched_get_priority_min( SCHED_RR );
      int max_prio = sched_get_priority_max( SCHED_RR );
      
      switch( curr_priority_ ) {
        case TeThreadIdlePriority :
        {
          new_prio_value = min_prio;
          break;
        }
        case TeThreadBelowNormalPriority :
        {
          new_prio_value = (int)( ( max_prio - min_prio ) / 4 );
          break;
        }
        case TeThreadNormalPriority :
        {
          new_prio_value = (int)( 2 * ( max_prio - min_prio ) / 4 );
          break;
        }
        case TeThreadAboveNormalPriority :
        {
          new_prio_value = (int)( 3 * ( max_prio - min_prio ) / 4 );
          break;
        }
        case TeThreadTimeCriticalPriority :
        {
          new_prio_value = max_prio;
          break;
        }
        default :
        {
          TEAGN_LOG_AND_THROW( "Invalid thread priority" )
          break;
        }                                      
      }
             
      int curr_policy = SCHED_RR;
      struct sched_param curr_sched_param;
      curr_sched_param.sched_priority = new_prio_value;
      
      if( pthread_setschedparam( thread_id_, curr_policy,
        &curr_sched_param ) == 0 ) {
        
        curr_priority_ = new_priority;
        return true;        
      } else {
        return false;
      }  
    #else
      #error "ERROR: Unsupported platform"
    #endif    
  } else {
    /* Thread not running */
    
    curr_priority_ = new_priority;
    
    return true;
  }
}


bool TeThread::waitToFinish()
{
  if( curr_status_ != TeThreadStopped ) {
    #if TePLATFORM == TePLATFORMCODE_MSWINDOWS
      TEAGN_TRUE_OR_RETURN( ( WaitForSingleObjectEx( 
        thread_handler_, INFINITE, false ) == WAIT_OBJECT_0 ),
        "Thread joinning failed" )
    #elif TePLATFORM == TePLATFORMCODE_LINUX
      TEAGN_TRUE_OR_RETURN( ( pthread_join( thread_id_, 0 ) == 0 ),
        "Thread joinning failed" )
      thread_id_ = 0;
    #else
      #error "ERROR: Unsupported platform"
    #endif  

    freeResources();
  }
  
  return curr_return_value_;
}


void TeThread::freeResources()
{
  #if TePLATFORM == TePLATFORMCODE_MSWINDOWS
    if( thread_handler_ != 0 ) {
      WaitForSingleObjectEx( thread_handler_, INFINITE, false );
            
      TEAGN_TRUE_OR_THROW( CloseHandle( thread_handler_ ),
        "Error closing thread handle" );
        
      thread_handler_ = 0;
    }
  #elif TePLATFORM == TePLATFORMCODE_LINUX
    if( thread_id_ != 0 ) {
      pthread_join( thread_id_, 0 );
      
      pthread_cancel( thread_id_ );
      
      thread_id_ = 0;
      
      TEAGN_TRUE_OR_THROW( ( pthread_attr_destroy( &thread_attr_ ) == 0 ),
        "Unable to destroy thread attributes" )        
    } 
  #else
    #error "ERROR: Unsupported platform"
  #endif  
}


void* TeThread::startFunctWrapper( void* params_ptr )
{
  TEAGN_DEBUG_CONDITION( (params_ptr != 0), 
    "Invalid thread exec params pointer" )
  TeThreadExecParams* thread_exec_params_ptr = 
    (TeThreadExecParams*)params_ptr;
 
  *( thread_exec_params_ptr->thread_return_value_ptr_ ) =
    thread_exec_params_ptr->thread_start_func_ptr_( 
    *( thread_exec_params_ptr->thread_user_params_ptr_ ) );
    
  thread_exec_params_ptr->thread_user_params_ptr_->clear();
  
  *( thread_exec_params_ptr->thread_status_ptr_ ) = TeThreadStopped;
  
  return 0;
}


TeThread::TeThread( const TeThread& )
{
}


const TeThread& TeThread::operator=( const TeThread& external )
{
  return external;
}

