/************************************************************************************
TerraLib - a library for developing GIS applications.
Copyright  2001-2007 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular purpose.
The library provided hereunder is on an "as is" basis, and the authors have no
obligation to provide maintenance, support, updates, enhancements, or modifications.
In no event shall INPE and Tecgraf / PUC-Rio be held liable to any party for direct,
indirect, special, incidental, or consequential damages arising out of the use
of this library and its documentation.
*************************************************************************************/
/*! \file TeThread.h
   \brief Thread class.
*/

#ifndef TETRHEAD_H
  #define TETRHEAD_H
  
  #include "TeThreadParameters.h"
  #include "TeSharedPtr.h"
  #include "TeDefines.h"
  
  #if TePLATFORM == TePLATFORMCODE_MSWINDOWS
    #include <windows.h>
    
    #define TeDelayThread( seconds ) Sleep( seconds * 1000)
  #elif TePLATFORM == TePLATFORMCODE_LINUX
    #include <pthread.h>
    #include <unistd.h>
    
    #define TeDelayThread( seconds ) sleep( seconds )
  #else
    #error "ERROR: Unsupported platform"
  #endif
  
  /**
   * @brief Thread handler class.
   * @author Emiliano F. Castejon <castejon@dpi.inpe.br>
   * @ingroup Utils
   */
  class TeThread {
    public :

      /**
       * @brief Type definition for a thread start function pointer.
       * @param params The thread parameters.
       * @return true if OK, false on erros.
       */      
      typedef bool (*TeThreadStartFunctT)( TeThreadParameters const& params );
      
      /** @typedef TeSharedPtr< TeThread > pointer
       * Type definition for an thread instance pointer. 
       */
      typedef TeSharedPtr< TeThread > pointer;      
              
      /**
       * @enum TeThreadStatus Possible thread status.
       */
      enum TeThreadStatus {
        /** @brief The thread is stoped. */
        TeThreadStopped,
        /** @brief The thread is running. */
        TeThreadRunning
      };
      
      /**
       * @enum TeThreadPriority Possible thread priorities.
       */
      enum TeThreadPriority {
        /** @brief Idle priority. */
        TeThreadIdlePriority,
        /** @brief Below normal priority. */
        TeThreadBelowNormalPriority,
        /** @brief Normal priority. */
        TeThreadNormalPriority,        
        /** @brief Above normal priority. */
        TeThreadAboveNormalPriority,       
        /** @brief Time critical priority. */
        TeThreadTimeCriticalPriority                 
      };      
      
      /**
       * @brief Default Constructor.
       */
      TeThread( TeThreadStartFunctT start_func_ptr );    

      /**
       * @brief Default Destructor
       */
      ~TeThread();
      
      /**
       * @brief Start the current thread following the supplied parameters.
       * @param params The thread parameters.
       * @return true if OK, false on errors.
       */      
      bool start( const TeThreadParameters& params );
      
      /**
       * @brief Get the current thread status.
       * @return The current thread status.
       */      
      TeThreadStatus getStatus() const;

      /**
       * @brief Returns the current thread execution return value.
       */      
      bool getReturnValue() const;
      
      /**
       * @brief Returns the current thread priority.
       * @return The current thread priority.
       */      
      TeThreadPriority getPriority() const;      
      
      /**
       * @brief Change the current thread priority.
       * @param new_priority The new thread priority.
       * @return true if OK, false on errors.
       */      
      bool setPriority( TeThreadPriority new_priority );       
            
      /**
       * @brief Block the current thread until the thread 
       * handled by this instance terminates.
       * @return true if OK, false on errors.
       */      
      bool waitToFinish();

    protected :
    
      /**
       * @brief Internal thread execution parameters.
       */     
      class TeThreadExecParams {
        public :
        
        /** @brief A pointer to the current user parameters instance. */
        TeThreadParameters* thread_user_params_ptr_;
        
        /** @brief A pointer to the current user thread start function. */
        TeThreadStartFunctT thread_start_func_ptr_;
        
        /** @brief A pointer to the current thread status variable. */
        TeThreadStatus* thread_status_ptr_;
        
        /** @brief A pointer to the current thread exit value variable. */
        bool* thread_return_value_ptr_;
        
        /** @brief Default constructor. */
        TeThreadExecParams()
        {
          thread_user_params_ptr_ = 0;
          thread_start_func_ptr_ = 0;
          thread_status_ptr_ = 0;
          thread_return_value_ptr_ = 0;        
        }
      };

      /**
       * @brief The current threat start functio pointer.
       */       
      TeThreadStartFunctT start_func_ptr_;
     
      /**
       * @brief The current threat parameters.
       */       
      TeThreadParameters params_;
     
      /**
       * @brief The current threat execution return value.
       */       
      bool curr_return_value_;
      
      /**
       * @brief The current threat status.
       */        
      TeThreadStatus curr_status_;
      
      /**
       * @brief A internal thread start params instance.
       */      
      TeThreadExecParams internal_thread_exec_params_;
      
      /**
       * @brief The current thread priority.
       */       
      TeThreadPriority curr_priority_;
      
      #if TePLATFORM == TePLATFORMCODE_MSWINDOWS
        /** @brief Thread handler. */
        HANDLE thread_handler_;
        
      #elif TePLATFORM == TePLATFORMCODE_LINUX
        /** @brief Pthread thread ID. */
        pthread_t thread_id_;
        
        /** @brief Pthread thread attributes. */
        pthread_attr_t thread_attr_;
      #else
        #error "ERROR: Unsupported platform"
      #endif    

      /**
       * @brief Free all thread allocated resources.
       */      
      void freeResources();

      /**
       * @brief A thread start function wrrapper.
       * @param params_ptr A void casted pointer to a const thread parameters
       * instance.
       */      
      static void* startFunctWrapper( void* params_ptr );
      
      
    private :
    
      /**
       * @brief Alternative Constructor.
       * @param external External reference.
       */
      TeThread( const TeThread& external );      
    
      /**
       * @brief Operator = overload.
       * @param external External reference.
       * @return A const reference.
       */
      const TeThread& operator=( const TeThread& external );
  };
  
#endif

