/************************************************************************************
TerraLib - a library for developing GIS applications.
Copyright  2001-2007 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular purpose.
The library provided hereunder is on an "as is" basis, and the authors have no
obligation to provide maintenance, support, updates, enhancements, or modifications.
In no event shall INPE and Tecgraf / PUC-Rio be held liable to any party for direct,
indirect, special, incidental, or consequential damages arising out of the use
of this library and its documentation.
*************************************************************************************/
/*! \file TeThreadLock.h
  \brief This file contains definitions about a class to deal with critical 
  region locking.
*/


#ifndef TETHREADLOCK_H
  #define TETHREADLOCK_H
  
  #include "TeDefines.h"
  
  #if TePLATFORM == TePLATFORMCODE_MSWINDOWS
    #include <windows.h>
  #elif TePLATFORM == TePLATFORMCODE_LINUX
    #include <pthread.h>
    #include  <errno.h>
  #else
    #error "ERROR: Unsupported platform"
  #endif  
 
  /**
   * @brief A class to deal with critical region locking.
   * @author Emiliano F. Castejon <castejon@dpi.inpe.br>
   */
  class TeThreadLock
  {
    public :
    
      /**
       * @brief Default constructor.
       */    
      TeThreadLock();

      /**
       * @brief Default destructor.
       */       
      ~TeThreadLock();
      
      /**
       * @brief Lock the current object instance.
       * @note If section is already busy then the current thread will be 
       * blocked until it's ready again.
       */       
      void lock();      
      
      /**
       * @brief Try to lock the current object instance.
       * @return true if OK, false if unable to lock.
       * @note If section is busy, this method will return false 
       * without blocking the current thread.
       */       
      bool tryLock();      
      
      /**
       * @brief Unlock the current object instance.
       */       
      void unLock();      
            
    protected :
   
      #if TePLATFORM == TePLATFORMCODE_MSWINDOWS
    
        /**
        * @brief The mutex instance.
        */      
        HANDLE m_access_;
        
        /**
        * @brief The mutex instance attributes.
        */           
        SECURITY_ATTRIBUTES m_sa_;
      
      #elif TePLATFORM == TePLATFORMCODE_LINUX
      
        /**
        * @brief The mutex instance.
        */        
        pthread_mutex_t m_access_;
      
      #else
        #error "Unsuported plataform"
      #endif   
   
    private :
    
      /**
       * Alternative constructor.
       */    
      TeThreadLock(  const TeThreadLock& ) {};
    
   
      /**
       * operator= overload.
       * @return A const reference to the current instance.
       */      
      const TeThreadLock& operator=( const TeThreadLock& ) { return *this; };
 
  };

#endif
