/************************************************************************************
TerraLib - a library for developing GIS applications.
Copyright  2001-2007 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular purpose.
The library provided hereunder is on an "as is" basis, and the authors have no
obligation to provide maintenance, support, updates, enhancements, or modifications.
In no event shall INPE and Tecgraf / PUC-Rio be held liable to any party for direct,
indirect, special, incidental, or consequential damages arising out of the use
of this library and its documentation.
*************************************************************************************/
/*! \file TeVisual.h
    \brief This file deals with the visual presentation characteristics of geometries in TerraLib
*/
#ifndef  __TERRALIB_INTERNAL_VISUAL_H
#define  __TERRALIB_INTERNAL_VISUAL_H

#include "TeDataTypes.h"
#include "TeAbstractFactory.h"

class TeRasterTransform;

#include <string>
#include <map>



//! A structure for supporting a color definition
struct TeColor
{
	//! Red component
	int red_;

	//! Green component 
	int green_; 	
	
	//! Blue component
	int blue_;	
	
	//! Color name
	std::string name_;
	
	//! Empty constructor
	TeColor () : red_(0),green_(0),blue_(0), name_("") {}

	//! Constructor with parameters
	TeColor (int r, int g, int b, const std::string& name="") : red_(r),green_(g),blue_(b), name_(name) {}

	//! Set parameters of colors
	void init (int r, int g, int b, const std::string& name="") {red_=r,green_=g,blue_=b;name_=name; }


	//! Returns TRUE if color1 is equal to color2 or FALSE if they are different.
	bool operator== (const TeColor& color)
	{
		return (red_==color.red_ && green_==color.green_ && blue_==color.blue_);
	}

	//! Assignment operator
	TeColor& operator= ( const TeColor& color )
	{
		if ( this != &color )
		{	
			red_ = color.red_;
			green_ = color.green_;
			blue_ = color.blue_;
			name_ = color.name_;
		}
		return *this;
	}
};


//! This class is used to store the standard presentation characteristics of vector geometries
class TeVisual
{
public :
	int			symbId_;			//!< Symbol id that define the style for the geometry filling
	TeColor		color_;				//!< Color used for geometry filling (line, point and text color)
	int			transparency_;		//!< Percentage of transparency for the geometry filling
	int			width_;				//! Width for line geometry
	
	int			contourSymbId_;			//!< Symbol id that define the style for the geometry outline
	TeColor		contourColor_;			//!< Contour color for the geometry outline
	int			contourTransparency_;	//!< Percentage of transparency for the geometry outline
	int			contourWidth_;			//!< Width for the geometry outline

	int			sizeValue_;				//!< Point size
	int			ptAngle_;				//!< Point angle
	
	// Text attributes
	std::string		family_;		//!< Text family
	bool		bold_;				//!< Flag to indicate if text is bold
	bool		italic_;			//!< Flag to indicate if text is italic
	double		alignmentVert_;		//!< Position of vertical alignment
	double		alignmentHoriz_;	//! Position of horizontal alignment
	int			tabSize_;			//!< Number of spaces defined in a tab character
	int			lineSpace_;			//!< Spacing between lines
	bool		fixedSize_;			//!< Flag to indicate that element has a fixed size, dont follow zoom operations	

	std::string		visualType_;		//!< Visual type used in factory strategy
	TeGeomRep   geomRep_;				//!< Geometric representation associated with this visual 

	//! Empty constructor
	TeVisual(TeGeomRep rep = TePOLYGONS);

	//! Destructor
	virtual ~TeVisual() {}

	//! Copy constructor
	TeVisual(const TeVisual& other);

	//! Assignment operator
	virtual TeVisual& operator=(const TeVisual& vis);

	//! Returns TRUE if param1 is equal to param2 or FALSE if they are different.
	virtual bool operator== (const TeVisual& vis);

	//! Returns visual to the default parameters 
	virtual void clear();

	//! Copies only the basic visual parameters from another visual
	virtual void setBasic (TeVisual &vis);

	//! Sets the color used for geometry filling and line, point and text color
	virtual void color(TeColor& color) 
	{	color_ = color;}

	//! Returns the color used for geometry filling and line, point and text color
	virtual TeColor& color() 
	{ return color_; }

	//! Sets the style for the filling of geometries
	virtual void style(int s) 
	{ symbId_ = s; }

	//! Returns the style for the filling of geometries
	virtual int	style() 
	{ return symbId_; }

	//! Sets the percentage of transparency for polygons and cells
	virtual void transparency(int t)
	{ transparency_ = t; }

	//! Returns the percentage of transparency used in polygons and cells
	virtual int transparency()
	{ return transparency_; }

	//! Sets the color for the polygons and cells outlines
	virtual void contourColor(TeColor& color) 
	{ contourColor_ = color;}

	//! Returns the color used in the polygons and cells outlines
	virtual TeColor& contourColor() 
	{ return contourColor_; }

	//! Sets the percentage of transparency for the polygon ans cells outlines
	virtual void contourTransparency(int t) 
	{ contourTransparency_ = t;}

	//! Returns the color used in the polygons and cells outlines
	virtual int	contourTransparency() 
	{ return contourTransparency_; }

	//! Sets the contour width for the polygons and cells outlines
	virtual void contourWidth(int w) 
	{ contourWidth_ = w; }

	//! Returns the contour width used in the polygons and cells outlines
	virtual int	contourWidth()  
	{ return contourWidth_; }

	//! Sets the contour style for the polygons and cells outlines
	virtual void contourStyle(int s) 
	{ contourSymbId_ = s; }

	//! Returns the contour style used in the polygons and cells outlines
	virtual int	contourStyle() 
	{ return contourSymbId_; }

	//! Sets the width
	virtual void width(int w) 
	{ width_ = w; }

	//! Returns the width
	virtual int	 width ()  
	{ return width_; }

	//! Sets the point size
	virtual void size (int s)  
	{ sizeValue_ = s; }

	//! Returns the point size
	virtual int	size ()  
	{ return sizeValue_; }

	//! Sets the text family for fonts
	virtual void family (std::string f) 
	{ family_ = f; }

	//! Returns fonts family
	virtual std::string family () 
	{ return family_ ; }

	//! Sets the bold style for texts
	virtual void bold (bool b) 
	{ bold_ = b; }

	//! Returns the bold style used in the texts
	virtual bool bold () 
	{ return bold_; }

	//! Sets the italic style for texts
	virtual void italic (bool i) 
	{ italic_ = i; }

	//! Returns the italic style used in the texts
	virtual bool italic () 
	{ return italic_; }

	//!  Sets a flag indicating if the text will have a fixed size
	virtual void fixedSize (bool i) 
	{ fixedSize_ = i; }

	//!  Retuns a flag indicating if an text have a fixed size
	virtual bool fixedSize () 
	{ return fixedSize_; }

	//!  Sets the vertical alignment of texts
	virtual void alignmentVert(double a) 
	{ alignmentVert_ = a; }

	//!  Retuns the vertical alignment of texts
	virtual double alignmentVert() 
	{ return alignmentVert_; }

	//!  Sets the horizontal alignment of texts
	virtual void alignmentHoriz(double a) 
	{ alignmentHoriz_ = a; }

	//!  Retuns the horizontal alignment of texts
	virtual double alignmentHoriz() 
	{ return alignmentHoriz_; }

	//! Sets the number of spaces defined in a tab character
	virtual void tabSize (int s) 
	{ tabSize_ = s; }

	//! Returns the number of spaces defined in a tab character
	virtual int	tabSize () 
	{ return tabSize_; }

	//! Sets the spacing between lines
	virtual void lineSpace (int s) 
	{ lineSpace_ = s; }

	//! Returns the spacing between lines
	virtual int	lineSpace () 
	{ return lineSpace_; }

	//! Returns the angle of points and symbols
	virtual int   ptAngle() 
	{return ptAngle_; }

	//! Sets the angle of points and symbols
	virtual void  ptAngle(int p) 
	{ptAngle_ = p;}

	//! Creates another visual pointer and copies its content
	virtual TeVisual* copy()
	{
		TeVisual* retval = new TeVisual();
		*retval = *this;
		return retval;
	}

	//! Returns a default visual object
	static TeVisual* DefaultObject()  
	{ return new TeVisual(); }
 
};

//! A factory to build TeVisual objects
class TeVisualFactory : public TeAbstractFactory<TeVisual,std::string>
{
public:
	TeVisualFactory(const std::string& facName) : TeAbstractFactory<TeVisual,std::string>(facName)
	{}
		virtual TeVisual* build ()
	{	return new TeVisual(); }
};

//! Export the type that represents the definition of visual for raster data
typedef TeRasterTransform TeRasterVisual;

#endif

